
/******************************************************************************
* MODULE     : multiply.gen.cc
* DESCRIPTION: multiplication of relaxed power series
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#module code_relaxed_multiply (C)

/******************************************************************************
* Naive multiplication
******************************************************************************/

struct product1_series_rep<C>: public series_rep<C> {
  series<C> f, g;
  product1_series_rep<C> (series<C> f2, series<C> g2):
    series_rep<C> (RELAXED), f(f2), g(g2) {}
  C next () {
    // cout << "Coefficient " << n << "\n";
    int i;
    C c (ZERO<C>);
    for (i=0; i<=n; i++) c += f[i]*g[n-i];
    // cout << "  -> " << c << "\n";
    return c;
  }
  operator tree () { return tree ("*", f, g); }
};

/******************************************************************************
* Karatsuba multiplication
******************************************************************************/

struct Karatsuba_rep<C>: public series_rep<C> {
  series<C>    f, g, lo, mid, hi;
  int          N;
  Karatsuba_rep<C> (series<C> f2, series<C> g2, int N2):
    series_rep<C> (RELAXED), f(f2), g(g2), N(N2) {
      value= zealous<C> (N<<1); }
  Karatsuba_rep<C> (series<C> f2, series<C> g2, series<C> h, int N2):
    series_rep<C> (RELAXED), f(f2), g(g2), N(N2) {
      lo   = h;
      value= h (0, N<<1);
      n    = N>>1; }
  C next ();
  operator tree () { return tree ("*", f, g); }
  friend class Karatsuba<C>;
  friend class product2_series_rep<C>;
};

series<C>
Karatsuba<C> (series<C> f, series<C> g, int N) {
  series<C> P= new Karatsuba_rep<C> (f, g, N);
  /*
  zealous<C> PP= P(0,2*N-1);
  cout << "f= " << f(0,N) << "\n";
  cout << "g= " << g(0,N) << "\n";
  cout << "P= " << PP << "\n";
  */
  return P;
}

series<C>
Karatsuba<C> (series<C> f, series<C> g, series<C> h, int N) {
  series<C> P= new Karatsuba_rep<C> (f, g, h, N);
  /*
  zealous<C> PP= P(0,2*N-1);
  cout << "f= " << f(0,N) << "\n";
  cout << "g= " << g(0,N) << "\n";
  cout << "P= " << PP << "\n";
  */
  return P;
}

C
Karatsuba_rep<C>::next () {
  int i, h= N>>1;

  if (N<Karatsuba_threshold<C>) {
    C c (0);
    if (n<N) for (i=0; i<=n; i++) c += f[i] * g[n-i];
    else for (i=n-(N-1); i<=N-1; i++) c += f[i] * g[n-i];
    return c;
  }

  if (n < h) {
    if (n==0) lo= Karatsuba<C> (f, g, N>>1);
    return lo[n];
  }

  if (n < N) {
    if (n==h) {
      series<C> ff= f>>h;
      series<C> gg= g>>h;
      mid= Karatsuba<C> (f+ff, g+gg, N>>1);
      hi = Karatsuba<C> (ff, gg, N>>1);
    }
    return lo[n]+ mid[n-h]- lo[n-h]- hi[n-h];
  }

  if (n == N) {
    for (i=0; i<N-1; i++) value[i+N]= hi[i];
    value->accumulate (mid (h, n-1)- lo (h, n-1)- hi (h, n-1), n);
    lo = series<C> ();
    mid= series<C> ();
    hi = series<C> ();
  }

  return value[n];
}

struct product2_series_rep<C>: public series_rep<C> {
  series<C> f, g, h;
  product2_series_rep<C> (series<C> f2, series<C> g2):
    series_rep<C> (RELAXED), f(f2), g(g2),
    h (Karatsuba<C> (f, g, Karatsuba_threshold<C>)) {}
  C next ();
  operator tree () { return tree ("*", f, g); }
};

C
product2_series_rep<C>::next () {
  if (n >= Karatsuba_threshold<C>) {
    int k=n;
    while ((k&1)==0) k= k>>1;
    if (k==1) h= Karatsuba<C> (f, g, h, n << 1);
  }
  return ((series<C>) h) [n];
}

/******************************************************************************
* Fast relaxed multiplication
******************************************************************************/

struct product3_series_rep<C>: public series_rep<C> {
  series<C> f, g;
  product3_series_rep<C> (series<C> f2, series<C> g2):
    series_rep<C> (RELAXED), f(f2), g(g2) {}
  C next ();
  operator tree () { return tree ("*", f, g); }
};

C
product3_series_rep<C>::next () {
  // cout << "Coefficient " << n << "\n";
  if ((2*n) >= N(value)) {
    int k=1;
    while (k <= (2*n)) k= k<<1;
    value= value (0, k);
  }

  int k= 2*(n+2), p= -1;
  while ((k&1)==0) {
    k= k>>1; p++;
    value->accumulate (multiply (f ((1<<p)-1, (2<<p)-1),
				 g (((k-1)<<p)-1, (k<<p)-1)),
		       (k<<p)-2);
    if (k==2) return value[n];
    value->accumulate (multiply (f (((k-1)<<p)-1, (k<<p)-1),
				 g ((1<<p)-1, (2<<p)-1)),
		       (k<<p)-2);
  }
  // cout << "-> " << value[n] << "\n";
  return value[n];
}

/******************************************************************************
* Variant of fast relaxed multiplication
******************************************************************************/

struct product4_series_rep<C>: public series_rep<C> {
  series<C> f, g;
  zealous<C> diag;
  product4_series_rep<C> (series<C> f2, series<C> g2):
    series_rep<C> (RELAXED), f(f2), g(g2), diag (0) {}
  C next ();
  void accumulate (int a, int b, int p, bool flag);
  operator tree () { return tree ("*", f, g); }
};

/*
static int indent= 0;
static string
spaces () {
  int i;
  string s(indent);
  for (i=0; i<indent; i++) s[i]= ' ';
  return s;
}
*/

C
product4_series_rep<C>::next () {
  // cout << "Coefficient " << n << "\n";
  if ((2*n) >= N(value)) {
    int k=1;
    while (k <= (2*n)) k= k<<1;
    value= value (0, k);
  }
  if (N(diag)<N(value)) diag= diag (0, N(value));

  C c= f[n]*g[n];
  value[n<<1] += c;
  diag [n<<1] += c;
  int k= n+2, p= 0;
  while ((k&1)==0) { k=k>>1; p++; }
  if (k==5) accumulate (2, 3, p, TRUE);

  k= 2*(n+2); p= -1;
  while (((k&1)==0) && (k!=4)) {
    k= k>>1; p++;
    if (p>0) accumulate ((k<<1)-2, (k<<1)-1, p-1, TRUE);
    accumulate (1, k-1, p, FALSE);
  }
  // cout << "-> " << value[n] << "\n";
  return value[n];
}

void
product4_series_rep<C>::accumulate (int a, int b, int p, bool flag) {
  // cout << "  Accumulate " << a << ", " << b << " [" << (1<<p) << "]\n";
  int i= (a<<p)-1, j= (b<<p)-1, k= 1<<p;
  zealous<C> t= multiply (f(i,i+k)+ f(j,j+k), g(i,i+k)+ g(j,j+k))
    -diag (i<<1, ((i+k)<<1)-1)- diag (j<<1, ((j+k)<<1)-1);
  value->accumulate (t, i+j);
  if (flag) diag->accumulate (t, i+j);
}

/******************************************************************************
* interface
******************************************************************************/

series<C>
operator * (series<C> f, series<C> g) {
  return new product3_series_rep<C> (f, g);
}

#endmodule // code_relaxed_multiply (C)
