
/******************************************************************************
* MODULE     : zealous.gen.cc
* DESCRIPTION: zealous power series
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <Univariate/zealous.gen.h>

#module code_zealous (C)
#import zealous (C)

void FFT_mul (C* R, C* P, C* Q, int n);

/******************************************************************************
* Truncation and shifting
******************************************************************************/

void
zealous_rep<C>::accumulate (zealous<C> f, int at) {
  int i;
  for (i=0; i<f->n; i++)
    a[i+at] += f[i];
}

zealous<C>
zealous<C>::operator () (int start, int end) {
  int i;
  zealous<C> f (end-start);
  for (i=start; i< (end <? rep->n); i++)
    f[i-start]= rep->a[i];
  return f;
}

zealous<C>
constant (C c, int k) {
  zealous<C> f (k);
  if (k>0) f[0]= c;
  return f;
}

zealous<C>
id (int n) {
  zealous<C> f (n);
  f[1]= 1;
  return f;
}

zealous<C>
operator << (zealous<C> f, int k) {
  int i, n (N(f));
  zealous<C> g (n+k);
  for (i=0; i<n; i++) g[i+k]= f[i];
  return g;
}

zealous<C>
operator >> (zealous<C> f, int k) {
  int i, n (N(f));
  zealous<C> g (n-k);
  for (i=0; i<n-k; i++) g[i]= f[i+k];
  return g;
}

/******************************************************************************
* Fast multiplication
******************************************************************************/

static void
mul (C* R, C* P, C* Q, int n) {
#ifdef Karatsuba_threshold<C>
  if (n<Karatsuba_threshold<C>) {
    // cout << "Full Naive " << n << "\n";
    int i, j;
    for (i=0; i<((n<<1)-1); i++) R[i]=0;
    for (i=0; i<n; i++)
      for (j=0; j<n; j++)
	R[i+j]= R[i+j]+ P[i]*Q[j];
    // cout << "Done\n";
    return;
  }
#else
  switch (n) {
  case 0:
    return;
  case 1:
    R[0]= P[0]*Q[0];
    return;
  case 2:
    R[0] = P[0]*Q[0];
    R[2] = P[1]*Q[1];
    R[1] = (P[0]+P[1])*(Q[0]+Q[1])-R[0]-R[2];
    return;
  case 3:
    R[0] = P[0]*Q[0];
    R[2] = P[1]*Q[1];
    R[4] = P[2]*Q[2];
    R[1] = (P[0]+P[1])*(Q[0]+Q[1])-R[0]-R[2];
    R[3] = (P[1]+P[2])*(Q[1]+Q[2])-R[2]-R[4];
    R[2]+= (P[0]+P[2])*(Q[0]+Q[2])-R[0]-R[4];
    return;
  }
#endif

  if (n>=FFT_threshold<C>) {
    FFT_mul (R, P, Q, n);
    return;
  }

  // cout << "Full Karatsuba " << n << "\n";
  int i;
  int l= (n+1)>>1;
  int r= n-l;
  C* T = new C[l<<2](ZERO<C>);
  C* SP= T + (l<<1);
  C* SQ= SP+ l;
  for (i=0; i<r; i++) {
    SP[i]= P[i]+ P[i+l];
    SQ[i]= Q[i]+ Q[i+l];
  }
  if (i<l) {
    SP[i]= P[i];
    SQ[i]= Q[i];
  }

  mul (R, P, Q, l); R[(l<<1)-1]= 0;
  mul (R+(l<<1), P+l, Q+l, r);
  mul (T, SP, SQ, l);

  for (i=0; i<((l<<1)-1); i++) T[i]   -= R[i];
  for (i=0; i<((r<<1)-1); i++) T[i]   -= R[i+(l<<1)];
  for (i=0; i<((l<<1)-1); i++) R[i+l] += T[i];  
  delete[] T;
  // cout << "Done\n";
}

static void
mul (C* R, C* P, C* Q, int n, int m) {
  if (n==m) mul (R, P, Q, n);
  else if (m>n) mul (R, Q, P, m, n);
  else if (m==0) return;
  else {
    int i;
    mul (R, P, Q, n-m, m);
    for (i=n; i<n+m-1; i++) R[i]=0;
    C* T= new C[(m<<1)-1](ZERO<C>);
    mul (T, P+n-m, Q, m);
    for (i=0; i<((m<<1)-1); i++) R[i+n-m] = R[i+n-m]+ T[i];
    delete[] T;
  }
}

static void
trunc_mul (C* R, C* P, C* Q, int n) {
#ifdef Karatsuba_threshold<C>
  if (n<Karatsuba_threshold<C>) {
    // cout << "Truncated Naive " << n << "\n";
    int i, j;
    for (i=0; i<n; i++) R[i]= ZERO<C>;
    for (i=0; i<n; i++) {
      // cout << "  " << i << "\n";
      for (j=0; j<n-i; j++)
	R[i+j]= R[i+j]+ P[i]*Q[j];
    } 
    // cout << "Done\n";
    return;
  }
#else
  switch (n) {
  case 0:
    return;
  case 1:
    R[0]= P[0]*Q[0];
    return;
  case 2:
    R[0] = P[0]*Q[0];
    R[1] = P[0]*Q[1]+P[1]*Q[0];
    return;
  case 3:
    R[0] = P[0]*Q[0];
    R[2] = P[1]*Q[1];
    R[1] = (P[0]+P[1])*(Q[0]+Q[1])-R[0]-R[2];
    R[2]+= P[0]*Q[2]+P[2]*Q[0];
    return;
  }
#endif

  // cout << "Truncated Karatsuba " << n << "\n";
  int i;
  int l = (n+1)>>1;
  int r = n-l;
  C* T  = new C[n](ZERO<C>);
  R[n-1]= ZERO<C>;
  mul (R, P, Q, l);
  trunc_mul (T, P+l, Q, r);
  for (i=0; i<r; i++) R[i+l] += T[i];
  trunc_mul (T, P, Q+l, r);
  for (i=0; i<r; i++) R[i+l] += T[i];
  delete[] T;
  // cout << "Done\n";
}

/******************************************************************************
* FFT multiplication
******************************************************************************/

zealous<C>
mul_unit (zealous<C> f, int r) {
  int i, n=N(f); r= r- (4*n*n);
  zealous<C> g(n);
  for (i=0; i<n; i++) {
    int j= (i-r)%(2*n);
    if (j<n) g[i]= f[j];
    else g[i]= -f[j-n];
  }
  return g;
}

void
DFT (zealous<C>* a, int n, int r, int step=1) {
  int i;
  if (n>=4) {
    DFT (a, n>>1, r<<1, step<<1);
    DFT (a+step, n>>1, r<<1, step<<1);
  }
  zealous<C>* b= new zealous<C>[n];
  for (i=0; i<(n/2); i++) {
    zealous<C> x= a[2*i*step];
    zealous<C> y= ((i==0)? a[(2*i+1)*step]: mul_unit (a[(2*i+1)*step], i*r));
    b[i]= x+y;
    b[i+(n/2)]= x-y;
  }
  for (i=0; i<n; i++) a[i*step]= b[i];
  delete[] b;
}

zealous<C>
cyclo_mul (zealous<C> f, zealous<C> g) {
  int n= N(f);
  if (n<FFT_threshold<C>) {
    zealous<C> h= multiply (f, g);
    return h (0, n) - h (n, n<<1);
  }
  else {
    int z= 0;
    while ((n&1)==0) { n=n>>1; z++; }
    n= n << z;
    int m= n >> (z/2);
    int d= n/m;
    zealous<C>* a= new zealous<C>[6*d](m);
    zealous<C>* b= a+ (2*d);
    zealous<C>* c= b+ (2*d);
    int i, j;
    for (j=0; j<d; j++)
      for (i=0; i<m; i++) {
	a[j][i]= f[d*i+j];
	b[j][i]= g[d*i+j];
      }

    DFT (a, 2*d, m/d);
    DFT (b, 2*d, m/d);
    for (i=0; i<(2*d); i++)
      c[i]= cyclo_mul (a[i], b[i]);
    DFT (c, 2*d, (2*d-1)*m/d);

    zealous<C> h (n);
    for (j=0; j<d; j++)
      for (i=0; i<m; i++) {
	C x= c[j][i];
	C y= i==0? -c[j+d][m-1]: c[j+d][i-1];
	h[d*i+j]= (x+y)/C(2*d);
      }
    delete[] a;
    return h;
  }
}

void
FFT_mul (C* R, C* P, C* Q, int n) {
  int i, N=16;
  while (TRUE) {
    if (3*N >= n) { N= 6*N; break; }
    if (4*N >= n) { N= 8*N; break; }
    N = N << 1;
  }

  zealous<C> f (N), g (N);
  for (i=0; i<n; i++) {
    f[i]= P[i];
    g[i]= Q[i];
  }
  zealous<C> h= cyclo_mul (f, g);
  for (i=0; i<(2*n-1); i++) R[i]= h[i];
}

/******************************************************************************
* Basic arithmetic
******************************************************************************/

zealous<C>
operator + (zealous<C> f, zealous<C> g) {
  int i, n (N(f));
  if (N(g)!=n) fatal_error ("unequal truncation orders", "operator +");
  zealous<C> h (n);
  for (i=0; i<n; i++) h[i]= f[i]+ g[i];
  return h;
}

zealous<C>
operator - (zealous<C> f, zealous<C> g) {
  int i, n (N(f));
  if (N(g)!=n) fatal_error ("unequal truncation orders", "operator -");
  zealous<C> h (n);
  for (i=0; i<n; i++) h[i]= f[i]- g[i];
  return h;
}

zealous<C>
operator * (C c, zealous<C> f) {
  int i, n (N(f));
  zealous<C> g (n);
  for (i=0; i<n; i++) g[i]= c*f[i];
  return g;
}

zealous<C>
multiply (zealous<C> f, zealous<C> g) {
  zealous<C> h (N(f)+N(g)-1);
  mul (h->a, f->a, g->a, N(f), N(g));
  return h;
}

zealous<C>
multiply (zealous<C> f, zealous<C> g, int order) {
  // return multiply (f, g) (0, order <? (N(f)+N(g)-1));
  // cout << "  ] " << N(f) << " * " << N(g) << " : " << order << "\n";
  if (order >= (N(f)+N(g)-1)) return multiply (f, g);
  if (N(g) > N(f)) return multiply (g, f, order);
  zealous<C> h (order);

  if (order <= N(g)) {
    trunc_mul (h->a, f->a, g->a, order);
    return h;
  }

  if (order <= N(f)) {
    mul (h->a, f->a, g->a, order-N(g), N(g));
    h->accumulate (f(order-N(g), order) * g, order-N(g));
    return h;
  }

  int l= (order+1)>>1;
  if (l < N(g)) {
    mul (h->a, f->a, g->a, l, l);
    h->accumulate (multiply (f (l, N(f)), g (0, l), order-l), l);
    h->accumulate (multiply (f (0, l), g (l, N(g)), order-l), l);
    return h;    
  }
  else {
    mul (h->a, f->a, g->a, order-N(g), N(g));
    h->accumulate (multiply (f (order-N(g), N(f)), g, N(g)), order-N(g));
    return h;
  }
}

zealous<C>
operator * (zealous<C> f, zealous<C> g) {
  if (N(g)!=N(f)) fatal_error ("unequal truncation orders", "operator *");
  zealous<C> h (N(f));
  trunc_mul (h->a, f->a, g->a, N(f));
  return h;
}

zealous<C>
inv (zealous<C> f) {
  int n= N(f);
  if (n==1) return constant (C(1)/f[0], n);
  int m= (n+1) >> 1;
  zealous<C> g= inv (f (0, m)) (0, n);
  return g- ((((f*g)>>m)*g(0,n-m))<<m);
}

zealous<C>
operator / (zealous<C> f, zealous<C> g) {
  if (N(g)!=N(f)) fatal_error ("unequal truncation orders", "operator /");
  return f*inv(g);
}

/******************************************************************************
* Special functions
******************************************************************************/

#ifdef has_divisible<C>
zealous<C>
operator / (zealous<C> f, C c) {
  int i, n (N(f));
  zealous<C> g (n);
  for (i=0; i<n; i++) g[i]= f[i]/c;
  return g;
}

zealous<C>
d (zealous<C> f) {
  int i, n (N(f));
  zealous<C> g (n-1);
  for (i=0; i<n-1; i++) g[i]= C(i+1)* f[i+1];
  return g;
}

zealous<C>
I (zealous<C> f) {
  int i, n (N(f));
  zealous<C> g (n+1);
  for (i=0; i<n; i++) g[i+1]= f[i]/C(i+1);
  return g;
}

zealous<C>
log (zealous<C> f) {
#ifdef has_special_functions<C>
  zealous<C> g= I(d(f)/f(0,N(f)-1));
  g[0]= log (f[0]);
  return g;
#else
  return I(d(f)/f(0,N(f)-1));
#endif
}

zealous<C>
exp (zealous<C> f) {
  int n= N(f);
  if (n==1) {
#ifdef has_special_functions<C>
    return constant (exp(f[0]), 1);
#else
    return constant (C(1), 1);
#endif
  }
  int m= (n+1)>>1;
  zealous<C> g= exp (f(0,m)) (0, n);
  return g- (log (g)- f)*g;
}
#endif

/******************************************************************************
* Composition
******************************************************************************/

static zealous<C>
compose (zealous<C> f, zealous<C>* g, int n) {
  int p= N(f), q= N(g[1]);
  if (p==1) return constant (f[0], 1);
  int l= ((p-1)*(q-1)+1) <? n;
  zealous<C> h1= compose (f(0, p>>1), g, n);
  zealous<C> h2= compose (f(p>>1, p), g, n);
  // cout << "    Table? " << (p>>1) << "\n";
  return h1 (0, l)+ multiply (h2, g[p>>1]) (0, l);
}

static void
setup_table (zealous<C>* g, int p, int n) {
  if ((p<=1) || (N(g[p])!=0)) return;
  int q= N(g[1]);
  setup_table (g, p>>1, n);
  setup_table (g, (p+1)>>1, n);
  // cout << "    Table: " << p << "\n";
  g[p]= multiply (g[p>>1], g[(p+1)>>1]) (0, (p*(q-1)+1) <? n);
}

zealous<C>
compose (zealous<C> f, zealous<C> g, int n) {
  zealous<C> G[n+1];
  if (n>1) G[1]= g;
  setup_table (G, n>>1, n);
  setup_table (G, (n+1)>>1, n);
  return compose (f, G, n);
}

#ifdef has_divisible<C>
static C
factorial<C> (int n) {
  int i;
  C f (1);
  for (i=1; i<=n; i++) f= f*C(i);
  return f;
}

zealous<C>
compose (zealous<C> f, zealous<C> g) {
  int n= N(f);
  if (N(g)!=n) {
    if (N(g)*N(g) < n) return compose (f, g, n);
    g= g (0, n);
  }

  // cout << "  Composing " << n << "\n";
  int p=0;
  while (((p+1)*(1<<(2*p+2))) <= n) p++;
  int q= 1<<p, r= (n+q-1)/q;
  zealous<C> g1 = g(0,q)(0,n);
  zealous<C> g2 = g-g1;
  zealous<C> dg1= d(g1);

  int i;
  zealous<C> D = f;
  for (i=1; i<r; i++) D= d(D);
  D= compose (D, g1, n+1-r) / factorial<C>(r-1);

  zealous<C> S = D (0, 0 >? (n+q-r*q)), T;
  for (i=r-1; i>0; i--) {
    zealous<C> c (n+1-i); c[0]= f[i-1];
    D= c+ I (dg1 (0, n-i)* (C(i) * D));
    T= (S * (g2>>q) (0, 0 >? (n-i*q))) << q;
    S= D (0, 0 >? (n+q-i*q))+ T (0, 0 >? (n+q-i*q));
  }
  // cout << "  Returning " << n << "\n";
  return S;
}

zealous<C>
revert (zealous<C> f) {
  int n= N(f);
  // cout << "Reverting " << n << "\n";
  if (n==1) return constant (0, 1);
  if (n==2) return constant (C(1)/f[1], 1) << 1;
  int m= (n+1) >> 1;
  zealous<C> g= revert (f(0,m)) (0,n);
  zealous<C> N= compose (f, g)- id (n);
  zealous<C> D= compose (d(f), g(0,n-1));
  // cout << "Returning " << n << "\n";
  return g- (((N>>1)/D)<<1);
}
#endif

/******************************************************************************
* Output
******************************************************************************/

ostream&
operator << (ostream& out, zealous<C> f) {
  int i;
  for (i=0; i<N(f); i++) {
    out << f[i];
    if (i==0);
    else if (i==1) out << " z";
    else out << " z^" << i;
    if (i<N(f)-1) out << " + ";
  }
  return out;
}

#endmodule // zealous (C)
