
/******************************************************************************
* MODULE     : lexer.gen.cc
* DESCRIPTION: dynamic lexer
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <lexer.gen.h>
#include <file.gen.h>

#include <array.gen.cc>
#include <list.gen.cc>
#include <hashset.gen.cc>
#include <hashmap.gen.cc>
#include <iterator.gen.cc>

#module code_lexer
#import lexer
#import file
#import code_resource (lexer_rep, lexer)

/******************************************************************************
* Lexical states
******************************************************************************/

#import code_array (lex_state)
#import code_list (lex_state)
#import code_hashset (lex_state)
#import code_hashmap (string, lex_state)
#import code_hashmap (string, array<lex_state>)
#import code_hashmap (lex_state, path)
#import code_iterator (lex_state)
#import code_hashset_iterator (lex_state)
#import code_hashmap_iterator (string, array<lex_state>)
#import code_hashmap_iterator (lex_state, path)

lex_state_rep::lex_state_rep (program name2, int p2, string jump2):
  next (array<lex_state> ()), name (name2), priority (p2), jump (jump2) {}

lex_state::lex_state (program name, int p, string jump):
  rep (new lex_state_rep (name, p, jump)) {}
int hash (lex_state st) { return ((int) st.rep); }
bool operator == (lex_state st1, lex_state st2) { return st1.rep == st2.rep; }
bool operator != (lex_state st1, lex_state st2) { return st1.rep != st2.rep; }
ostream& operator << (ostream& out, lex_state st) { return out << st->name; }

/******************************************************************************
* Subroutines for constructing programs
******************************************************************************/

static program
def_language (program p) {
  int i;
  array<program> a (N(p)-1);
  a[0]= program ("|", p[0]->start, p[0]->end);
  for (i=2; i<N(p); i++) a[i-1]=p[i];
  return program (a, p->start, p->end);
}

static program
atom (string s) {
  array<program> a (1);
  a[0]= s;
  return program (a);
}

static program
operator * (program p, program q) {
  int i;
  array<program> a (N(p)+1);
  for (i=0; i<N(p); i++) a[i]=p[i];
  a[i]= q;
  return program (a, p->start, q->end);
}

/******************************************************************************
* Building the lexer
******************************************************************************/

void
lexer_rep::insert_definition (string name, program p) {
  lex_state st1= insert_state (atom (name));
  lex_state st2= insert_state (atom (name) * p);
  p= def_language (p);
  convert (p, st1, st2);
  init (name) = st1;
  term (name) = st2;
  def  (name) = p;
}

lex_state
lexer_rep::insert_state (program name, int priority, string jump) {
  lex_state st (name, priority, jump);
  table->insert (st);
  return st;
}

void
lexer_rep::insert_transition (lex_state st1, lex_state st2, string s) {
  if (N(st1->next[s])==0) st1->next (s)= array<lex_state> ();
  st1->next (s) << (st2);
}

void
lexer_rep::convert (program p, lex_state st1, lex_state st2) {
  // cout << "Converting " << p << "\n";
  if (is_string (p)) {
    string& s (p->s);
    if (s=="<eof>") {
      insert_transition (st1, st2, "<eof>");
      return;
    }
    if (!def->contains (s)) {
      cerr << "Lexical error: unknown symbol '" << s << "' in grammar\n";
      return;
    }
    convert (def[s], st1, st2);
  }
  else if (is_func (p, "'", 1) && is_string (p[1])) {
    int i;
    string s (p[1]->s);
    program base_name= st1->name;
    if (N(s)==0) { insert_transition (st1, st2, ""); return; }
    for (i=0; i<N(s)-1; i++) {
      lex_state stm= insert_state (base_name * program (s (0, i+1)));
      insert_transition (st1, stm, s (i, i+1));
      st1=stm;
    }
    insert_transition (st1, st2, s (i, i+1));
  }
  else if (is_func (p, "-", 2)) {
    if ((p[1]->status!=NONCOMPILED) || (p[2]->status!=NONCOMPILED)) return;
    char c;
    char c1=p[1]->s[0];
    char c2=p[2]->s[0];
    for (c=c1; c<=c2; c++)
      insert_transition (st1, st2, string (c));
  }
  else if (is_func (p, "*", 1)) {
    lex_state stm= insert_state (st1->name * p);
    convert (p[1], stm, stm);
    insert_transition (st1, stm, "");
    insert_transition (stm, st2, "");
  }
  else if (is_func (p, "+", 1)) {
    lex_state stm= insert_state (st1->name * p);
    convert (p[1], st1, stm);
    convert (p[1], stm, stm);
    insert_transition (stm, st2, "");
  }
  else if (is_func (p, "?", 1)) {
    convert (p[1], st1, st2);
    if (st1!=st2) insert_transition (st1, st2, "");
  }
  else if (is_func (p, ";")) {
    if (N(p)==1) return;
    int i;
    for (i=1; i<N(p)-1; i++) {
      lex_state stm= insert_state (st1->name * p[i]);
      convert (p[i], st1, stm);
      st1=stm;
    }
    convert (p[i], st1, st2);
  }
  else if (is_func (p, "|")) {
    int i;
    for (i=1; i<N(p); i++)
      convert (p[i], st1, st2);
  }
  else if ((N(p)==0) || (p[0]->status != NONCOMPILED));
}

void
lexer_rep::insert_rule (string name, program p, int pri) {
  if ((N(p)!=2) && ((N(p)!=3) || (!is_string (p[2])))) return;
  if (!is_composite (p[1])) return;
  string jump= name;
  if (N(p)==3) jump= p[2]->s;
  lex_state st1= init[name];
  lex_state st2= insert_state (atom (name) * p[0], pri, jump);
  convert (p[0], st1, st2);
}

void
lexer_rep::insert_symbol (string name, program p) {
  lex_state st1= insert_state (atom (name));
  // lex_state st2= insert_state (atom (name) * p);
  init (name) = st1;
  // term (name) = st2;
  def  (name) = p;

  int i;
  for (i=1; i<N(p); i++)
    insert_rule (name, p[i], i);
}

lexer_rep::lexer_rep (string name, program p):
  rep<lexer> (name), init (lex_state ()),
  term (lex_state ()), def (program ()),
  lex_H (path ())
{
  int i;
  if (p->status!=COMPOSITE) {
    cerr << "Lexical error: syntax error in grammar\n"; return; }
  for (i=0; i<N(p); i++) {
    if ((p[i]->status!=COMPOSITE) ||
	(N(p[i])<2) ||
	(!is_string (p[i][0])) ||
	((is_string (p[i][0], "def")) && (!is_string(p[i][1])))) {
      cerr << "Lexical error: syntax error in grammar\n";
      return;
    }
    if (is_string (p[i][0], "def")) insert_definition (p[i][1]->s, p[i]);
    else insert_symbol (p[i][0]->s, p[i]);
  }
}

lexer_rep::~lexer_rep () {
  iterator<lex_state> it= iterate (table);
  while (it->busy ()) {
    lex_state st= it->next();
    iterator<string> inner= iterate (st->next);
    while (inner->busy ()) {
      string s= inner->next ();
      st->next (s)= array<lex_state> (0);
    }
  }
}

lexer
load_lexer (string name) {
  if (resource<lexer> -> contains (name)) return lexer (name);
  string s, fname= name * ".lex";
  if (debug (0)) cout << "TeXmacs] Loading " << fname << "\n";
  load_string ("$TEXMACS_GRAMMAR_PATH", fname, s, TRUE);
  program p (lisp_prg_to_program (s, 0));
  return new lexer_rep (name, p);
}

/******************************************************************************
* Lexing a string
******************************************************************************/

void
lexer_rep::set_input (string s, int offset, string lex_symb2, bool flag) {
  lex_s   = s;
  lex_symb= lex_symb2;
  lex_last= 0;
  lex_pos = 0;
  lex_off = offset;
  lex_eof = FALSE;
  lex_H   = hashmap<lex_state,path> (path());
  lex_r   = "";
  lex_t   = "";
  lex_flag= flag;

  if (!init->contains (lex_symb)) {
    cerr << "Error: no symbol '" << lex_symb << "' in lexical grammar\n";
    return;
  }

  lex_H (init[lex_symb])= path ();
  lex ("");
}

static int
compare_priority (path p1, path p2) {
  if (nil (p1) && nil (p2)) return 0;
  // cout << "      Compare " << p1 << " with " << p2 << "\n";
  if (nil (p1)) return 1;
  if (nil (p2)) return -1;
  if (p1->item < p2->item) return 1;
  if (p1->item > p2->item) return -1;
  if (p1->next->item < p2->next->item) return -1;
  if (p1->next->item > p2->next->item) return 1;
  return compare_priority (p1->next->next, p2->next->next);
}

void
lexer_rep::lex (string c) {
  // cout << lex_s (0,lex_pos) << "[" << c << "]:  " << lex_H << "\n";
  if (c!="<eof>") lex_pos += N(c);

  // advance by c
  int i;
  hashmap<lex_state,path> H= path ();
  iterator<lex_state> it= iterate (lex_H);
  while (it->busy()) {
    lex_state st= it->next();
    // cout << "  " << st << " ->\n";
    path p= lex_H[st];
    if (N(c)==0) {
      if ((!H->contains (st)) ||
	  (compare_priority (revert (H[st]), revert (lex_H[st])) < 0))
	H (st)= lex_H [st];
      if ((N(st->jump) != 0) && init->contains (st->jump)) {
	lex_state next_st= init[st->jump];
	path q= path (lex_pos, path (st->priority, p));
	if ((!H->contains (next_st)) ||
	    (compare_priority (revert (H[next_st]), revert (q)) < 0))
	  H (next_st)= q;
      }
    }
    array<lex_state> next (st->next[c]);
    for (i=0; i<N(next); i++) {
      // cout << "    " << next[i] << "\n";
      if (H->contains (next[i]))
	if (compare_priority (revert (H[next[i]]), revert (p)) >= 0)
	  continue;
      H(next[i])= p;
    }
  }

  // extra epsilon move if H!=lex_H
  bool flag= (N(c)>0) || (N(H)!=N(lex_H));
  if (!flag) {
    iterator<lex_state> it= iterate (lex_H);
    while (it->busy()) {
      lex_state st= it->next();
      flag= flag || (!H->contains (st)) || (H[st]!=lex_H[st]);
    }
  }
  if (flag) {
    if (lex_flag && (N(H)==0))
      cerr << "Lexical error: skipping '" << c << "'\n";
    else { lex_H= H; lex (""); }
  }
}

bool
lexer_rep::produce () {
  // Test whether there should be a production

  if (N(lex_H)==0) return FALSE;

  int rule=-1, next=-1;
  iterator<lex_state> it= iterate (lex_H);
  while (it->busy()) {
    lex_state st= it->next();
    path p= revert (lex_H[st]);
    if (nil(p)) return FALSE;
    if (rule==-1) {
      rule= p->item;
      next= p->next->item;
    }
    else {
      if (rule!=p->item) return FALSE;
      if (next!=p->next->item) return FALSE;
    }
  }

  hashmap<lex_state,path> H= path();
  it= iterate (lex_H);
  while (it->busy()) {
    lex_state st= it->next();
    path p= revert (lex_H[st]);
    H(st)= revert (p->next->next);
  }
  lex_H= H;

  // Execute the production

  int i;
  program P= def[lex_symb][rule];
  program p= P[1];
  // cout << "Produce " << p << ", " << next << "\n";
  for (i=0; i<N(p); i++) {
    if (is_string (p[i])) {
      if (N(lex_r) == 0) lex_rst= lex_last;
      if (is_string (p[i], "$")) lex_r << lex_s (lex_last, next);
      else lex_r << copy (p[i]->s);
    }
    else if (is_func (p[i], "'", 1)) {
      if (N(lex_r) == 0) lex_rst= lex_last;
      else lex_r << copy (p[i][1]->s);
    }
    else if (is_composite (p[i])) {
      int ii;
      program pp= p[i];
      for (ii=0; ii<N(pp); ii++)
	if (is_string (pp[ii])) {
	  if (is_string (pp[ii], "$")) lex_t << lex_s (lex_last, next);
	  else lex_t << copy (pp[ii]->s);
	}
	else if (is_func (pp[ii], "'", 1))
	  lex_t << copy (pp[ii][1]->s);
    }
  }
  lex_last= next;
  if (N(P)==3) lex_symb= P[2]->s;

  return TRUE;
}

void
lexer_rep::get (string& s, string& tp, location& start, location& end) {
  if (!init->contains (lex_symb)) {
    s    = "";
    tp   = end;
    start= lex_off;
    end  = lex_off+ N(lex_s);
    return;
  }

  while (TRUE) {
    if (produce () && (N(lex_t) != 0)) {
      s    = lex_r;
      tp   = lex_t;
      start= lex_off+ lex_rst;
      end  = lex_off+ lex_last;
      lex_r= "";
      lex_t= "";
      return;
    }

    if (lex_eof) {
      s    = "";
      tp   = "end";
      start= lex_off+ N(lex_s);
      end  = lex_off+ N(lex_s);
      return;
    }

    if (lex_pos<N(lex_s)) lex (lex_s (lex_pos, lex_pos+1));
    else { lex_eof= TRUE; lex ("<eof>"); }
  }
}

/******************************************************************************
* Recognize a string using the lexer
******************************************************************************/

void
lexer_rep::saturate (hashset<lex_state>& S1, string c) {
  hashset<lex_state> S2;
  iterator<lex_state> it= iterate (S1);
  while (it->busy()) {
    int i;
    lex_state st= it->next();
    if (N(c)==0) S2->insert (st);
    array<lex_state> next (st->next[c]);
    for (i=0; i<N(next); i++) S2->insert (next[i]);
  }
  if (!(S1==S2)) {
    S1=S2;
    saturate (S1, "");
  }
}

bool
lexer_rep::recognizes (string s, string tp) {
  if ((!init->contains (tp)) || (!term->contains (tp))) {
    cerr << "Lexical error: '" << tp << "' undefined in grammar\n";
    return FALSE;
  }

  int i;
  hashset<lex_state> S;
  S->insert (init[tp]);
  saturate (S, "");
  for (i=0; i<N(s); i++)
    saturate (S, s (i, i+1));
  return S->contains (term[tp]);  
}

bool
lexer_rep::recognizes (string s, string tp, string& result) {
  if (!init->contains (tp)) {
    cerr << "Lexical error: no symbol '" << tp << "' in grammar\n";
    return FALSE;
  }

  if (term->contains (tp)) {
    bool flag= recognizes (s, tp);
    result= copy (s);
    return flag;
  }
  else {
    set_input (s, 0, tp, FALSE);
    while (TRUE) {
      if (produce () && (N(lex_t) != 0)) {
	result= lex_r;
	return TRUE;
      }
      if (lex_eof) return FALSE;
      if (lex_pos<N(lex_s)) lex (lex_s (lex_pos, lex_pos+1));
      else { lex_eof= TRUE; lex ("<eof>"); }
      if (N(lex_H)==0) return FALSE;
    }
  }
}

#endmodule // code_lexer
