
/******************************************************************************
* MODULE     : parser.gen.cc
* DESCRIPTION: Dynamic parser
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <parser.gen.h>
#include <iterator.gen.h>
#include <file.gen.h>

#module code_parser
#import parser
#import hashset_iterator (string)
#import file
#import code_resource (parser_rep, parser)

/******************************************************************************
* states and symbols
******************************************************************************/

#import code_array (symbol)
#import code_array (state)
#import code_hashmap (string, symbol)
#import code_hashmap (string, state)
#import code_hashmap_iterator (string, symbol)
#import code_hashmap_iterator (string, state)

symbol_rep::symbol_rep (string s):
  name (s), init (state (TRUE)), term (state (TRUE)) {}

symbol::symbol (string name): rep (new symbol_rep (name)) {}
bool operator == (symbol sy1, symbol sy2) { return sy1.rep == sy2.rep; }
bool operator != (symbol sy1, symbol sy2) { return sy1.rep != sy2.rep; }
ostream& operator << (ostream& out, symbol sy) {
  (void) sy; return out << "(symbol)"; }

state_rep::state_rep ():
  production (program ()), epsilon (state ()),
  token_table (state ()), symbol_table (state ()) {}

state::state (bool dummy): rep (new state_rep ()) { (void) dummy; }
bool operator == (state st1, state st2) { return st1.rep == st2.rep; }
bool operator != (state st1, state st2) { return st1.rep != st2.rep; }
ostream& operator << (ostream& out, state st) {
  (void) st; return out << "(state)"; }

/******************************************************************************
* construction of the parser
******************************************************************************/

void
parser_rep::insert_symbol (string name) {
  symbol sy (name);
  symbol_table (name) = sy;
  state_table (name * " | .") = sy->init;
  state_table (name * " | " * name * " .") = sy->term;
}

void
parser_rep::insert_method (string name, program p1, program p2) {
  // cout << "  " << name  << " : " << p1 << " -> " << p2 << "\n";
  string cumul= name * " |";
  symbol sy= symbol_table [name];
  if (nil (sy)) { grammar_error ("unknown symbol " * name); return; }
  int i;
  state st=sy->init;
  if (p1->status!=COMPOSITE)
    { grammar_error ("syntax error in rule"); return; }
  for (i=0; i<N(p1); i++) {
    if (p1[i]->status!=NONCOMPILED)
      { grammar_error ("syntax error in rule"); return; }
    string s= p1[i]->s;
    cumul=cumul * " " * s;
    if (i==0) {
      if (symbol_table->contains (s))
	sy->symbol_table->insert (s);
      else sy->token_table->insert (s);
    }

    if ((i==0) && (s==name)) st= sy->term;
    else if (st->token_table->contains (s)) st= st->token_table [s];
    else if (st->symbol_table->contains (s)) st= st->symbol_table [s];
    else {
      state next (TRUE);
      state_table (cumul * " .")= next;
      if (symbol_table->contains (s)) st->symbol_table (s)= next;
      else st->token_table (s)= next;
      st= next;
    }
  }
  if (is_string (st->production, "")) {
    st->production= p2;
    st->epsilon   = sy->term;
  }
  else grammar_error ("state " * cumul * " has already a production");
}

void
parser_rep::complete_automaton () {
  bool flag;
  do {
    flag=FALSE;
    iterator<string> it= iterate (symbol_table);
    while (it->busy ()) {
      symbol sy= symbol_table [it->next()];
      iterator<string> it2= iterate (sy->symbol_table);
      while (it2->busy ()) {
	symbol sy2= symbol_table [it2->next()];
	iterator<string> insert= iterate (sy2->token_table);
	while (insert->busy ()) {
	  string token= insert->next ();
	  if (!sy->token_table->contains (token)) {
	    sy->token_table->insert (token);
	    flag= TRUE;
	  }
	}
      }
    }
  } while (flag);
}

void
parser_rep::grammar_error (string s) {
  cerr << "Error in grammar: " << s << "\n";
}

/******************************************************************************
* Constructors
******************************************************************************/

parser_rep::parser_rep (string name, program p):
  rep<parser> (name), state_table (state ()), symbol_table (symbol ())
{
  int i;
  if (p->status!=COMPOSITE) { grammar_error ("syntax error"); return; }
  for (i=0; i<N(p); i++) {
    if ((p[i]->status!=COMPOSITE) ||
	((N(p[i])&1)==0) ||
	(!is_string (p[i][0]))) {
      grammar_error ("syntax error");
      return;
    }
    insert_symbol (p[i][0]->s);
  }
  for (i=0; i<N(p); i++) {
    // cout << "\nHandling " << p[i] << "\n\n";
    int j;
    for (j=1; j<N(p[i]); j+=2)
      insert_method (p[i][0]->s, p[i][j], p[i][j+1]);
  }
  complete_automaton();

  /*
  cout << "------------------- symbols -------------------\n";
  iterator<string> syt= iterate (symbol_table);
  while (syt->busy ()) {
    string s= syt->next ();
    cout << s << " ->\n";
    symbol sy= symbol_table [s];
    cout << "  " << sy->token_table << "\n";
    cout << "  " << sy->symbol_table << "\n";
  }
  cout << "------------------- states --------------------\n";
  iterator<string> stt= iterate (state_table);
  while (stt->busy ()) {
    string s= stt->next ();
    cout << s;
    state st= state_table [s];
    if (st->production == "") cout << "\n";
    else cout << " -> " << st->production << "\n";
  }
  */
}

parser_rep::~parser_rep () {
  iterator<string> it1= iterate (symbol_table);
  while (it1->busy ()) {
    symbol sy= symbol_table[it1->next()];
    sy->init= state ();
    sy->term= state ();
  }
  iterator<string> it2= iterate (state_table);
  while (it2->busy ()) {
    state st= state_table[it2->next()];
    st->epsilon= state ();
    st->token_table = hashmap<string,state> (state ());
    st->symbol_table= hashmap<string,state> (state ());
  }
}

parser
load_parser (string name) {
  if (resource<parser> -> contains (name)) return parser (name);
  string s, fname= name * ".grm";
  if (debug (0)) cout << "TeXmacs] Loading " << fname << "\n";
  load_string ("$TEXMACS_GRAMMAR_PATH", fname, s, TRUE);
  program p (lisp_prg_to_program (s, 0));
  return new parser_rep (name, p);
}

/******************************************************************************
* using the parser
******************************************************************************/

void
parser_rep::set_input (lexer lx) {
  parse_lx= lx;
  read ();
}

void
parser_rep::read () {
  parse_lx->get (parse_s, parse_type, parse_start, parse_end);
  // cout << "Got " << parse_s << " : " << parse_type << "\n";
  if (parse_type == "error") { error ("lexical error"); read(); return; }
  if (parse_type == "empty") read();
}

program
parser_rep::produce (array<program> a, program p) {
  int i;

  if (is_func (p, "<*", 2)) {
    program p1= produce (a, p[1]);
    program p2= produce (a, p[2]);
    array<program> b (N(p2)+1);
    b[0]= p1;
    for (i=0; i<N(p2); i++) b[i+1]= p2[i];
    return program (b);
  }
  if (is_func (p, "*<", 2)) {
    program p1= produce (a, p[1]);
    program p2= produce (a, p[2]);
    array<program> b (N(p1)+1);
    for (i=0; i<N(p1); i++) b[i]= p1[i];
    b[i]= p2;
    return program (b);
  }

  if (p->status==COMPOSITE) {
    array<program> b (N(p));
    for (i=0; i<N(p); i++)
      b[i]= produce (a, p[i]);
    return b;
  }
  else {
    if (is_int (p->s)) {
      i= as_int (p->s);
      return a[i-1];
    }
    else return program (p->s);
  }
}

program
parser_rep::parse (string the_type) {
  // cout << "Parsing " << the_type << "\n";
  if (!symbol_table->contains (the_type)) {
    error ("no symbol '" * the_type * "' in grammar of parser");
    return program ("error", parse_start, parse_start);
  }

  symbol sy= symbol_table [the_type];
  state  st= sy->init;
  
  array<program> a;
  while (TRUE) {
    state next= st->token_table (parse_type);
    if (!nil (next)) {
      a << program (parse_s, parse_start, parse_end);
      read ();
    }
    else {
      iterator<string> it= iterate (st->symbol_table);
      while (it->busy()) {
	string sy_name= it->next ();
	symbol sy= symbol_table [sy_name];
	if (sy->token_table->contains (parse_type)) {
	  a << parse (sy_name);
	  next= st->symbol_table [sy_name];
	  break;
	}
      }
    }

    if (!nil (next)) st= next;
    else {
      if (st==sy->term) return a[0];
      if (!is_string (st->production, "")) {
	program result= produce (a, st->production);
	if (N(a)==0) {
	  result->start = parse_start;
	  result->end   = parse_start;
	}
	else {
	  result->start = a[0]->start;
	  result->end   = a[N(a)-1]->end;
	}
	a   = array<program> (1);
	a[0]= result;
	st  = st->epsilon;
      }
      else {
	if (parse_type == "end") {
	  error ("premature end");
	  return program ("error", parse_start, parse_start);
	}
	else error ("syntax error");
	read ();
      }
    }
  }
}

program
parser_rep::parse () {
  program p= parse ("main");
  if (parse_type != "end") error ("superfluous tokens at end of file");
  return p;
}

void
parser_rep::error (string s) {
  cerr << "Parse error: " << s << " in '" << parse_s << "'\n";
}

#endmodule // code_parser
