
/******************************************************************************
* MODULE     : tiger.gen.cc
* DESCRIPTION: Dynamic tiger
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <analyze.gen.h>
#include <tiger.gen.h>
#include <file.gen.h>

#module code_tiger
#import analyze
#import tiger
#import file
#import code_resource (tiger_rep, tiger)

/******************************************************************************
* construction of the tiger
******************************************************************************/

void
tiger_rep::insert_symbol (string name) {
  array<program> a;
  rules_table (name)= a;
}

void
tiger_rep::insert_method (string name, program p1, program p2) {
  // cout << "  " << name  << " : " << p1 << " -> " << p2 << "\n";

  program& rules (rules_table(name));
  int i, n=N(rules);
  array<program> new_rules (n+1);
  for (i=0; i<n; i++) new_rules[i]= rules[i];
  array<program> rule(2); rule[0]=p1; rule[1]=p2;
  new_rules[i]= program (rule);
  rules_table (name)= program (new_rules);

  // for (i=0; i<n+1; i++)
  //   cout << "    " << new_rules[i] << "\n";
}

void
tiger_rep::grammar_error (string s) {
  cerr << "Error in grammar: " << s << "\n";
}

/******************************************************************************
* Constructors
******************************************************************************/

tiger_rep::tiger_rep (string name, program p):
  rep<tiger> (name), rules_table (program ())
{
  int i;
  if (p->status!=COMPOSITE) { grammar_error ("syntax error"); return; }
  for (i=0; i<N(p); i++) {
    if ((p[i]->status!=COMPOSITE) ||
	((N(p[i])&1)==0) ||
	(!is_string (p[i][0]))) {
      grammar_error ("syntax error");
      return;
    }
    insert_symbol (p[i][0]->s);
  }
  for (i=0; i<N(p); i++) {
    // cout << "\nHandling " << p[i] << "\n\n";
    int j;
    for (j=1; j<N(p[i]); j+=2)
      insert_method (p[i][0]->s, p[i][j], p[i][j+1]);
  }
}

tiger
load_tiger (string name) {
  if (resource<tiger> -> contains (name)) return tiger (name);
  string s, fname= name * ".tgr";
  if (debug (0)) cout << "TeXmacs] Loading " << fname << "\n";
  load_string ("$TEXMACS_GRAMMAR_PATH", fname, s, TRUE);
  program p (lisp_prg_to_program (s, 0));
  return new tiger_rep (name, p);
}

/******************************************************************************
* using the tiger
******************************************************************************/

void
tiger_rep::set_lexer (lexer lx) {
  tiger_lx= lx;
}

bool
tiger_rep::match (program p, program l) {
  // cout << "Match " << p << " ? " << l << "\n";

  if (is_string (l, ".")) return TRUE;
  if (is_func (l, "\'", 1)) return p==l[1];
  if (is_string (l)) {
    string dummy;
    if (!is_string (p)) return FALSE;
    return tiger_lx->recognizes (p->s, l->s, dummy);
  }
  if (!is_composite (l)) return FALSE;
  if (!is_composite (p)) return FALSE;

  int i, nl= N(l), np= N(p), wild_card=-1;
  for (i=0; i<nl; i++)
    if (is_string (l[i], "*")) {
      if (wild_card!=-1) {
	error ("multiple wild card (incorrect grammar)", p);
	return FALSE;
      }
      wild_card= i;
    }

  if (wild_card==-1) {
    if (np!=nl) return FALSE;
    for (i=0; i<nl; i++)
      if (!match (p[i], l[i])) return FALSE;
    return TRUE;
  }
  else {
    if (np<nl) return FALSE;
    for (i=0; i<wild_card; i++)
      if (!match (p[i], l[i])) return FALSE;
    for (i++; i<nl; i++)
      if (!match (p[i+np-nl], l[i])) return FALSE;
    return TRUE;
  }
}

program
tiger_rep::get_sub_program (program p, program l, program where, int offset) {
  if (offset==N(where)) {
    // cout << "Done " << p << ", " << l << ", " << where << "\n";
    if ((N(p)==0) && (N(l)==0) && (!is_string (l, "."))) {
      string what;
      tiger_lx->recognizes (p->s, l->s, what);
      return program (what);
    }
    return p;
  }
  if (N(p)==0) {
    error ("can not descend (incorrect grammar)", p);
    return p;
  }
  if (!is_string (where[offset])) {
    error ("bad path indication (incorrect grammar)", p);
    return p;
  }
  int child= as_int (where[offset]->s)-1;
  if (child>=N(l)) {
    error ("invalid child (incorrect grammar)", p);
    return p;
  }
  
  int i, nl= N(l), np= N(p), wild_card=nl;
  for (i=0; i<nl; i++)
    if (is_string (l[i], "*"))
      wild_card= i;
    
  if (child<wild_card)
    return get_sub_program (p[child], l[child], where, offset+1);
  if (child>wild_card)
    return get_sub_program (p[child+np-nl], l[child], where, offset+1);
  if ((offset+1)!=N(where))
    error ("can not descend after wild card (incorrect grammar)", p);
  array<program> a (np+1-nl);
  for (i=0; i<np+1-nl; i++)
    a[i]= p[i+wild_card];
  return program (a);
}

void
tiger_rep::produce (program p, program l, program r, hashset<string>& X) {
  // cout << "Produce " << p << " via " << r << "\n";

  int i;
  for (i=0; i<N(r); i++) {
    if (is_string (r[i])) tiger_out << r[i]->s;
    if (is_composite (r[i]) && (N(r[i])>0)) {
      program subp= get_sub_program (p, l, r[i], 1);
      if (is_string (r[i][0], "$")) {
	if (!is_string (subp))
	  error ("string expected (incorrect grammar)", subp);
	else tiger_out << subp->s;
      }
      else if (N(r[i])>1) {
	hashset<string> Y;
	convert (subp, r[i][0]->s, Y);
      }
      else convert (subp, r[i][0]->s, X);
    }
  }
}

void
tiger_rep::convert (program p, string type, hashset<string>& X) {
  // cout << "Converting " << p << " : " << type << " " << X << "\n";

  if (X->contains (type)) {
    tiger_out << "?";
    error ("cyclic match of type " * type, p);
    return;
  }
  if (!rules_table->contains (type)) {
    tiger_out << "?";
    error ("unknown symbol " * type, p);
    return;
  }

  program rules (rules_table (type));
  int i, n= N(rules);
  for (i=0; i<n; i++)
    if (match (p, rules[i][0])) {
      X->insert (type);
      produce (p, rules[i][0], rules[i][1], X);
      return;
    }

  tiger_out << "?";
  error ("no match of type " * type, p);
}

string
tiger_rep::convert (program p, string type) {
  hashset<string> X;
  tiger_out= "";
  convert (p, type, X);
  return tiger_out;
}

void
tiger_rep::error (string s, program p) {
  cerr << "Tiger error: " << s << " in " << p << "\n";
}

#endmodule // code_tiger
