
/******************************************************************************
* MODULE     : totex.gen.cc
* DESCRIPTION: conversion of edit trees into TeX strings
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#module code_totex
#import hashmap (string, tree)

/******************************************************************************
* The totexer structure
******************************************************************************/

struct to_tex {
  string  buffer;
  int     last_space, line_count;
  bool    lf, no_lf;
  bool    latex_flag;
  bool    disable_output;
  hashmap<string,tree> env;

public:
  to_tex (bool latex_flag);

  void   print (string s);
  void   remove (int n=1);
  void   sep (bool text_also= FALSE, bool force= FALSE);
  void   ret ();
  string last (int n=1);

  void   print_cmd (string cmd);
  void   tex_arg (tree t, bool flag=FALSE);
  void   tex_string (string s);
  void   set (string s, tree t);
  void   assign (string s, tree t);
  void   reset (string s);
  bool   in_mode (string which);
  string get_decl (string var, string val);
  string get_cmd (string var, string val);
  void   tex_set_var (string var, string val);
  void   tex_assign (string var, tree val);
  void   tex_reset_var (string var);

  void   tex_apply (tree t);
  int    tex_special_env (string& s);
  void   tex_begin (tree t);
  void   tex_end (tree t);

  void   tex_script (string s, tree t);
  void   tex_matrix_table (tree t, bool flag);
  void   tex (tree t);
  void   preamble (string style, string lan);
  void   postamble ();
};

/******************************************************************************
* Basic output subroutines
******************************************************************************/

void
to_tex::print (string s) {
  if (disable_output) return;
  if (lf) {
    if (no_lf) s= "\n" * s;
    else s = "\n\n" * s;
    lf= no_lf= FALSE;
  }

  int i, n=N(s);
  for (i=0; i<n; i++) {
    // remove superfluous space
    if (N(buffer)>0) {
      if (s[i]==' ') {
	char c= buffer[N(buffer)-1];
	if (c==' ') continue;
	if (line_count==0) continue;
	if (in_mode ("math")) {
	  if ((c=='_') || (c=='^') || (c=='$') || (c=='{') || (c=='['))
	    continue;
	}
      }
      if ((s[i]==']') || (s[i]=='}')) {
	char c= buffer[N(buffer)-1];
	if ((c==' ') && in_mode ("math")) {
	  buffer[N(buffer)-1]= s[i];
	  for (last_space= N(buffer)-1; last_space>=0; last_space--)
	    if (buffer[last_space]==' ') break;
	  continue;
	}
      }
    }

    // do line breaking
    if (s[i]==' ') last_space=N(buffer);
    buffer << s[i];
    line_count++;
    if (s[i]=='\n') line_count=0;
    if ((line_count>=79) && (last_space!=-1)) {
      buffer[last_space]='\n';
      line_count=N(buffer)-last_space;
      last_space=-1;
    }
  }
}

void
to_tex::remove (int n) {
  if (disable_output) return;
  if (N(buffer)<=n) buffer= "";
  else buffer.resize (N(buffer)-n);
}

void
to_tex::sep (bool text_also, bool force) {
  if (disable_output) return;
  if (force) {
    buffer << " ";
    line_count++;
    return;
  }
  if (text_also || in_mode ("math"))
    if ((N(buffer)>0) && (buffer[N(buffer)-1]!=' '))
      print (" ");
}

void
to_tex::ret () {
  if (line_count>0) print ("\n");
}

string
to_tex::last (int n) {
  if (N(buffer)<n) return "";
  else return buffer (N(buffer)-n, N(buffer));
}

/******************************************************************************
* Text output
******************************************************************************/

void
to_tex::print_cmd (string cmd) { 
  if (in_mode ("math")) { sep(); print (cmd); sep(); }
  else print ("{" * cmd * "}");
}

void
to_tex::tex_arg (tree t, bool flag) {
  if (flag && (N(t)==0) && (N(t->label)==1)) {
    print (t->label);
  }
  else {
    print ("{");
    tex (t);
    print ("}");
  }
}

void
to_tex::tex_string (string s) {
  int i;
  for (i=0; i<N(s); i++)
    if (s[i]=='<') {
      i++;
      int start=i;
      while ((i<N(s)) && (s[i]!='>')) i++;
      string r= "\\" * s (start, i);
      if ((N(r)>=6) && (r (0, 6) == "\\left-"))
	r= "\\left" * r (6, N(r));
      if ((N(r)>=5) && (r (0, 5) == "\\mid-"))
	r= r (5, N(r));
      if ((N(r)>=7) && (r (0, 7) == "\\right-"))
	r= "\\right" * r (7, N(r));
      if ((N(r)>=7) && (r (0, 5) == "\\big-"))
	r= "\\big" * r (5, N(r));
      if (r == "\\bigsum") r= "\\sum";
      if (r == "\\bigprod") r= "\\prod";
      if (r == "\\bigint") r= "\\int";
      if (r == "\\bigoint") r= "\\oint";
      if (r == "\\less") r= "<";
      if (r == "\\gtr")  r= ">";
      print_cmd (r);
    }
    else {
      if (in_mode ("math")) {
	if (s[i]=='\'') { print (s[i]); continue; }
	sep ();
	if (s[i]=='*') continue;
	if (s[i]=='{') { print ("\\{"); continue; }
	if (s[i]=='}') { print ("\\}"); continue; }
	if (is_alpha (s[i])) {
	  int start=i;
	  while ((i<N(s)) && is_alpha (s[i])) i++;
	  i--;
	  if (i!=start) {
	    string ss= s (start, i+1);
	    if (command_type ["\\" * ss] == "operator")
	      print ("\\" * ss * " ");
	    else print ("\\mathrm{" * ss * "}\\;");
	    continue;
	  }
	}
	if (is_numeric (s[i])) {
	  while ((i<N(s)) && is_numeric (s[i])) {
	    print (s[i]);
	    i++;
	  }
	  i--;
	  continue;
	}
	if ((s[i]=='#') && (i<(N(s)-1)) && is_numeric (s[i+1])) {
	  print (s[i++]);
	  print (s[i]);
	  continue;
	}
      }
      print (s[i]);
    }
}
  
/******************************************************************************
* Routines concerning the environment
******************************************************************************/

void
to_tex::assign (string var, tree val) {
  if (is_func (env[var], BACKUP, 2))
    env(var)= tree (BACKUP, val, env[var][1]);
  else env(var)= val;
}

void
to_tex::set (string var, tree val) {
  env(var)= tree (BACKUP, val, env[var]);
}

void
to_tex::reset (string var) {
  if (is_func (env[var], BACKUP, 2))
    env(var)= env[var][1];
}

bool
to_tex::in_mode (string which) {
  tree t= env[MODE];
  if (is_func (t, BACKUP, 2)) t=t[0];
  return t->label == which;
}

string
to_tex::get_decl (string var, string val) {
  if (var==FONT_SIZE) {
    int sz= (int) ((as_double (val)*10.0) + 0.5);
    if (sz<=5) return "\\tiny";
    else if (sz<=6) return "\\scriptsize";
    else if (sz<=7) return "\\footnotesize";
    else if (sz<=9) return "\\small";
    else if (sz<=11) return "\\normalsize";
    else if (sz<=13) return "\\large";
    else if (sz<=15) return "\\Large";
    else if (sz<=18) return "\\LARGE";
    else if (sz<=22) return "\\huge";
    else return "\\Huge";
  }
  if (var==TEXT_FAMILY) {
    if (val == "rm") return "\\rmfamily";
    if (val == "ss") return "\\ssfamily";
    if (val == "tt") return "\\ttfamily";
  }
  if (var==TEXT_SERIES) {
    if (val == "medium") return "\\mdseries";
    if (val == "bold") return "\\bfseries";
  }
  if (var==TEXT_SHAPE) {
    if (val == "right") return "\\upshape";
    if (val == "slanted") return "\\slshape";
    if (val == "italic") return "\\itshape";
    if (val == "small-caps") return "\\scshape";
  }
  if (var==COLOR) return "\\" * val;
  return "";
}

string
to_tex::get_cmd (string var, string val) {
  if (var==TEXT_FAMILY) {
    if (val == "rm") return "\\textrm";
    if (val == "ss") return "\\textss";
    if (val == "tt") return "\\texttt";
  }
  if (var==TEXT_SERIES) {
    if (val == "medium") return "\\textmd";
    if (val == "bold") return "\\textbf";
  }
  if (var==TEXT_SHAPE) {
    if (val == "right") return "\\textup";
    if (val == "slanted") return "\\textsl";
    if (val == "italic") return "\\textit";
    if (val == "small-caps") return "\\textsc";
  }
  if (var==MATH_FONT) {
    if (val == "cal") return "\\mathcal";
    if (val == "Euler") return "\\mathfrak";
    if (val == "Bbb") return "\\mathbb";
  }
  if (var==MATH_SERIES) {
    if (val == "bold") return "\\mathbf";
  }
  return "";
}

static tree
arg_to_sharp (tree t, tree args, int n) {
  if (N(t)==0) return t;

  if (is_func (t, APPLY, 1)) {
    int i;
    for (i=0; i<n; i++)
      if (t[0] == args[i]) {
	string s ("#0");
	s[1]= (char) (((int) '1') + i);
	return s;
      }
  }

  int i;
  tree r (t->label, N(t));
  for (i=0; i<N(t); i++)
    r[i]= arg_to_sharp (t[i], args, n);
  return r;
}

void
to_tex::tex_assign (string var, tree val) {
  assign (var, val->label);
  string decl= get_cmd (var, val->label);
  if (decl == "") {
    if (is_func (val, FUNCTION)) {
      sep ();
      print ("\\newcommand{\\" * var * "}");
      if (N(val)>1) print ("[" * as_string (N(val)-1) * "]");
      print ("{");
      tex (arg_to_sharp (val[N(val)-1], val, N(val)-1));
      print ("}");
      sep ();
    }
    else if (is_func (val, ENVIRONMENT)) {
      sep ();
      print ("\\newenvironment{" * var * "}");
      if (N(val)>2) print ("[" * as_string (N(val)-2) * "]");
      print ("{");
      tex (arg_to_sharp (val[N(val)-2], val, N(val)-2));
      print ("}{");
      tex (arg_to_sharp (val[N(val)-1], val, N(val)-2));
      print ("}");
      sep ();
    }
    else {
      sep ();
      print ("\\newcommand{\\" * var * "}{");
      tex (val);
      print ("}");
      sep ();      
    }
  }
  else {
    sep (TRUE);
    print (decl);
    sep (TRUE);
  }
}

void
to_tex::tex_set_var (string var, string val) {
  set (var, val);
  if (var==MODE) {
    if (val == "text") print ("\\text{");
    if (val == "math") {
      if (last()=="$") sep (TRUE, TRUE);
      print ("$");
    }
  }
  else {
    string cmd= get_cmd (var, val);
    if (cmd == "") {
      string decl= get_cmd (var, val);
      if (decl != "") print ("{" * decl * " ");
    }
    else print (cmd * "{");
  }
}

void
to_tex::tex_reset_var (string var) {
  if (N(env[var])==0) return;
  string val (env[var][0]->label);
  if (var==MODE) {
    if (val == "text") print ("}");
    if (val == "math") {
      if (last()=="$") sep (TRUE, TRUE);
      print ("$");
    }
  }
  if ((get_cmd (var, val) != "") || (get_decl (var, val) != "")) print ("}");
  reset (var);
}
  
/******************************************************************************
* User and system commands and environments
******************************************************************************/

void
to_tex::tex_apply(tree t) {
  if ((N(t)==1) && (t[0]=="item")) {
    no_lf= TRUE;
    sep (TRUE);
    print ("\\item");
    sep (TRUE);
  }
  else if ((N(t)==2) && (t[0]=="item*")) {
    no_lf= TRUE;
    sep (TRUE);
    print ("\\item["); tex (t[1]); print ("]");
    sep (TRUE);
  }
  else if ((N(t)==2) && (t[0]=="bibitem")) {
    no_lf= TRUE;
    sep (TRUE);
    print ("\\bibitem{"); tex (t[1]); print ("}");
    sep (TRUE);
  }
  else if ((N(t)==3) && (t[0]=="bibitem*")) {
    no_lf= TRUE;
    sep (TRUE);
    print ("\\bibitem");
    print ("["); tex (t[1]); print ("]");
    print ("{"); tex (t[2]); print ("}");
    sep (TRUE);
  }
  else if (N(t)==1) print_cmd ("\\" * t[0]->label);
  else {
    int i;
    sep(); print ("\\"); print (t[0]->label);
    for (i=1; i<N(t); i++) tex_arg (t[i]);
    sep();
  }
}

#define ENV_NORMAL       0
#define ENV_LIST         1
#define ENV_EQUATION     2
#define ENV_BIBLIOGRAPHY 3

int
to_tex::tex_special_env (string& s) {
  if ((N(s)>=7) && (s(0,7)=="itemize")) {
    if (s == "itemize-minus") s= "itemizeminus";
    else if (s == "itemize-dot") s= "itemizedot";
    else if (s == "itemize-arrow") s= "itemizearrow";
    else s= "itemize";
    return ENV_LIST;
  }
  if ((N(s)>=9) && (s(0,9)=="enumerate")) {
    if (s == "enumerate-numeric") s= "enumeratenumeric";
    else if (s == "enumerate-roman") s= "enumerateroman";
    else if (s == "enumerate-Roman") s= "enumerateromancap";
    else if (s == "enumerate-alpha") s= "enumeratealpha";
    else if (s == "enumerate-Alpha") s= "enumeratealphacap";
    else s= "enumerate";
    return ENV_LIST;
  }
  if ((N(s)>=11) && (s(0,11)=="description")) {
    s= "description";
    return ENV_LIST;
  }

  if ((s == "equation") || (s == "equation*")) return ENV_EQUATION;
  if ((s == "eqnarray") || (s == "eqnarray*")) return ENV_EQUATION;

  if (s == "bibliography") return ENV_BIBLIOGRAPHY;

  return ENV_NORMAL;
}

void
to_tex::tex_begin (tree t) {
  int i;
  string s= t[0]->label;

  switch (tex_special_env (s)) {
  case ENV_LIST:
    no_lf= TRUE;
    break;
  case ENV_EQUATION:
    set (MODE, "math");
    no_lf= TRUE;
    break;
  case ENV_BIBLIOGRAPHY:
    if (N(t)==4) {
      ret ();
      print ("\\bibliographystyle{" * t[2]->label * "}\n");
      print ("\\bibliography{" * t[3]->label * "}\n");
      disable_output= TRUE;
      return;
    }
  }

  sep();
  print ("\\begin{" * s * "}");
  for (i=1; i<N(t); i++) tex_arg (t[i]);
  ret();
}

void
to_tex::tex_end (tree t) {
  int env_type;
  string s= t[0]->label;
  env_type= tex_special_env (s);

  if (env_type != ENV_BIBLIOGRAPHY) {
    ret();
    print ("\\end{" * s * "}");
    sep();
  }

  switch (env_type) {
  case ENV_LIST:
    no_lf= TRUE;
    break;
  case ENV_EQUATION:
    no_lf= TRUE;
    reset (MODE);
    break;
  case ENV_BIBLIOGRAPHY:
    disable_output= FALSE;
    break;
  }
}

/******************************************************************************
* mathematics
******************************************************************************/

void
to_tex::tex_script (string s, tree t) {
  bool flag;
  while (last()==" ") remove ();
  flag= (last(2)=="\\;");
  if (flag) remove (2);
  print (s);
  tex_arg (t, TRUE);
  if (flag) print ("\\;");
}

void
to_tex::tex_matrix_table (tree t, bool is_table) {
  if (N(t)<2) return;
  int rows= as_int (t[N(t)-1]->label);
  int cols= as_int (t[N(t)-2]->label);
  if (N(t)!=(rows*cols+2)) return;
  int i, j;

  sep();
  print ("\\begin{array}{");
  if (is_table) print ("|");
  for (j=0; j<cols; j++)
    if (is_table) print ("c|");
    else print ("c");
  print ("}");
  sep();

  if (is_table) print_cmd ("\\hline");
  for (i=0; i<rows; i++) {
    for (j=0; j<cols; j++) {
      if (j>0) { sep(); print ("&"); sep(); }
      tex (t[i*cols+j]);
    }
    print ("\\\\"); sep();
    if (is_table) print_cmd ("\\hline");
  }

  sep();
  print ("\\end{array}");
  sep();
}

/******************************************************************************
* to_tex main routine
******************************************************************************/

void
to_tex::tex (tree t) {
  int i;

  // cout << "t] " << t << "\n";
  if (N(t)==0) {
    tex_string (t->label);
    return;
  }

  switch (math_primary (t->label)) {
  case Document:
    for (i=0; i<N(t)-1; i++) {
      if (t[i] == "") no_lf= FALSE;
      else tex (t[i]);
      lf= TRUE;
    }
    tex (t[i]);
    return;
  case Paragraph:
    tex (t[0]);
    return;
  case Concat:
    for (i=0; i<N(t); i++) tex (t[i]);
    return;
  case Format:
    if (t[0]->label == NO_FIRST_INDENT) print ("\\noindent ");
    if (t[0]->label == LINE_BREAK) print ("\\linebreak");
    if (t[0]->label == NEW_LINE) print ("\n\n");
    if (t[0]->label == PAGE_BREAK) print ("\\pagebreak");
    if (t[0]->label == NEW_PAGE) print ("\\newpage");
    if (t[0]->label == LINE_SEP) { print (" & "); }
    if (t[0]->label == NEXT_LINE) print ("\\\\\n");
    return;
  case Hspace:
    if (N(t)==1) print ("\\hspace{" * t[0]->label * "}");
    else print ("\\hspace{" * t[1]->label * "}");
    return;
  case Vspace_before:
    return;
  case Vspace_after:
    if (N(t)==1) print ("\\vspace{" * t[0]->label * "}");
    else print ("\\vspace{" * t[1]->label * "}");
    return;
  case Space:
    return;
  case Htab:
    print ("\\hspace*{\\fill} ");
    return;
  case Split:
  case Move:
  case Resize:
    return;

  case Left_sub:
    sep();
    print ("\\sideset{_");
    tex_arg (t[0], TRUE);
    print ("}{}");
    sep();
    return;
  case Left_sup:
    sep();
    print ("\\sideset{^");
    tex_arg (t[0], TRUE);
    print ("}{}");
    sep();
    return;
  case Right_sub:
    tex_script ("_", t[0]);
    return;
  case Right_sup:
    tex_script ("^", t[0]);
    return;
  case Frac:
    if (N(t)<2) return;
    if (latex_flag) { sep(); print ("\\frac"); }
    else print ("{");
    tex_arg (t[0]);
    if (!latex_flag) print (" \\over ");
    tex_arg (t[1]);
    if (latex_flag) sep();
    else print ("}");
    return;
  case Sqrt:
    if (N(t)==1) {
      sep();
      print ("\\sqrt");
      tex_arg (t[0]);
      sep();
    }
    else {
      sep();
      print ("\\sqrt[");
      tex (t[1]);
      print ("]");
      tex_arg (t[0]);
      sep();
    }
    return;
  case Wide :
    if (N(t)<2) return;
    {
      string s= t[1]->label;
      sep();
      if (s == "^") print ("\\hat");
      else if (s == "~") print ("\\tilde");
      else if (s == "<check>") print ("\\check");
      else if (s == "<acute>") print ("\\acute");
      else if (s == "<grave>") print ("\\grave");
      else if (s == "<vect>") print ("\\vec");
      else print ("\\hat");
      tex_arg (t[0]);
      sep();
    }
    return;
  case Neg :
    sep();
    print ("\\not");
    tex_arg (t[0]);
    sep();
    return;
  case Tree :
    return;
  case Matrix:
    tex_matrix_table (t, FALSE);
    return;
  case Table:
    tex_matrix_table (t, TRUE);
    return;
  case Mosaic:
  case Mosaic_item:
    return;

  case Inactive:
  case Symbol:
  case Latex:
  case Hybrid:
  case Backup:
  case Quote:
    return;
  case Assign:
    if (N(t)<2) return;
    tex_assign (t[0]->label, t[1]);
    return;
  case Set:
    if (N(t)<2) return;
    tex_set_var (t[0]->label, t[1]->label);
    return;
  case Reset:
    tex_reset_var (t[0]->label);
    return;
  case Apply:
    tex_apply (t);
    return;
  case Begin:
    tex_begin (t);
    return;
  case End:
    tex_end (t);
    return;
  case Function:
  case Environment:
    return;
  case Label:
    sep();
    print ("\\label{");
    print (t[0]->label);
    print ("}");
    sep();
    return;
  case Reference:
    sep();
    print ("\\ref{");
    print (t[0]->label);
    print ("}");
    sep();
    return;
  case Write:
    return;

  case Plus:
  case Minus:
  case Times:
  case Merge:
  case Number:
  case Translate:
    return;

  case Graphics:
  case Point:
  case Line:
  case Arc:
  case Bezier:
  case Postscript:
    return;

  default:
    cout << "\nTree: " << t << "\n";
    fatal_error ("not yet implemented", "tex");
  }
}

/******************************************************************************
* generation of the preamble and driver program
******************************************************************************/

void
to_tex::preamble (string style, string lan) {
  print ("\\documentclass{" * style * "}\n");

  if (lan == "dutch") print ("%\\usepackage{dutch}\n");
  if (lan == "french") print ("\\usepackage{french}\n");
  if (lan == "german") print ("%\\usepackage{german}\n");

  print ("\\IfFileExists{TeXmacs.sty}\n");
  print ("  {\\usepackage{TeXmacs}}\n");
  print ("  {\\usepackage{$TEXMACS_PATH/misc/latex/TeXmacs}}%$\n");
  print ("\n\\begin{document}\n");

  if (lan == "dutch") print ("\\SelectDutch\n");
  if (lan == "french") print ("\\SelectFrench\n");
  if (lan == "german") print ("\\SelectGerman\n");

  print ("\n");
}

void
to_tex::postamble () {
  print ("\n\n\\end{document}\n");
}

to_tex::to_tex (bool latex_flag2): env (tree ("?")) {
  last_space     = -1;
  line_count     = 0;
  latex_flag     = latex_flag2;
  lf             = FALSE;
  no_lf          = FALSE;
  disable_output = FALSE;
  env[MODE]      = "text";
}

/******************************************************************************
* interface
******************************************************************************/

static string
edit_to_tex_latex (
  tree t, string style, string mode, string lan,
  bool latex_flag, bool document_flag)
{
  initialize_tex_conversions ();
  (void) mode;
  to_tex tt (latex_flag);
  if (document_flag) tt.preamble (style, lan);
  tt.env (MODE)= mode;
  tt.tex (t);
  if (document_flag) tt.postamble ();

  /*
  cout << "Edit to latex" << "\n";
  cout << "  t    = " << t     << "\n";
  cout << "  style= " << style << "\n";
  cout << "  mode = " << mode  << "\n";
  cout << "  lan  = " << lan   << "\n";
  cout << "---\n";
  cout << tt.buffer << "\n";
  cout << "---\n";
  */

  return tt.buffer;
}

string
edit_to_tex (tree t, string mode) {
  return edit_to_tex_latex (t, "article", mode, "english", FALSE, FALSE);
}

string
edit_to_tex_document (tree t, string style, string lan) {
  return edit_to_tex_latex (t, style, "text", lan, FALSE, TRUE);
}

string
edit_to_latex (tree t, string mode) {
  return edit_to_tex_latex (t, "article", mode, "english", TRUE, FALSE);
}

string
edit_to_latex_document (tree t, string style, string lan) {
  return edit_to_tex_latex (t, style, "text", lan, TRUE, TRUE);
}

#endmodule // code_totex
