
/******************************************************************************
* MODULE     : parser.gen.h
* DESCRIPTION: dynamic parser
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <program.gen.h>
#include <hashmap.gen.h>
#include <hashset.gen.h>
#include <resource.gen.h>

#module parser
#import program
#import hashset (string)
#import resource (parser_rep, parser)

#define TRANSIT  0
#define PRODUCE  1
#define TERMINAL 2

/******************************************************************************
* Some ``automatic'' declarations
******************************************************************************/

class symbol_rep;
class symbol {
#import concrete_null (symbol, symbol_rep)
  symbol (string name);
  friend bool operator == (symbol sy1, symbol sy2);
  friend bool operator != (symbol sy1, symbol sy2);
  friend ostream& operator << (ostream& out, symbol sy);
};

class state_rep;
class state {
#import concrete_null (state, state_rep)
  state (bool dummy);
  friend bool operator == (state st1, state st2);
  friend bool operator != (state st1, state st2);
  friend ostream& operator << (ostream& out, state st);
};

#define no_tree_converter<symbol>
#define no_tree_converter<state>
#import array (symbol)
#import array (state)
#import hashmap (string, symbol)
#import hashmap (string, state)

/******************************************************************************
* Each symbol 'sy' corresponds to a set of parse rules.
*   - When parsing a string of type 'sy', we start at the initial state and
*     continue until we meet the terminal state.
*   - The token table contains the set of possible tokens with which
*     the parse rules start and the symbol table to the set of
*     possible starting symbols.
******************************************************************************/

class symbol_rep: public concrete_struct {
  string           name;
  state            init;
  state            term;
  hashset<string>  token_table;
  hashset<string>  symbol_table;

public:
  symbol_rep (string name);
  friend class parser_rep;
};
#import code_concrete_null (symbol, symbol_rep)

/******************************************************************************
* Each parse rule associated to a symbol 'sy' amounts to a sequence of states.
*   - We either advance by an epsilon move, a token or a symbol.
*   - We advance by a symbol 'sy2' whenever the next token matches
*     a starting token for 'sy2'.
*   - At the before last state of the rule, we produce the result.
*   - The last state coincides with the terminal state of 'sy';
*     we move to it using an epsilon move.
******************************************************************************/

class state_rep: public concrete_struct {
  program                production;    // production
  state                  epsilon;       // next state after epsilon movement
  hashmap<string,state>  token_table;   // next state after token
  hashmap<string,state>  symbol_table;  // next state after symbol

public:
  state_rep ();
  friend class parser_rep;
};
#import code_concrete_null (state, state_rep)

/******************************************************************************
* Once the parser has been constructed from a grammar program (.grm file),
* the parser is ready for use.
******************************************************************************/

struct parser_rep: rep<parser> {
  hashmap<string,state>  state_table;   // the grammar states
  hashmap<string,symbol> symbol_table;  // the grammar symbols

  lexer    parse_lx;     // the lexer which provides the tokens
  string   parse_s;      // the next token read by the lexer
  string   parse_type;   // the lexer type of the token
  location parse_start;  // the start of the token in the source code
  location parse_end;    // the end of the token in the source code

public:
  /*********** Building the parser *************/
  parser_rep (string name, program p);
  ~parser_rep ();
  void insert_symbol (string name);
  void insert_method (string name, program p1, program p2);
  void complete_automaton ();

  /****************** Parsing  *****************/
  void set_input (lexer lx);
  void read ();
  program produce (array<program> a, program p);
  program parse (string sy);
  program parse ();

  /************** Error handling ***************/
  void grammar_error (string s);
  void error (string s);
};

parser load_parser (string name);

#endmodule // parser
