
/******************************************************************************
* MODULE     : notify.gen.cc
* DESCRIPTION: undertake the appropriate action after
*              notification of changes in edit tree
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <data.gen.h>
#include <edit_modify.gen.h>
#include <edit_cursor.gen.h>
#include <tm_buffer.gen.h>

#module code_notify
#import data
#import edit_modify
#import edit_cursor
#import tm_buffer

/******************************************************************************
* Constructors and destructors
******************************************************************************/

edit_modify::edit_modify (tm_buffer buf): edit_base (buf) {}
edit_modify::~edit_modify () {}

/******************************************************************************
* modification routines
******************************************************************************/

void
edit_modify::assign (tree& t, path p, tree u) {
  // cout << "Assign " << p << ", " << u << "\n";
  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_assign (t, p, u);

  subtree (t, p)= u;
  finished (t);
}

void
edit_modify::insert (tree& t, path p, tree u) {
  // cout << "Insert " << p << ", " << u << "\n";
  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_insert (t, p, u);

  insert_at (t, p, u);
  finished (t);
}

void
edit_modify::remove (tree& t, path p, int nr) {
  // cout << "Remove " << p << ", " << nr << "\n";
  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_remove (t, p, nr);

  remove_at (t, p, nr);
  finished (t);
}

void
edit_modify::split (tree& t, path p) {
  // cout << "Split " << p << "\n";
  if (N(p)<2) fatal_error ("path too short in split", "editor::split");
  tree& st (subtree (t, path_up (path_up (p))));
  int  l1 = last_item (path_up (p));
  int  l2 = last_item (p);

  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_split (t, p);

  if (N(st[l1])==0) {
    string s1, s2;
    ::split (st[l1]->label, l2, s1, s2);
    st[l1]= s2;
    st= insert_one (st, l1, tree (s1));
  }
  else {
    tree st1, st2;
    ::split (st[l1], l2, st1, st2);
    st[l1]= st2;
    st= insert_one (st, l1, st1);
  }
  finished (t);
}

void
edit_modify::join (tree& t, path p) {
  // cout << "Join " << p << "\n";
  if (N(p)<1) fatal_error ("path too short in join", "editor::join");
  tree& st (subtree (t, path_up (p)));
  int  l1 = last_item (p);
  int  l2 = N (st[l1]);
  if (l2==0) l2= N (st[l1]->label);
  if (l1+1 >= N(st)) fatal_error ("invalid join", "editor::join");
  if ((N(st[l1])==0) != (N(st[l1+1])==0)) {
    cerr << "Arg[1]= " << st[l1] << "\n";
    cerr << "Arg[2]= " << st[l1+1] << "\n";
    fatal_error ("either both or no arguments in join should be strings",
		 "editor::join");
  }

  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_join (t, p);

  if (N(st[l1])==0) st[l1]->label << st[l1+1]->label;
  else st[l1] << st[l1+1]->a;
  st= ::remove (st, l1+1, 1);
  finished (t);
}

void
edit_modify::ins_unary (tree& t, path p, string op) {
  // cout << "Insert unary " << p << ", " << op << "\n";
  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_ins_unary (t, p, op);

  tree& st (subtree (t, p));
  st= tree (op, st);
  finished (t);
}

void
edit_modify::rem_unary (tree& t, path p) {
  // cout << "Remove unary " << p << "\n";
  tree& st (subtree (t, p));
  if (N(st)!=1) fatal_error ("not a unary tree", "editor::rem_unary");

  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->notify_rem_unary (t, p);

  st= st[0];
  finished (t);
}

void
edit_modify::finished (tree& t) {
  int i;
  for (i=0; i<N(buf->vws); i++)
    ((tm_view) (buf->vws[i]))->ed->post_notify (t);
}

/*****************************************************************************/
// Cursor handling after notification of changes in document
/*****************************************************************************/

void
edit_modify::notify_assign (tree& t, path p, tree u) { (void) u;
  if ((&t)!=(&et)) return;
  if (p<tp) tp= path_down (p, 0);

  if (nil (p)) typeset_notify_assign (N(u));
  else typeset_notify_modify (p->item);
}

void
edit_modify::notify_insert (tree& t, path p, tree u) {
  if ((&t)!=(&et)) return;
  if ((N(p)>=2) && path_inf (path_up (p), tp));
  else if (path_inf (p, tp) || (p <= tp)) {
    int nr= (N(u)==0? N(u->label): N(u));
    tp[N(p)-1] += nr;
  }
  
  if (atom (p)) typeset_notify_insert (p->item, N(u));
  else typeset_notify_modify (p->item);
}

void
edit_modify::notify_remove (tree& t, path p, int nr) {
  if ((&t)!=(&et)) return;
  if ((N(p)>=2) && path_inf (path_up (p), tp));
  else {
    tree& st (subtree (et, path_up (p)));
    path p2= path_add (p, nr, N(p)-1);
    if ((path_up (p) <= tp) && (N (st) == nr))
      tp= path_down (path_up (p), 0);
    else if (path_inf (p2, tp) || (p2 <= tp)) tp[N(p)-1] -= nr;
    else if (path_inf (p, tp)  || (p <= tp)) {
      if ((N(st)==0) || (last_item(p)==0)) tp= copy (p);
      else tp= path_down (path_dec (p), right_index (st [last_item(p)-1]));
    }
  }

  if (atom (p)) typeset_notify_remove (p->item, nr);
  else typeset_notify_modify (p->item);
}

void
edit_modify::notify_split (tree& t, path p) {
  if ((&t)!=(&et)) return;
  if (!(path_up (p, 2) <= path_up (tp)));
  else if (path_up (p, 2) == path_up (tp));
  else if (tp[N(p)-2]<p[N(p)-2]);
  else if (tp[N(p)-2]>p[N(p)-2]) tp[N(p)-2] ++;
  else if ((path_up (p) == path_up (tp)) &&
	   (N (subtree (et, path_up (p))) != 0)) {
    if (last_item (tp)==1) tp[N(p)-2] ++;
  }
  else if (tp[N(p)-1]<p[N(p)-1]);
  else {
    tp[N(p)-2] ++;
    tp[N(p)-1] -= last_item (p);
  }

  typeset_notify_modify (p->item);
  if (N(p) == 2) typeset_notify_insert (p->item, 1);
}

void
edit_modify::notify_join (tree& t, path p) {
  if ((&t)!=(&et)) return;
  tree& st (subtree (et, p));
  bool flag = (N(st)==0);
  int  extra= (flag? N(st->label): N(st));
  if (!(path_up (p) <= path_up (tp)));
  else if (path_up (p) == path_up (tp));
  else if (tp[N(p)-1] <= p[N(p)-1]);
  else {
    if ((tp[N(p)-1] == p[N(p)-1]+1) &&
	(flag || (path_inc (p) != path_up (tp)))) tp[N(p)] += extra;
    tp[N(p)-1] --;
  }

  typeset_notify_modify (p->item);
  if (N(p) == 1) typeset_notify_remove (p->item+1, 1);
}

void
edit_modify::notify_ins_unary (tree& t, path p, string op) { (void) op;
  if ((&t)!=(&et)) return;
  if (p <= path_up (tp)) {
    path add= path (0, tail (tp, N(p)));
    tp= merge (copy (p), add);
  }

  if (nil (p)) typeset_notify_assign (1);
  else typeset_notify_modify (p->item);
}

void
edit_modify::notify_rem_unary (tree& t, path p) {
  if ((&t)!=(&et)) return;
  if (p == path_up (tp)) {
    if (last_item (tp)==1)
      tp[N(tp)-1]= right_index (subtree (et, path_down (p, 0)));
  }
  else if (p <= path_up (tp)) {
    path add= tail (tp, N(p)+1);
    tp= merge (p, add);
  }

  if (nil (p)) typeset_notify_assign (N(et[0]));
  else typeset_notify_modify (p->item);
}

void
edit_modify::post_notify (tree& t) {
  if ((&t)!=(&et)) return;
  notify_change (THE_TREE);
  set_cursor (correct_cursor (et, tp));
  buf->modification_flag= TRUE;
}

#endmodule // code_notify
