
/******************************************************************************
* MODULE     : glue.gen.cc
* DESCRIPTION: Glue for linking TeXmacs commands to guile
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <server.gen.h>

#module code_glue
#import server
#include <string.h>
#include <gh.h>

typedef SCM (*FN)();

extern string tree_to_string (tree t);
extern void initialize_glue_editor ();
extern void initialize_glue_server ();

/******************************************************************************
* The server
******************************************************************************/

static server* ptr_the_sv= NULL;

server
get_server () {
  if (ptr_the_sv==NULL)
    fatal_error ("TeXmacs server not yet started", "get_server");
  return *ptr_the_sv;
}

/******************************************************************************
* Booleans
******************************************************************************/

#define SCM_ASSERT_BOOL(flag,arg,rout) \
  SCM_ASSERT (gh_boolean_p (flag), flag, arg, rout)

static SCM
bool_to_scm (bool flag) {
  return gh_bool2scm (flag);
}

static bool
scm_to_bool (SCM flag) {
  return gh_scm2bool (flag);
}

/******************************************************************************
* Integers
******************************************************************************/

#define SCM_ASSERT_INT(i,arg,rout) \
  SCM_ASSERT (SCM_INUMP (i), i, arg, rout);

/*static*/
SCM
int_to_scm (int i) {
  return gh_long2scm ((long) i);
}

static int
scm_to_int (SCM i) {
  return (int) gh_scm2long (i);
}

/******************************************************************************
* Strings
******************************************************************************/

#define SCM_ASSERT_STRING(s,arg,rout) \
  SCM_ASSERT (SCM_NIMP (s) && SCM_STRINGP (s), s, arg, rout)

static SCM
string_to_scm (string s) {
  char* _s= as_charp (s);
  SCM r= gh_str2scm (_s, strlen (_s));
  delete[] _s;
  return r;
}

static string
scm_to_string (SCM s) {
  int len_r;
  char* _r= gh_scm2newstr (s, &len_r);
  string r= _r;
  free (_r);
  return r;
}

/******************************************************************************
* Trees
******************************************************************************/

static long tree_tag;

#define SCM_ASSERT_TREE(t,arg,rout) \
  SCM_ASSERT ((SCM_NIMP (t) && SCM_CAR (t) == tree_tag), t, arg, rout)

static SCM
tree_to_scm (tree t) {
  SCM tree_smob;
  SCM_NEWCELL (tree_smob);
  SCM_SETCDR (tree_smob, new tree (t));
  SCM_SETCAR (tree_smob, tree_tag);
  return tree_smob;
}

static tree
scm_to_tree (SCM tree_smob) {
  return *((tree*) SCM_CDR (tree_smob));
}

static SCM
mark_tree (SCM tree_smob) {
  return SCM_BOOL_F;
}

static scm_sizet
free_tree (SCM tree_smob) {
  tree *ptr = (tree *) SCM_CDR (tree_smob);
  delete ptr;
  return sizeof (tree); // should be replaced by total size of the tree
}

static int
print_tree (SCM tree_smob, SCM port, scm_print_state *pstate) {
  tree   t= scm_to_tree (tree_smob);
  string s= "<tree " * tree_to_string (t) * ">";
  scm_display (string_to_scm (s), port);
  return 1;
}

static SCM
cmp_tree (SCM t1, SCM t2) {
  return gh_bool2scm (scm_to_tree (t1) == scm_to_tree (t2));
}

/******************************************************************************
* Programs
******************************************************************************/

static long program_tag;

#define SCM_ASSERT_PROGRAM(p,arg,rout) \
  SCM_ASSERT ((SCM_NIMP (p) && SCM_CAR (p) == program_tag), p, arg, rout)

/*static*/
SCM
program_to_scm (program p) {
  SCM program_smob;
  SCM_NEWCELL (program_smob);
  SCM_SETCDR (program_smob, new program (p));
  SCM_SETCAR (program_smob, program_tag);
  return program_smob;
}

static program
scm_to_program (SCM program_smob) {
  return *((program*) SCM_CDR (program_smob));
}

static SCM
mark_prg (SCM program_smob) {
  return SCM_BOOL_F;
}

static scm_sizet
free_prg (SCM program_smob) {
  program *ptr = (program *) SCM_CDR (program_smob);
  delete ptr;
  return sizeof (program); // should be replaced by total size of the program
}

static int
print_prg (SCM program_smob, SCM port, scm_print_state *pstate) {
  program p= scm_to_program (program_smob);
  string  s= "<program " * program_to_lisp_expr (p) * ">";
  scm_display (string_to_scm (s), port);
  return 1;
}

static SCM
cmp_prg (SCM p1, SCM p2) {
  return gh_bool2scm (scm_to_program (p1) == scm_to_program (p2));
}

/******************************************************************************
* Pointers
******************************************************************************/

static long pointer_tag;

#define SCM_ASSERT_POINTER(p,arg,rout) \
  SCM_ASSERT ((SCM_NIMP (p) && SCM_CAR (p) == pointer_tag), p, arg, rout)

/*static*/
SCM
pointer_to_scm (pointer p) {
  SCM pointer_smob;
  SCM_NEWCELL (pointer_smob);
  SCM_SETCDR (pointer_smob, p);
  SCM_SETCAR (pointer_smob, pointer_tag);
  return pointer_smob;
}

/*static*/
pointer
scm_to_pointer (SCM pointer_smob) {
  return (pointer) SCM_CDR (pointer_smob);
}

static SCM
mark_ptr (SCM pointer_smob) {
  return SCM_BOOL_F;
}

static scm_sizet
free_ptr (SCM pointer_smob) {
  (void) pointer_smob;
  return 0;
}

static int
print_ptr (SCM pointer_smob, SCM port, scm_print_state *pstate) {
  pointer p= scm_to_pointer (pointer_smob);
  string  s= "<pointer " * as_string ((int) p) * ">";
  scm_display (string_to_scm (s), port);
  return 1;
}

static SCM
cmp_ptr (SCM p1, SCM p2) {
  return gh_bool2scm (scm_to_pointer (p1) == scm_to_pointer (p2));
}

/******************************************************************************
* References to programs
******************************************************************************/

typedef program& program_ref;

#define SCM_ASSERT_PROGRAM_REF(p,arg,rout) \
  SCM_ASSERT ((SCM_NIMP (p) && SCM_CAR (p) == pointer_tag), p, arg, rout)

/*static*/
SCM
program_ref_to_scm (program_ref p) {
  return pointer_to_scm ((pointer) &p);
}

/*static*/
program_ref
scm_to_program_ref (SCM pointer_smob) {
  return *((program*) scm_to_pointer (pointer_smob));
}

/******************************************************************************
* Initialization
******************************************************************************/

static void
real_initialize_glue (pointer ptr_sv) {
  ptr_the_sv= (server*) ptr_sv;
  tree_tag  = scm_make_smob_type_mfpe
    ("tree", sizeof(tree), mark_tree, free_tree, print_tree, cmp_tree);
  program_tag = scm_make_smob_type_mfpe
    ("program", sizeof(program), mark_prg, free_prg, print_prg, cmp_prg);
  pointer_tag = scm_make_smob_type_mfpe
    ("pointer", 0, mark_ptr, free_ptr, print_ptr, cmp_ptr);
  initialize_glue_editor ();
  initialize_glue_server ();
}

extern "C" {
  void initialize_glue (pointer ptr) {
    real_initialize_glue (ptr);
  }
}

#endmodule // code_glue
