
/******************************************************************************
* MODULE     : program.gen.cc
* DESCRIPTION: Intern representation of uncompiled and compiled programs
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <program.gen.h>

#module code_program
#import program

/******************************************************************************
* The program class
******************************************************************************/

program_rep::program_rep (string s2, location start2, location end2):
  status (NONCOMPILED), s(s2), start(start2), end(end2) {}

program_rep::program_rep (instance ins2, location start2, location end2):
  status (COMPILED), ins(ins2), start(start2), end(end2) {}

program_rep::program_rep (array<program> a2, location start2, location end2):
  a(a2), status (COMPOSITE), start(start2), end(end2) {}

program::program (char* s, location start, location end):
  rep (new program_rep (s, start, end)) {}

program::program (string s, location start, location end):
  rep (new program_rep (s, start, end)) {}

program::program (instance ins, location start, location end):
  rep (new program_rep (ins, start, end)) {}

program::program (array<program> a, location start, location end):
  rep (new program_rep (a, start, end)) {}

program&
program::operator [] (int i) {
  return rep->a[i];
}

int
N (program p) {
  return N (p->a);
}

bool
operator == (program p1, program p2) {
  return
    (p1->status==p2->status) && (p1->s==p2->s) &&
    (p1->ins==p2->ins) && (p1->a==p2->a);
}

bool
operator != (program p1, program p2) {
  return
    (p1->status!=p2->status) || (p1->s!=p2->s) ||
    (p1->ins!=p2->ins) || (p1->a!=p2->a);
}

program
composite (program cmd, program arg1, location start, location end) {
  array<program> a(2);
  a[0]= cmd;
  a[1]= arg1;
  return program (a, start, end);
}

program
composite (program cmd, program arg1, program arg2,
	   location start, location end) {
  array<program> a(3);
  a[0]= cmd;
  a[1]= arg1;
  a[2]= arg2;
  return program (a, start, end);
}

bool
is_string (program p) {
  return (p->status == NONCOMPILED);
}

bool
is_instance (program p) {
  return (p->status == COMPILED);
}

bool
is_composite (program p) {
  return (p->status == COMPOSITE);
}

bool
is_compiled (program p) {
  if (p->status == COMPILED) return TRUE;
  if (p->status == NONCOMPILED) return FALSE;
  if (N(p)==0) return FALSE;
  return is_compiled (p[0]);
}

bool
is_string (program p, char* s) {
  return (p->status == NONCOMPILED) && (p->s == s);
}

bool
is_string (program p, string s) {
  return (p->status == NONCOMPILED) && (p->s == s);
}

bool
is_func (program p, char* s, int n) {
  return
    (p->status == COMPOSITE) &&
    ((n==-1) || (N(p) == (n+1))) &&
    is_string (p[0], s);
}

bool
is_func (program p, string s, int n) {
  return
    (p->status == COMPOSITE) &&
    ((n==-1) || (N(p) == (n+1))) &&
    is_string (p[0], s);
}

/******************************************************************************
* Typed programs
******************************************************************************/

typed_program_rep::typed_program_rep (program p2, type t2):
  p (p2), t (t2) {}

typed_program::typed_program (program p, type t):
  rep (new typed_program_rep (p, t)) {}

typed_program::operator tree () {
  return tree (":", (tree) rep->p, (tree) rep->t);
}

ostream&
operator << (ostream& out, typed_program tp) {
  return out << tp->p << " : " << tp->t;
}

bool
operator == (typed_program tp1, typed_program tp2) {
  return (tp1->p==tp2->p) && (tp1->t==tp2->t);
}

bool
operator != (typed_program tp1, typed_program tp2) {
  return (tp1->p!=tp2->p) || (tp1->t!=tp2->t);
}

/******************************************************************************
* Converting a program to a lisp expression
******************************************************************************/

static string
slash (string s) {
  if (N(s)==0) return "\\e";

  int i;
  string r;
  for (i=0; i<N(s); i++)
    switch (s[i]) {
    case '(':
    case ')':
    case ' ':
    case ':':
    case '\'':
    case '\\':
      r << "\\" << s[i];
      break;
    case '\"': // "
      if (((i==0) && (s[N(s)-1]=='\"')) || // "
	  ((i==(N(s)-1)) && (s[0]=='\"'))) // "
	r << s[i];
      else r << "\\" << s[i];
      break;
    case ((char) 0):
      r << "\\0";
      break;
    case '\t':
      r << "\\t";
      break;
    case '\n':
      r << "\\n";
      break;
    default:
      r << s[i];
    }
  return r;
}

program::operator tree () {
  int i;
  switch (rep->status) {
  case NONCOMPILED:
    return slash (rep->s);
  case COMPILED:
    return slash (rep->ins->name);
  case COMPOSITE:
    {
      tree t ("()", N(rep->a));
      for (i=0; i<N(rep->a); i++)
	t[i]= ((tree) rep->a[i]);
      return t;
    }
  default:
    fatal_error ("Unexpected type", "program::operator tree ()");
  }
}

static void
program_to_lisp_expr (string& out, program p, bool flag=TRUE) {
  int i;
  switch (p->status) {
  case NONCOMPILED:
    out << slash (p->s);
    break;
  case COMPILED:
    out << slash (p->ins->name);
    break;
  case COMPOSITE:
    if (flag && is_func (p, "\'", 1)) {
      out << "\'";
      program_to_lisp_expr (out, p[1], !is_func (p[1], ":", 2));
    }
    else if (flag && is_func (p, ":", 2)) {
      program_to_lisp_expr
	(out, p[1], (!is_func (p[1], ":", 2)) && (!is_func (p[1], "\'", 1)));
      out << ":";
      program_to_lisp_expr
	(out, p[2], (!is_func (p[2], ":", 2)) && (!is_func (p[2], "\'", 1)));
    }
    else {
      out << "(";
      for (i=0; i<N(p); i++) {
	if (i>0) out << " ";
	program_to_lisp_expr (out, p[i]);
      }
      out << ")";
    }
    break;
  default:
    fatal_error ("Unexpected type", "program::operator tree ()");
  }
}

string
program_to_lisp_expr (program p) {
  string out;
  program_to_lisp_expr (out, p);
  return out;
}

ostream&
operator << (ostream& out, program p) {
  return out << program_to_lisp_expr (p);
}

/******************************************************************************
* Converting a lisp expression to a program
******************************************************************************/

static bool
is_spc (char c) {
  return (c==' ') || (c=='\t') || (c=='\n');
}

static string
unslash (string s, int start, int end) {
  int i;
  string r;
  for (i=start; i<end; i++)
    if ((s[i]=='\\') && ((i+1)<end))
      switch (s[++i]) {
      case '0': r << ((char) 0); break;
      case 'e': break;
      case 'n': r << '\n'; break;
      case 't': r << '\t'; break;
      default: r << s[i];
      }
    else r << s[i];
  return r;
}

program lisp_to_program (string s, int& i, location root);

program
lisp_get_type (program p, string s, int& i, location root) {
  int start=i;
  while ((i<N(s)) && is_spc(s[i])) i++;
  if ((i>=N(s)) || (s[i]!=':')) { i=start; return p; }

  array<program> a (3);
  a[0]= program (":", root+i, root+i+1);
  a[1]= p;
  i++;
  while ((i<N(s)) && is_spc(s[i])) i++;
  a[2]= lisp_to_program (s, i, root);
  return program (a, p->start, root+i);
}

program
lisp_to_program (string s, int& i, location root) {
  for (; i<N(s); i++)
    switch (s[i]) {

    case ' ':
    case '\t':
    case '\n':
      break;

    case '(':
      {
	int start=i;
	array<program> a;
	i++;
	while (TRUE) {
	  while ((i<N(s)) && is_spc(s[i])) i++;
	  if ((i==N(s)) || (s[i]==')')) break;
	  a << lisp_to_program (s, i, root);
	}
	if (i<N(s)) i++;
	program p (a, root+start, root+i);
	return lisp_get_type (p, s, i, root);
      }

    case '\'':
      {
	int start= i++;
	array<program> a(2);
	a[0]= program ("\'", root+start, root+start+1);
	a[1]= lisp_to_program (s, i, root);
	return program (a, root+start, root+i);
      }
      
    case '\"':
      { // "
	int start= i++;
	while ((i<N(s)) && (s[i]!='\"')) { // "
	  if ((i<N(s)-1) && (s[i]=='\\')) i++;
	  i++;
	}
	if (i<N(s)) i++;
	program p (unslash (s, start, i), root+start, root+i);
	return lisp_get_type (p, s, i, root);
      }
    
    case ';':
      if ((i+1<N(s)) && (s[i+1]==';')) {
	while ((i<N(s)) && (s[i]!='\n')) i++;
	break;
      }

    default:
      {
	int start= i;
	while ((i<N(s)) && (!is_spc(s[i])) && (s[i]!=':') &&
	       (s[i]!='(') && (s[i]!=')')) {
	  if ((i<N(s)-1) && (s[i]=='\\')) i++;
	  i++;
	}
	program p (unslash (s, start, i), root+start, root+i);
	return lisp_get_type (p, s, i, root);
      }
    }

  return program ("", root+i, root+i);
}

program
lisp_expr_to_program (string s, location root) {
  int i=0;
  return lisp_to_program (s, i, root);
}

program
lisp_prg_to_program (string s, location root) {
  array<program> a;
  int i=0;
  while ((i<N(s)) && (is_spc (s[i]) || s[i]==')')) i++;
  while (i<N(s)) {
    a << lisp_to_program (s, i, root+i);
    while ((i<N(s)) && (is_spc (s[i]) || s[i]==')')) i++;
  }
  return program (a, root, root+i);
}

#endmodule // program
