
/******************************************************************************
* MODULE     : compile.gen.cc
* DESCRIPTION: Interactive compilation of programs
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <program.gen.h>
#include <compiler.gen.h>

#module code_compiler
#import program
#import compiler

#define TP  typed_program
#define LTP list<typed_program>

program lisp_to_program (string s, int& i, location root);

/******************************************************************************
* Initialization
******************************************************************************/

compiler::compiler (): rep (new compiler_rep ()) {}

compiler_rep::compiler_rep ():
  env (hashmap<string,array<instance>> (array<instance> ()))
{
  primitive ("skip", "(-> Void Void)", prg_skip);
  primitive ("define!", "System", prg_define);
  primitive ("var!", "System", prg_var);
  primitive ("lambda!", "System", prg_lambda);
  primitive ("seq!", "System", prg_seq);
  primitive ("if!", "System", prg_if);
  primitive ("if_else!", "System", prg_if_else);
  primitive ("Function_coerce!", "System", prg_function_coerce);
  primitive ("Function_to_tree!", "System", prg_function_to_tree);
  instance ins_false (integer_instance ("false", "Bool", FALSE));
  env->define_var (ins_false);
  instance ins_true (integer_instance ("true", "Bool", TRUE));
  env->define_var (ins_true);
}

void
compiler_rep::primitive (string name, string t, pointer (*routine) (program)) {
  int i=0;
  program p (lisp_to_program (t, i, 0));
  instance ins (system_instance (name, p, routine));
  env->define_var (ins);
}

/******************************************************************************
* Main compilation routine
******************************************************************************/

LTP
compiler_rep::compile (program p) {
  switch (p->status) {

  case NONCOMPILED: {
    string& s (p->s);
    if (is_int (s)) return compile_integer (p);
    if ((N(s)>=2) && (s[0]=='\42') && (s[N(s)-1]=='\42'))
      return compile_string (p);

    int i;
    LTP l;
    array<instance> a= env [s];
    for (i=0; i<N(a); i++)
      l= LTP (TP (program (a[i], p->start, p->end), a[i]->t), l);
    if (ok (l, p)) return l;

    error (p, "syntax error");
    return LTP ();
  }

  case COMPOSITE: {
    if (N(p)==0)
      fatal_error ("can't execute empty list", "compiler_rep::compile");

    if (is_func (p, "define", 2)) return compile_define (p, FALSE);
    if (is_func (p, "var", 2)) return compile_define (p, TRUE);
    if (is_func (p, "lambda") && (N(p)>=3)) return compile_lambda (p);
    if (is_func (p, "\'", 1)) return compile_quote (p);
    if (is_func (p, "seq")) return compile_seq (p);
    if (is_func (p, "if", 2)) return compile_if (p);
    if (is_func (p, "if", 3)) return compile_if_else (p);
    
    LTP ltp= compile (p[0]), res, l;
    if (nil (ltp)) return ltp;

    res= LTP ();
    l  = ltp;
    while (!nil (l)) {
      res << compile_composite (l->item, p, 0);
      l= l->next;
    }
    if (ok (res, p)) return res;

    res= LTP ();
    l  = ltp;
    while (!nil (ltp)) {
      res << compile_composite (ltp->item, p, 2);
      ltp= ltp->next;
    }
    if (ok (res, p)) return res;

    error (p, "syntax error or type mismatch");
    return LTP ();
  }

  default:
    fatal_error ("not yet implemented", "compiler_rep::compile");
  }
}

/******************************************************************************
* Definition of variables and constants
******************************************************************************/

LTP
compiler_rep::compile_integer (program p) {
  instance ins (integer_instance (p->s, "Int", as_int (p->s)));
  return LTP (TP (program (ins, p->start, p->end), "Int"));
}

LTP
compiler_rep::compile_string (program p) {
  string s= p->s (1, N(p->s)-1);
  instance ins (string_instance (p->s, "String", s));
  return LTP (TP (program (ins, p->start, p->end), "String"));
}

LTP
compiler_rep::compile_quote (program p) {
  instance ins (program_instance ("quote-expr", "Program", p[1]));
  return LTP (TP (program (ins, p->start, p->end), "Program"));
}

LTP
compiler_rep::compile_define (program p, bool dynamic) {
  string  name;
  type    t;
  program init (p[2]);
  LTP     Init= compile (init);
  program INIT;
  program d;

  if (is_string (p[1])) {
    if (!ok (Init, p)) return LTP ();
    if (N(Init)!=1) {
      error (p, "ambiguous definition");
      return LTP ();
    }
    name= p[1]->s;
    INIT= Init->item->p;
    t   = Init->item->t;
  }
  else {
    if ((N(p[1])!=3) ||
	(!is_string (p[1][0], ":")) ||
	(!is_string (p[1][1])))
      {
	error (p, "incorrect typed definition");
	return LTP ();
      }
    name= p[1][1]->s;
    t   = p[1][2];
    if (!of_type (Init, t, init)) return LTP ();
    INIT= filter (Init, t, init);
  }

  instance ins;
  type r= composite ("&", t);
  if (dynamic) {
    ins= variable_instance (name, r);
    d  = system ("var!");
  }
  else {
    instance aux= variable_instance (name, r);
    if (!match (r, t, 1)) {
      error (p, "missing constructor in definition");
      cerr << "How  : " << r << " -> " << t << "\n";
      return LTP ();
    }
    program act= coerce (aux, r, t);
    ins= define_instance (name, t, act);
    d  = system ("define!");
  }
  env->define_var (ins);

  return LTP (TP (composite (d, ins, INIT, p->start, p->end), "Void"));
}

/******************************************************************************
* compilation of composite expressions
******************************************************************************/

LTP
compiler_rep::compile_composite (TP first, program arg, int max_coerce) {
  program cmd (first->p);
  type t (first->t);
  
  if (!is_func (t, "->", 2)) return LTP ();
  type dom (t[1]);
  type im (t[2]);

  int i;
  array<program> a (N(arg));
  a[0]=cmd;
  if ((N(a)==1) && (!is_string (dom, "Void"))) return LTP ();
  for (i=1; i<N(a); i++) {
    LTP l= compile (arg[i]);
    while (TRUE) {
      if (nil (l)) return LTP ();
      if ((N(a)==2) &&
	  (match (l->item->t, dom, max_coerce))) {
	a[i]= coerce (l->item->p, l->item->t, dom);
	break;
      }
      if (is_func (dom, "Tuple", 1) &&
	  match (l->item->t, dom[1], max_coerce))  {
	a[i]= coerce (l->item->p, l->item->t, dom[1]);
	break;
      }
      if (is_func (dom, "Cross", N(a)-1) &&
	  match (l->item->t, dom[i], max_coerce))  {
	a[i]= coerce (l->item->p, l->item->t, dom[i]);
	break;
      }
      l=l->next;
    }
  }
  return LTP (TP (program (a, arg->start, arg->end), im));
}

/******************************************************************************
* Compilation of lambda expressions
******************************************************************************/

LTP
compiler_rep::compile_lambda (program p) {
  int i;
  program arg= p[1];
  if (!is_composite (arg)) {
    error (arg, "arguments must be a list");
    return LTP ();
  }

  env->level_up ();
  array<program> ARGS (N(arg));
  array<program> TYPES (N(arg)+1);
  TYPES[0]= program ("Cross");

  for (i=0; i<N(arg); i++) {
    if ((N(arg[i])!=3) || (!is_string (arg[i][0], ":"))) {
      error (arg[i], "arguments of lambda should be typed");
      env->level_down ();
      return LTP ();
    }
    program var= arg[i][1];
    type    t  = arg[i][2];
    type    r  = composite ("&", t);
    if (!is_string (var)) {
      error (var, "variable name expected here");
      env->level_down ();
      return LTP ();
    }
    instance ins= variable_instance (var->s, r);
    ARGS[i]     = program (ins, var->start, var->end);
    TYPES[i+1]  = t;
    env->define_var (ins);
  }

  program body= p[2];
  if (N(p)>3) {
    array<program> a (N(p)-1);
    a[0]= program ("seq", p[0]->start, p[0]->end);
    for (i=1; i<N(a); i++) a[i]= p[i+1];
    body= program (a, p[2]->start, p[N(p)-1]->end);
  }

  program ARG (ARGS);
  type    TYPE (TYPES);
  if (N(TYPES)==1) TYPE= type ("Void");
  if (N(TYPES)==2) TYPE= TYPE[1];
  LTP ltp= compile (body);
  env->level_down ();

  LTP ret;
  while (!nil (ltp)) {
    TP tp      = ltp->item;
    program lp = composite (system ("lambda!"), ARG, tp->p, p->start, p->end);
    type lt    = composite ("->", TYPE, tp->t);
    ret        = LTP (TP (lp, lt), ret);
    ltp        = ltp->next;
  }

  return ret;
}

/******************************************************************************
* Control routines
******************************************************************************/

LTP
compiler_rep::compile_seq (program p) {
  int i;
  array<program> a (N(p));
  a[0]= system ("seq!");
  for (i=1; i<N(p)-1; i++)
    a[i]= filter (compile (p[i]), type ("Void"), p[i]);
  if (N(p)==1) return LTP (TP (a, type ("Void")));

  LTP ltp= compile (p[i]);
  LTP ret;
  while (!nil (ltp)) {
    a[i]       = ltp->item->p;
    program rp (copy (a), p->start, p->end);
    type rt    = ltp->item->t;
    ret        = LTP (TP (rp, rt), ret);
    ltp        = ltp->next;
  }
  return ret;
}

LTP
compiler_rep::compile_if (program p) {
  program cond= filter (compile (p[1]), type ("Bool"), p[1]);
  program prg= filter (compile (p[2]), type ("Void"), p[2]);
  array<program> a (N(p));
  a[0]= system ("if!");
  a[1]= cond;
  a[2]= prg;
  return LTP (TP (program (a, p->start, p->end), type ("Void")));
}

LTP
compiler_rep::compile_if_else (program p) {
  program cond= filter (compile (p[1]), type ("Bool"), p[1]);
  LTP prg1= compile (p[2]);
  LTP prg2= compile (p[3]);
  LTP ret;
  while (!nil (prg1)) {
    LTP counter= prg2;
    while (!nil (counter)) {
      if (prg1->item->t == counter->item->t) {
	array<program> a (N(p));
	a[0]= system ("if_else!");
	a[1]= cond;
	a[2]= prg1->item->p;
	a[3]= counter->item->p;
	a   = copy (a);
	ret = LTP (TP (program (a, p->start, p->end), prg1->item->t), ret);
      }
      counter= counter->next;
    }
    prg1= prg1->next;
  }
  return ret;
}

/******************************************************************************
* Type checking and error handling
******************************************************************************/

bool
compiler_rep::ok (list<typed_program> ltp, program p) {
  if (nil (ltp)) return FALSE;
  while (!nil (ltp)) {
    list<typed_program> ltp2= ltp->next;
    while (!nil (ltp2)) {
      if (ltp2->item->t==ltp->item->t) {
	error (p, "ambiguous type");
	return FALSE;
      }
      ltp2= ltp2->next;
    }
    ltp=ltp->next;
  }
  return TRUE;
}

bool
compiler_rep::of_type (list<typed_program> ltp, type t, program p) {
  // ERROR: Should also test coercions
  while (!nil (ltp)) {
    if (ltp->item->t==t) {
      list<typed_program> ltp2= ltp->next;
      while (!nil (ltp2)) {
	if (ltp2->item->t==t) {
	  error (p, "ambiguous expression");
	  return FALSE;
	}
	ltp2= ltp2->next;
      }
      return TRUE;
    }
    ltp=ltp->next;
  }
  cerr << "Error: " << t << " expected\n";
  cerr << "In   : " << p << "\n";
  return FALSE;
}

program
compiler_rep::filter (list<typed_program> ltp, type t, program p) {
  if (!of_type (ltp, t, p)) return program ("compilation error");
  while (!nil (ltp)) {
    if (ltp->item->t==t)
      return ltp->item->p;
    ltp= ltp->next;
  }
  fatal_error ("this situation should not occur", "compiler_rep::filter");
}

bool
compiler_rep::exists (string name, type t) {
  array<instance> a= env[name];
  int i=0;
  for (i=0; i<N(a); i++)
    if (a[i]->t==t) return TRUE;
  return FALSE;
}

instance
compiler_rep::find (string name, type t) {
  array<instance> a= env[name];
  int i=0;
  for (i=0; i<N(a); i++)
    if (a[i]->t==t) return a[i];
  fatal_error ("instance not found", "compiler_rep::find");
}

instance
compiler_rep::system (string name) {
  return find (name, "System");
}

void
compiler_rep::error (program p, string s) {
  cerr << "Error: " << s << "\n";
  cerr << "In   : " << p << "\n";
}

#undef TP
#undef LTP

#endmodule // code_compiler
