
/******************************************************************************
* MODULE     : boxes.gen.cc
* DESCRIPTION: Important routines for all boxes
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <Boxes/boxes.gen.h>
#include <array.gen.cc>

#module code_boxes
#import boxes
#import tree

/*****************************************************************************/
// Default settings for virtual routines
/*****************************************************************************/

int box_rep::subnr () { return 0; }
box box_rep::subbox (int i) { (void) i; return box (); }
path box_rep::left_path () { return path (0); }
path box_rep::right_path () { return path (1); }
double box_rep::left_slope () { return 0.0; }
double box_rep::right_slope () { return 0.0; }
SI box_rep::left_correction () { return (SI) (-y3*left_slope ()); }
SI box_rep::right_correction () { return (SI) (y4*right_slope ()); }
SI box_rep::left_sub_correction (double level) { (void) level; return 0; }
SI box_rep::left_sup_correction (double level) { (void) level; return 0; }
SI box_rep::right_sub_correction (double level) { (void) level; return 0; }
SI box_rep::right_sup_correction (double level) { (void) level; return 0; }
SI box_rep::sub_base () { return y1; }
SI box_rep::sup_base () { return y2; }

/*****************************************************************************/
// Positioning routines
/*****************************************************************************/

bool
outside (SI x, SI delta, SI x1, SI x2) {
  return
    (x<x1) || ((x==x1) && (delta<0)) ||
    (x>x2) || ((x==x2) && (delta>=0));
}

SI
distance (box b, SI x, SI y, SI delta) {
  x-=b->x0;
  y-=b->y0;
  int dx, dy;
  if (x <=  b->x1) dx = b->x1- x- (delta<0? 1:0);
  else if (x >=  b->x2) dx = x- b->x2+ (delta<0? 0:1);
  else dx = 0;
  if (y <  b->y1) dy = b->y1- y;
  else if (y >= b->y2) dy = y- b->y2;
  else dy = 0;
  return dx+dy;
}

SI
get_delta (SI x, SI x1, SI x2) {
  if (x1==x2) return 0;
  if (x==x1) return -1;
  if (x==x2) return 1;
  return 0;
}

/*****************************************************************************/
// Routines concerning the cursor
/*****************************************************************************/

cursor
box_rep::std_cursor (path p) {
  /*
  cursor cu ((p->item<=0 ? x1 : x2), (y1+y2)>>1);
  cu->y1 = (y1-y2)>>1;
  cu->y2 = cu->y1+ (y2-y1);
  return cu;
  */

  bool flag= (p->item==0);
  double slope= flag? left_slope (): right_slope ();
  cursor cu (flag ? x1: x2, 0);
  cu->y1= y1; cu->y2= y2;
  cu->x1= (SI) (cu->y1* slope);
  cu->x2= (SI) (cu->y2* slope);
  return cu;
}

path
box_rep::std_path (SI x, SI y, SI delta) { (void) y;
  SI m= (x1+x2)>>1;
  if ((x<m) || ((x==m) && (delta<0))) return path (0);
  else return path (1);
}

path
box_rep::get_path (SI x, SI y, SI delta) {
  return std_path (x, y, delta);
}

cursor
box_rep::get_cursor (path p) {
  return std_cursor (p);
}

path
box_rep::var_get_path (SI x, SI y, SI delta) {
  return get_path (x, y, delta + get_delta (x, x1, x2));
}

cursor
box_rep::var_get_cursor (path p) {
  cursor cu (get_cursor (p));
  cu->delta -= get_delta (cu->x0, x1, x2);
  return cu;
}

int
reindex (int i, int item, int n) {
  if (item<0) item=0;
  if (item>n) item=n;
  if (i==0) return item;
  if ((i <= (item<<1)) && (i <= ((n-item)<<1))) {
    int d=(i+1)>>1;
    if (((i+1)&1)==0) return item-d;
    else return item+d;
  }
  if (i > (item<<1)) return i;
  return n-i;
}

path box_rep::encode (path p) { return p; }
path box_rep::decode (path p) { return p; }

/*****************************************************************************/
// Redrawing boxes
/*****************************************************************************/

int nr_painted= 0;

void
clear_rectangles (ps_device dev, rectangles l) {
  while (!nil (l)) {
    rectangle r (l->item);
    dev->clear (r->x1- dev->ox, r->y1- dev->oy,
		r->x2- dev->ox, r->y2- dev->oy);
    l= l->next;
  }
}

void
box_rep::redraw (ps_device dev, path p, rectangles& l) {
  if ((nr_painted>=10) && dev->check_event (INPUT_EVENT)) return;
  dev->move_origin (x0, y0);
  SI X1 (x3-dev->pixel), Y1 (y3-dev->pixel);
  SI X2 (x4+dev->pixel), Y2 (y4+dev->pixel);
  if (dev->is_visible (X1, Y1, X2, Y2)) {
    rectangles ll;
    l= rectangles();

    if (!nil (p)) p= decode (p);
    int i, item=-1, n=subnr ();
    bool flag= (nil (p) || nil (p->next));
    if (!flag) item= p->item;
    for (i=0; i<n; i++) {
      int k= reindex (i, item, n-1);
      if (flag || (i!=0)) subbox (k)->redraw (dev, path (), ll);
      else subbox (k)->redraw (dev, p->next, ll);
      while (!nil(ll)) { l= rectangles (ll->item, l); ll= ll->next; }
    }

    if (!dev->check_event (EVENT_STATUS)) {
      dev->outer_round (X1, Y1, X2, Y2);
      l= rectangle (X1+ dev->ox, Y1+ dev->oy, X2+ dev->ox, Y2+ dev->oy);
      display (dev);
      nr_painted++;
    }
  }
  dev->move_origin (-x0, -y0);
}

/*****************************************************************************/
// The cursor class
/*****************************************************************************/

cursor::cursor (SI x, SI y, SI delta, SI x1, SI y1, SI x2, SI y2):
  rep (new cursor_rep)
{
  rep->x0= x ; rep->y0= y ; rep->delta= delta;
  rep->x1= x1; rep->y1= y1; rep->x2= x2; rep->y2= y2;
}

cursor
copy (cursor cu) {
  return cursor (cu->x0, cu->y0, cu->delta, cu->x1, cu->y1, cu->x2, cu->y2);
}

/*****************************************************************************/
// Selections
/*****************************************************************************/

selection::selection (rectangles rs, path start, path end):
  rep (new selection_rep)
{
  rep->rs   = rs;
  rep->start= start;
  rep->end  = end;
}

selection
box_rep::select (path start, path end) {
  if ((subnr ()==0) || atom (start) || atom (end)) {
    if (start==end) return selection (rectangles(), start, end);
  }
  else {
    if (start->item==end->item) {
      selection sel= subbox (start->item)->select (start->next, end->next);
      return selection (translate (sel->rs, x0, y0),
			path (start->item, sel->start),
			path (end->item, sel->end));
    }
  }
  return selection (rectangle (x0+x1, y0+y1, x0+x2, y0+y2),
		    left_path (), right_path ());
}

/*****************************************************************************/
// Miscellaneous routines
/*****************************************************************************/

bool box::operator == (box b2) { return rep==b2.rep; }
bool box::operator != (box b2) { return rep!=b2.rep; }

#import code_array (box)

box::operator tree () { return tree (*rep); }
ostream& operator << (ostream& out, box b) { return out << ((tree) b); }

#endmodule // code_boxes
