
/******************************************************************************
* MODULE     : concat.gen.cc
* DESCRIPTION: Concatenations of arrays of boxes
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <Boxes/composite.gen.h>

#module code_concat_boxes
#import composite_boxes

/*****************************************************************************/
// The concat_box representation
/*****************************************************************************/

struct concat_box_rep: public composite_box_rep {
  concat_box_rep (array<box> bs, array<SI> spc);
  operator tree ();

  void      position (array<SI> spc);
  int       index_left (int i);
  int       count_left (int i);
  int       count_right (int i);
  double    left_slope ();
  double    right_slope ();

  virtual cursor    get_cursor (path p);
  virtual path      get_path (SI x, SI y, SI delta);
  virtual path      left_path ();
  virtual path      right_path ();
  virtual selection select (path start, path end);
};

concat_box_rep::operator tree () {
  int i, n= N(bs);
  tree t ("concat", n);
  for (i=0; i<n; i++) t[i]= (tree) bs[i];
  return t;
}

/*****************************************************************************/
// Concatenation boxes
/*****************************************************************************/

void
concat_box_rep::position (array<SI> spc) {
  int i;
  if (N(bs)==0)
    fatal_error ("concat of zero boxes", "concat_box_rep::position");
  x1 = bs[0]->x1;
  x2 = 0;
  for (i=0; i<N(bs); i++) {
    x2 += spc[i];
    bs[i]->x0= x2;
    bs[i]->y0= 0;
    x2 += bs[i]->x2;
  }
  composite_box_rep::position ();
}

concat_box_rep::concat_box_rep (array<box> bs2, array<SI> spc) {
  bs = bs2;
  position (spc);
}

int
concat_box_rep::index_left (int i) {
  while ((i>0) && (bs[i-1]->x0+ bs[i-1]->x1== bs[i]->x0+ bs[i]->x1)) i--;
  return i;
}

cursor
concat_box_rep::get_cursor (path p) {
  if (nil (p->next)) return std_cursor (p);
  else {
    int i=p->item;
    cursor cu (bs[i]->var_get_cursor (p->next));
    cu->x0 += bs[i]->x0;
    cu->y0 += bs[i]->y0;
    if (cu->x0 == bs[N(bs)-1]->x0+ bs[N(bs)-1]->x2) cu->delta -= N(bs)-1 - i;
    else if (cu->x0 == bs[i]->x0+ bs[i]->x1) cu->x0 += i- index_left (i);
    return cu;
  }
}

int
concat_box_rep::count_left (int i) {
  int n;
  for (n=1; (i+n<N(bs)) && (i+n>0); n++)
    if ((bs[i+n]->x0+ bs[i+n]->x1) > (bs[i]->x0+ bs[i]->x1)) return n;
  return n;
}

int
concat_box_rep::count_right (int i) {
  int n;
  for (n=1; (i-n<N(bs)) && (i-n>0); n++)
    if ((bs[i]->x0+ bs[i]->x2) > (bs[i-n]->x0+ bs[i-n]->x2)) return n;
  return n;
}

path
concat_box_rep::get_path (SI x, SI y, SI delta) {
  int i, n;
  bool flag;

  SI x_1= bs[N(bs)-1]->x0+ bs[N(bs)-1]->x2;
  if (x >= x_1) {
    n= count_right (N(bs)-1);
    if ((delta >= 0) || (x > x_1)) i=N(bs)-1; else i=N(bs)+delta;
    if ((N(bs)-1)-i<n) {
      int d= ((i==N(bs)-1)? delta: 0);
      path p= bs[i]->var_get_path (x- bs[i]->x0, y- bs[i]->y0, d);
      return path (i, p);
    }
    i=N(bs)-n;
    path p= bs[i]->var_get_path (x- bs[i]->x0, y- bs[i]->y0, delta+ (n-1));
    return path (i, p);
  }

  for (i=0; i<N(bs); i++) {
    SI x_i= bs[i]->x0+ bs[i]->x1;
    if ((x < x_i) || ((x == x_i) && (delta < 0))) {
      path p= bs[i]->var_get_path (x- bs[i]->x0, y- bs[i]->y0, delta);
      return path (i, p);
    }

    n= count_left (i);
    if (x == bs[i]->x0+ bs[i]->x1) {
      /*
      if ((i>0) && (bs[i-1]->x0+ bs[i-1]->x2 < x)) {
	path p= bs[i]->var_get_path (x- bs[i]->x0, y- bs[i]->y0, delta);
	return path (i, p);
      }
      */
      if (delta<0) { // i can not be zero here
	i--;
	path p= bs[i]->var_get_path (x- bs[i]->x0, y- bs[i]->y0, delta);
	return path (i, p);
      }
      if (delta<n) {
	i += delta;
	path p= bs[i]->var_get_path (x- bs[i]->x0, y- bs[i]->y0, 0);
	return path (i, p);
      }
      i += n-1;
      path p= bs[i]->var_get_path (x- bs[i]->x0, y- bs[i]->y0, delta- (n-1));
      return path (i, p);
    }

    i += n-1;
    if (i+1==N(bs)) flag=TRUE;
    else {
      box sb1= bs[i];
      box sb2= bs[i+1];
      int ex1= x-(sb1->x0+ sb1->x2);
      int ex2= (sb2->x0+ sb2->x1)-x;
      flag= ((ex2>ex1+1) || ((ex2>=ex1) && (delta<0)));
    }
    if (flag) {
      path p= bs[i]->var_get_path (x- bs[i]->x0, y- bs[i]->y0, delta);
      return path (i, p);
    }
  }

  return std_path (x, y, delta);
}

double
concat_box_rep::left_slope () {
  if (N(bs)==0) return 0.0;
  return bs[0]->left_slope ();
}

double
concat_box_rep::right_slope () {
  if (N(bs)==0) return 0.0;
  return bs[N(bs)-1]->right_slope ();
}

path
concat_box_rep::left_path () {
  if (N(bs)==0) return path (0);
  return path (0, bs[0]->left_path());
}

path
concat_box_rep::right_path () {
  if (N(bs)==0) return path (1);
  return path (N(bs)-1, bs[N(bs)-1]->right_path());
}

selection
concat_box_rep::select (path start, path end) {
  if (N(bs)==0) return selection (rectangles (), start, end);
  if (atom (start) && (start == end))
    return selection (rectangles (), start, end);
  if ((!atom (start)) && (!atom (end)) && (start->item==end->item)) {
    selection sel= subbox (start->item)->select (start->next, end->next);
    return selection (translate (sel->rs, x0, y0),
		      path (start->item, sel->start),
		      path (end->item, sel->end));
  }
  else {
    int  i;
    int  i1 = atom (start)? 0: start->item;
    int  i2 = atom (end)? N(bs)-1: end->item;
    box  sb1= bs[i1];
    box  sb2= bs[i2];
    path sp1= atom (start)? sb1->left_path (): start->next;
    path sp2= atom (end)? sb2->right_path (): end->next;
    selection sel1= sb1->select (sp1, sb1->right_path ());
    selection sel2= sb2->select (sb2->left_path (), sp2);
    rectangles rs = sel1->rs; rs << sel2->rs;
    for (i=i1+1; i<i2; i++) {
      box& sb (bs[i]);
      selection sel= sb->select (sb->left_path (), sb->right_path ());
      rs << sel->rs;
    }
    path p1 (i1, sel1->start), p2 (i2, sel2->end);
    if (nil (rs)) return selection (rectangles (), p1, p2);
    rectangle r  = least_upper_bound (rs);
    return selection (rectangle (x0+r->x1, y0+y1, x0+r->x2, y0+y2), p1, p2);
  }
}

/*****************************************************************************/
// box construction routines
/*****************************************************************************/

box
concat_box (array<box> bs, array<SI> spc) {
  return new concat_box_rep (bs, spc);
}

#endmodule // code_concat_boxes
