
/******************************************************************************
* MODULE     : math.gen.cc
* DESCRIPTION: Boxes for mathematical typesetting:
*              fractions, roots, negations, limits and scripts
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <Boxes/composite.gen.h>

#module code_math_boxes
#import composite_boxes

/******************************************************************************
* Miscellaneous routines
******************************************************************************/

void
italic_correct (box b, bool left_flag, bool right_flag) {
  if (left_flag)  b->x1 -= b->left_correction ();
  if (right_flag) b->x2 += b->right_correction ();
}

void
italic_restore (box b, bool left_flag, bool right_flag) {
  if (left_flag)  b->x1 += b->left_correction ();
  if (right_flag) b->x2 -= b->right_correction ();
}

SI
italic_correction (box L, box R, SI dy) {
  double slope_L= L->right_slope ();
  double slope_R= R->left_slope ();

  if (slope_L == slope_R) {
    if (dy==0) return 0;
    return (SI) (slope_R * ((double) dy));
  }

  if (slope_L < slope_R) {
    if (L->y3 >= R->y3+ dy) {
      SI dx_A= (SI) (slope_L * ((double) L->y3));
      SI dx_B= (SI) (slope_R * ((double) (L->y3- dy)));
      return dx_A- dx_B;
    }
    else {
      SI dx_A= (SI) (slope_L * ((double) (R->y3+ dy)));
      SI dx_B= (SI) (slope_R * ((double) R->y3));
      return dx_A- dx_B;
    }
  }
  else {
    if (L->y4 <= R->y4+ dy) {
      SI dx_A= (SI) (slope_L * ((double) L->y4));
      SI dx_B= (SI) (slope_R * ((double) (L->y4- dy)));
      return dx_A- dx_B;
    }
    else {
      SI dx_A= (SI) (slope_L * ((double) (R->y4+ dy)));
      SI dx_B= (SI) (slope_R * ((double) R->y4));
      return dx_A- dx_B;
    }
  }
}

/******************************************************************************
* Fractions
******************************************************************************/

struct frac_box_rep: public composite_box_rep {
  frac_box_rep (box b1, box b2, SI bar_w, SI bar_sep, SI bar_y, color c);
  operator tree () { return tree ("frac", bs[0], bs[1]); }
  path get_path (SI x, SI y, SI delta);
};

frac_box_rep::frac_box_rep (box b1, box b2, SI bar_w,
			    SI bar_sep, SI bar_y, color c) {
  italic_correct (b1, TRUE, FALSE);
  italic_correct (b2, TRUE, FALSE);

  x2   = max (b1->w(), b2->w());
  insert (b1, (x2>>1)- (b1->w()>>1)- b1->x1,
	      bar_y+ bar_sep+ (bar_w>>1)- b1->y1);
  insert (b2, (x2>>1)- (b2->w()>>1)- b2->x1,
	      bar_y- bar_sep- (bar_w>>1)- b2->y2);
  if (x2<=PIXEL) insert (line_box (0, 0, PIXEL, 0, bar_w, c), 0, bar_y);
  else insert (line_box (0, 0, x2, 0, bar_w, c), 0, bar_y);
  position ();
  x1= 0;
  y1= bar_y- b2->h()- (bar_w>>1)- bar_sep;
  x2= max (max(b1->w(), b2->w()), PIXEL);
  y2= bar_y+ b1->h()+ (bar_w>>1)+ bar_sep;

  italic_restore (b1, TRUE, FALSE);
  italic_restore (b2, TRUE, FALSE);
}

path
frac_box_rep::get_path (SI x, SI y, SI delta) {
  if (outside (x, delta, x1, x2)) return std_path (x, y, delta);
  else {
    int i ((y>= bs[2]->y0)?0:1);
    path p= bs[i]->var_get_path (x- bs[i]->x0, y- bs[i]->y0, delta);
    return path (i, p);
  }
}

/******************************************************************************
* Square roots
******************************************************************************/

struct sqrt_box_rep: public composite_box_rep {
  sqrt_box_rep (box b1, box b2, box sqrtb, SI wsqrt, color c);
  operator tree () { return tree ("sqrt", bs[0]); }
  path get_path (SI x, SI y, SI delta);
};

sqrt_box_rep::sqrt_box_rep (box b1, box b2, box sqrtb, SI wsqrt, color c) {
  insert (b1, 0, 0);
  if (!nil (b2)) {
    SI X= - sqrtb->w();
    SI M= X / 3;
    SI Y= sqrtb->y1;
    if (sqrtb->h() < 3*sqrtb->w()) Y += sqrtb->h()>>1;
    else Y += (sqrtb->w()*3)>>1;
    insert (b2, min (X, M- b2->x2), Y- b2->y1);
  }
  insert (sqrtb, -sqrtb->x2, 0);
  insert (line_box (-wsqrt>>1, sqrtb->y2, b1->x2, sqrtb->y2, wsqrt, c), 0, 0);
  
  position ();
  left_justify ();
  y1 -= wsqrt;
  y2 += wsqrt;
}

path
sqrt_box_rep::get_path (SI x, SI y, SI delta) {
  if (outside (x, delta, x1, x2)) return std_path (x, y, delta);
  else {
    int i (( (x>bs[0]->x0) || ((x>=bs[0]->x0) && (delta>=0)) ) ? 0:1 );
    if (N(bs)==3) i=0;
    path p= bs[i]->var_get_path (x- bs[i]->x0, y- bs[i]->y0, delta);
    return path (i, p);
  }
}

/******************************************************************************
* Negation boxes
******************************************************************************/

struct neg_box_rep: public composite_box_rep {
  neg_box_rep (box b1, SI wneg, SI delta, color c);
  operator tree () { return tree ("neg", bs[0]); }
  path get_path (SI x, SI y, SI delta);
};

neg_box_rep::neg_box_rep (box b, SI wneg, SI delta, color c) {
  SI X= (b->x3 + b->x4- PIXEL) >> 1;
  SI Y= (b->y3 + b->y4- PIXEL) >> 1;
  SI DX, DY;

  insert (b, 0, 0);
  if ((3*(b->x4-b->x3)) > (2*(b->y4-b->y3))) {
    DY= delta + ((b->y4 - b->y3)>>1);
    DX= DY>>1;
  }
  else {
    DX= delta + ((b->x4 - b->x3)>>1);
    DY= DX;
  }
  insert (line_box (X+DX, Y+DY, X-DX, Y-DY, wneg, c), 0, 0);
  position ();
}

path
neg_box_rep::get_path (SI x, SI y, SI delta) {
  if (outside (x, delta, x1, x2)) return std_path (x, y, delta);
  else {
    path p= bs[0]->var_get_path (x- bs[0]->x0, y- bs[0]->y0, delta);
    return path (0, p);
  }
}

/******************************************************************************
* Tree boxes
******************************************************************************/

struct tree_box_rep: public composite_box_rep {
  SI  border;
  tree_box_rep (array<box> bs, SI hsep, SI vsep, SI line_w, color line_c);
  operator tree () { return tree ("tree box"); }
  path get_path (SI x, SI y, SI delta);
};

tree_box_rep::tree_box_rep(
  array<box> bs, SI hsep, SI vsep, SI line_w, color line_c)
{
  int i, n= N(bs), w= 0, h= MIN_SI, x, x0, up;
  for (i=1; i<n; i++) w += bs[i]->w();
  for (i=1; i<n; i++) h = max (h, bs[i]->y2);
  w += (n-2)*hsep;
  x0= 0; if (bs[0]->w()>w) { x0= (bs[0]->w()-w)>>1; w= bs[0]->w(); }
  up= bs[0]->y1- vsep;

  insert (bs[0], (w>>1)- ((bs[0]->x1+bs[0]->x2)>>1), 0);
  for (x=x0, i=1; i<n; i++) {
    SI xi= x- bs[i]->x1;
    SI yi= up- h;
    insert (bs[i], xi, yi);
    x += bs[i]->w()+ hsep;
  }

  for (x=x0, i=1; i<n; i++) {
    SI xi= x+ (bs[i]->w()>>1);
    SI yi= up+ bs[i]->y2- h;
    insert (line_box (w>>1, bs[0]->y1, xi, yi, line_w, line_c), 0, 0);
    x += bs[i]->w()+ hsep;
  }

  position ();
  border= up+ (vsep>>1);
}

path
tree_box_rep::get_path (SI x, SI y, SI delta) {
  if (outside (x, delta, x1, x2)) return std_path (x, y, delta);
  else {
    int i=0;
    if (y < border) {
      int j, d=MAX_SI;
      for (j=1; j<((N(bs)+1)>>1); j++)
	if (distance (bs[j], x, y, delta)< d) {
	  d= distance (bs[j], x, y, delta);
	  i= j;
	}
    }
    path p= bs[i]->var_get_path (x- bs[i]->x0, y- bs[i]->y0, delta);
    return path (i, p);
  }
}

/******************************************************************************
* limits
******************************************************************************/

struct lim_box_rep: public composite_box_rep {
  box ref;
  lim_box_rep (box ref, box lo, box hi);
  operator tree () { return tree ("lim", bs[0]); }
};

lim_box_rep::lim_box_rep (box ref2, box lo, box hi): ref (ref2) {
  SI X, Y;
  insert (ref, 0, 0);
  if (!nil (lo)) {
    Y= ref->y1;
    X= ((SI) (ref->right_slope ()* (Y+lo->y2-lo->y1)))+
       ((ref->x1 + ref->x2)>>1);
    insert (lo, X- ((lo->x1 + lo->x2)>>1), Y-lo->y2);
  }
  if (!nil (hi)) {
    Y= ref->y2;
    X= ((SI) (ref->right_slope ()*(Y+hi->y2-hi->y1)))+
       ((ref->x1 + ref->x2)>>1);
    insert (hi, X- ((hi->x1 + hi->x2)>>1), Y-hi->y1);
  }
  position ();
  left_justify ();
}

/******************************************************************************
* wide hats, tildas, etc...
******************************************************************************/

struct wide_box_rep: public composite_box_rep {
  box ref;
  wide_box_rep (box ref, box hi, SI sep);
  operator tree () { return tree ("wide", bs[0]); }
  path   get_path (SI x, SI y, SI delta);
  double left_slope ();
  double right_slope ();
};

wide_box_rep::wide_box_rep (box ref2, box hi, SI default_y): ref (ref2) {
  SI X, Y;
  insert (ref, 0, 0);
  Y= ref->y4;
  X= ((SI) (ref->right_slope () * Y)) + ((ref->x1 + ref->x2)>>1);
  insert (hi, X- ((hi->x1 + hi->x2)>>1), Y- default_y);
  position ();
  left_justify ();
  x1= ref->x1;
  x2= ref->x2;
}

path
wide_box_rep::get_path (SI x, SI y, SI delta) {
  if (outside (x, delta, x1, x2)) return std_path (x, y, delta);
  else {
    path p= bs[0]->var_get_path (x- bs[0]->x0, y- bs[0]->y0, delta);
    return path (0, p);
  }
}

double
wide_box_rep::left_slope () {
  return ref->left_slope ();
}

double
wide_box_rep::right_slope () {
  return ref->right_slope ();
}

/******************************************************************************
* dummy script and prime boxes
******************************************************************************/

struct dummy_script_box_rep: public composite_box_rep {
  dummy_script_box_rep (box b1, box b2);
  operator tree () { return tree ("dummy script"); }
};

dummy_script_box_rep::dummy_script_box_rep (box b1, box b2) {
  SI mid, mid1, mid2;
  if (nil (b1)) {
    if (nil (b2)) mid= 0;
    else mid= (b2->y1+ b2->y2) >> 1;
  }
  else {
    if (nil (b2)) mid= (b1->y1+ b1->y2) >> 1;
    else {
      mid1= (b1->y1+ b1->y2) >> 1;
      mid2= (b2->y1+ b2->y2) >> 1;
      mid = (mid1+ mid2) >> 1;
    }
  }
  if (!nil (b1)) insert (b1, 0, mid- b1->y2);
  if (!nil (b2)) insert (b2, 0, mid- b2->y1);
  position ();
  left_justify ();
  if (nil (b1)) y1 -= b2->h();
  if (nil (b2)) y2 += b1->h();
}

struct dummy_prime_box_rep: public composite_box_rep {
  dummy_prime_box_rep (box b);
  operator tree () { return tree ("dummy prime"); }
  cursor get_cursor (path p);
  path get_path (SI x, SI y, SI delta);
};

dummy_prime_box_rep::dummy_prime_box_rep (box b) {
  SI mid= (b->y1+ b->y2) >> 1;
  insert (b, 0, mid- b->y1);
  position ();
  left_justify ();
  y1 -= b->h();
}

cursor
dummy_prime_box_rep::get_cursor (path p) {
  return composite_box_rep::get_cursor (path (0, p));
}

path
dummy_prime_box_rep::get_path (SI x, SI y, SI delta) {
  path p= composite_box_rep::get_path (x, y, delta);
  if (p == path (0)) return bs[0]->left_path ();
  if (p == path (1)) return bs[0]->right_path ();
  return p->next;
}

/******************************************************************************
* side boxes
******************************************************************************/

struct side_box_rep: public composite_box_rep {
  int nr_left;
  int nr_right;
  side_box_rep (box ref, box l1, box l2, box r1, box r2, double level);
  operator tree () { return tree ("side", bs[0]); }
  path get_path (SI x, SI y, SI delta);
};

side_box_rep::side_box_rep (box ref, box l1, box l2,
			    box r1, box r2, double level) {
  insert (ref, 0, 0);
  SI base= ref->sub_base()+ ((SI) ((ref->sup_base()-ref->sub_base())*level));

  nr_left= 0;
  if (!nil (l1)) nr_left++;
  if (!nil (l2)) nr_left++;
  nr_right= 0;
  if (!nil (r1)) nr_right++;
  if (!nil (r2)) nr_right++;

  if (!nil (l1)) {
    SI dy= min (ref->sub_base()- l1->sub_base(), base- l1->y2);
    SI dx= ::italic_correction (l1, ref, -dy);
    dx += l1->right_sup_correction (1.0-level);
    dx += ref->left_sub_correction (level);
    insert (l1, -l1->x2- dx, dy);
  }
  if (!nil (l2)) {
    SI dy= max (ref->sup_base()- l2->sup_base(), base- l2->y1);
    SI dx= ::italic_correction (l2, ref, -dy);
    dx += l2->right_sub_correction (1.0-level);
    dx += ref->left_sup_correction (level);
    insert (l2, -l2->x2- dx, dy);
  }

  if (!nil (r1)) {
    SI dy= min (ref->sub_base()- r1->sub_base(), base- r1->y2);
    SI dx= ::italic_correction (ref, r1, dy);
    dx += ref->right_sub_correction (level);
    dx += r1->left_sup_correction (1.0-level);
    insert (r1, ref->x2+ dx, dy);
  }
  if (!nil (r2)) {
    SI dy= max (ref->sup_base()- r2->sup_base(), base- r2->y1);
    SI dx= ::italic_correction (ref, r2, dy);
    dx += ref->right_sup_correction (level);
    dx += r2->left_sub_correction (1.0-level);
    insert (r2, ref->x2+ dx, dy);
  }

  position ();
  left_justify ();
}

path
side_box_rep::get_path (SI x, SI y, SI delta) {
  if (outside (x, delta, x1, x2))
    return std_path (x, y, delta);
  
  int i= 0;
  int k= nr_left+ 1;
  SI  xx, yy;

  if (nr_left>0) {
    if (nr_left == 1) xx= bs[1]->x0+ bs[1]->x2;
    else xx= max (bs[1]->x0+ bs[1]->x2, bs[2]->x0+ bs[2]->x2);
    if ((x<xx) || ((x==xx) && (delta<0))) {
      i= 1;
      if (nr_left == 2) {
	yy= (bs[1]->y0+ bs[1]->y2+ bs[2]->y0+ bs[2]->y1) >> 1;
	if (y >= yy) i= 2;
      }
    }
  }

  if (nr_right>0) {
    if (nr_right == 1) xx= bs[k]->x0+ bs[k]->x1;
    else xx= min (bs[k]->x0+ bs[k]->x1, bs[k+1]->x0+ bs[k+1]->x1);
    if ((x>=xx) || ((x==xx) && (delta>=0))) {
      i= k;
      if (nr_right == 2) {
	yy= (bs[k]->y0+ bs[k]->y2+ bs[k+1]->y0+ bs[k+1]->y1) >> 1;
	if (y >= yy) i= k+1;
      }
    }
  }
      
  return path (i, bs[i]->var_get_path (x- bs[i]->x0, y- bs[i]->y0, delta));
}

/******************************************************************************
* macro expansions
******************************************************************************/

struct macro_box_rep: public composite_box_rep {
  macro_box_rep (box b);
  operator tree () { return tree ("macro", (tree) bs[0]); }
  cursor get_cursor (path p);
  path get_path (SI x, SI y, SI delta);
  path left_path ();
  path right_path ();
};

macro_box_rep::macro_box_rep (box b) { insert (b, 0, 0); position (); }
cursor macro_box_rep::get_cursor (path p) { return std_cursor (p); }
path macro_box_rep::get_path (SI x, SI y, SI delta) {
  return std_path (x, y, delta); }
path macro_box_rep::left_path () { return path (0); }
path macro_box_rep::right_path () { return path (1); }

/******************************************************************************
* changing size parameters in boxes
******************************************************************************/

struct change_box_rep: public composite_box_rep {
  bool child_flag;
  change_box_rep (bool child_flag);
  operator tree () { return tree ("change", (tree) bs[0]); }
  cursor get_cursor (path p);
  path get_path (SI x, SI y, SI delta);
  path left_path ();
  path right_path ();
};

change_box_rep::change_box_rep (bool flag): child_flag (flag) {}

cursor
change_box_rep::get_cursor (path p) {
  if (child_flag) return composite_box_rep::get_cursor (p);
  return composite_box_rep::get_cursor (path (0, p));
}

path
change_box_rep::get_path (SI x, SI y, SI delta) {
  if (child_flag) return composite_box_rep::get_path (x, y, delta);
  path p= composite_box_rep::get_path (x, y, delta);
  if (p == path (0)) return bs[0]->left_path ();
  if (p == path (1)) return bs[0]->right_path ();
  return p->next;
}

path
change_box_rep::left_path () {
  if (child_flag) return composite_box_rep::left_path ();
  return bs[0]->left_path ();
}

path
change_box_rep::right_path () {
  if (child_flag) return composite_box_rep::right_path ();
  return bs[0]->right_path ();
}

/******************************************************************************
* moving and resizing boxes
******************************************************************************/

struct move_box_rep: public change_box_rep {
  move_box_rep (box b, SI x, SI y, bool child_flag);
  operator tree () { return tree ("move", (tree) bs[0]); }
};

move_box_rep::move_box_rep (box b, SI x, SI y, bool child_flag):
  change_box_rep (child_flag)
{
  insert (b, x, y);
  position ();
}

struct resize_box_rep: public change_box_rep {
  resize_box_rep (box b, SI x1, SI y1, SI x2, SI y2, bool child_flag);
  operator tree () { return tree ("resize", (tree) bs[0]); }
};

resize_box_rep::resize_box_rep (box b, SI X1, SI Y1, SI X2, SI Y2, bool F):
  change_box_rep (F)
{
  insert (b, 0, 0);
  position ();
  x1= X1; y1= Y1;
  x2= X2; y2= Y2;
}

/******************************************************************************
* box construction routines
******************************************************************************/

box
frac_box (box b1, box b2, SI bar_w, SI bar_sep, SI bar_y, color c) {
  return new frac_box_rep (b1, b2, bar_w, bar_sep, bar_y, c);
}

box
sqrt_box (box b1, box b2, box sqrtb, SI wsqrt, color c) {
  return new sqrt_box_rep (b1, b2, sqrtb, wsqrt, c);
}

box
neg_box (box b, SI wneg, SI delta, color c) {
  return new neg_box_rep (b, wneg, delta, c);
}

box
tree_box (array<box> bs, SI hsep, SI vsep, SI line_w, color line_c) {
  return new tree_box_rep (bs, hsep, vsep, line_w, line_c);
}

box
limit_box (box ref, box lo, box hi) {
  return new lim_box_rep (ref, lo, hi);
}

box
wide_box (box ref, box hi, SI default_y) {
  return new wide_box_rep (ref, hi, default_y);
}

box
script_box (box b1, box b2) {
  return new dummy_script_box_rep (b1, b2);
}

box
prime_box (box b) {
  return new dummy_prime_box_rep (b);
}

box
left_script_box (box ref, box b1, box b2, double level) {
  return new side_box_rep (ref, b1, b2, box (), box (), level);
}

box
right_script_box (box ref, box b1, box b2, double level) {
  return new side_box_rep (ref, box (), box (), b1, b2, level);
}

box
side_box (box ref, box l1, box l2, box r1, box r2, double level) {
  return new side_box_rep (ref, l1, l2, r1, r2, level);
}

box
macro_box (box b) {
  return new macro_box_rep (b);
}

box
move_box (box b, SI x, SI y, bool child_flag) {
  return new move_box_rep (b, x, y, child_flag);
}

box
resize_box (box b, SI x1, SI y1, SI x2, SI y2, bool child_flag) {
  return new resize_box_rep (b, x1, y1, x2, y2, child_flag);
}

#endmodule // code_math_boxes
