
/******************************************************************************
* MODULE     : matrix.gen.cc
* DESCRIPTION: Matrix boxes.
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <Boxes/boxes.gen.h>
#include <Boxes/construct.gen.h>

#module code_matrix_boxes
#import boxes
#import construct_boxes

/******************************************************************************
* Matrix boxes
******************************************************************************/

struct mb_type {
  bool ro;     // TRUE => read only field
  bool hflag;  // TRUE if positioned horizontally by hpos
  double hpos; // 0 for left, 0.5 for center, 1 for right
  bool vflag;  // TRUE if positioned vertically by hpos
  double vpos; // 0 for bottom, 0.5 for center, 1 for top
  bool filled; // fill box or not
  color col;   // fill color
};

struct matrix_box_rep: public box_rep {
  int nr;
  box* bs;
  int* bx;     // left column of baby box
  int* by;     // upper row of baby box
  int* bw;     // width in columns
  int* bh;     // height in rows
  mb_type* bt; // type of box

  int nrx;     // number of columns
  int nry;     // number of rows
  SI* cx1;     // left x coordinate of each column
  SI* cx2;     // right x coordinate of each comumn
  SI* ry1;     // bottom y coordinate of each row
  SI* ry2;     // top y coordinate of each row

  matrix_box_rep (int nrx2, int nry2);
  ~matrix_box_rep ();
  operator tree () { return tree ("matrix"); }

  void   insert (box b, int x, int y, int w, int h);
  void   format ();
  void   ready ();
  void   display (ps_device dev);
  int    subnr ();
  box    subbox (int i);
  path   get_path (SI x, SI y, SI delta);
  cursor get_cursor (path p);
  SI     sub_base ();
  SI     sup_base ();
};

matrix_box_rep::matrix_box_rep (int nrx2, int nry2) {
  int i;
  int max=nrx2*nry2;
  bs = new box[max];
  bx = new int[max];
  by = new int[max];
  bw = new int[max];
  bh = new int[max];
  bt = new mb_type[max];
  nr = 0;

  nrx= nrx2;
  nry= nry2;
  cx1= new SI[nrx];
  cx2= new SI[nrx];
  ry1= new SI[nry];
  ry2= new SI[nry];
  for (i=0; i<nrx; i++) { cx1[i]=0; cx2[i]=0; }
  for (i=0; i<nry; i++) { ry1[i]=0; ry2[i]=0; }
}

matrix_box_rep::~matrix_box_rep () {
  delete[] bs;
  delete[] bx;
  delete[] by;
  delete[] bw;
  delete[] bh;
  delete[] bt;
  delete[] cx1;
  delete[] cx2;
  delete[] ry1;
  delete[] ry2;
}

int
matrix_box_rep::subnr () {
  return nr;
}

box
matrix_box_rep::subbox (int i) {
  return bs[i];
}

SI
compute_size (SI* a1, SI* a2, int start, int n) {
  int i;
  SI sum=0;
  for (i=start; i<start+n; i++) sum+=a2[i]-a1[i];
  return sum;
}

void
matrix_box_rep::insert (box b, int x, int y, int w, int h) {
  if (nr==nrx*nry) fatal_error ("Bad matrix", "matrix_box_rep::insert");
  bs[nr]= b;
  bx[nr]= x;
  by[nr]= y;
  bw[nr]= w;
  bh[nr]= h;

  bt[nr].ro    = FALSE;
  bt[nr].hflag = TRUE;
  bt[nr].hpos  = 0.5;
  bt[nr].vflag = FALSE;
  bt[nr].vpos  = 0.5;
  bt[nr].filled= FALSE;

  if (w!=1) bt[nr].hflag = TRUE;
  if (h!=1) bt[nr].vflag = TRUE;

  nr++;
}

void
matrix_box_rep::format () {
  int k, size;
  for (size=1; size<=nr; size++)
    for (k=0; k<nr; k++)
      if (bw[k]*bh[k]==size) {
	box& b (bs[k]);
	int& x (bx[k]);
	int& y (by[k]);
	int& w (bw[k]);
	int& h (bh[k]);

	if (bt[k].hflag) {
	  SI w1= b->w ();
	  SI w2= compute_size (cx1, cx2, x, w);
	  if (w2<w1) {
	    int i;
	    SI delta= (w1-w2)/(2*w);
	    for (i=x; i<x+w; i++) {
	      cx1[i]-=delta;
	      cx2[i]+=delta;
	    }
	  }
	}
	else {
	  cx1[x]= min (cx1[x], b->x1);
	  cx2[x]= max (cx2[x], b->x2);
	}
	
	if (bt[k].vflag) {
	  SI h1= b->h ();
	  SI h2= compute_size (ry1, ry2, y, h);
	  if (h2<h1) {
	    int i;
	    SI delta= (h1-h2)/(2*h);
	    for (i=y; i<y+h; i++) {
	      ry1[i]-=delta;
	      ry2[i]+=delta;
	    }
	  }
	}
	else {
	  ry1[y]= min (ry1[y], b->y1);
	  ry2[y]= max (ry2[y], b->y2);
	}
      }
}

void
matrix_box_rep::ready () {
  int i;
  x1= x3= 0;
  x2= x4= compute_size (cx1, cx2, 0, nrx);
  y1= y2= 0;
  y2= y4= compute_size (ry1, ry2, 0, nry);
  for (i=0; i<nr; i++) {
    box& b(bs[i]);

    if (bt[i].hflag) {
      SI bx1= compute_size (cx1, cx2, 0, bx[i]);
      SI bx2= compute_size (cx1, cx2, 0, bx[i]+bw[i])-b->w();
      b->x0= ((SI) (((double) bx1)*(1-bt[i].hpos)+
		    ((double) bx2)*bt[i].hpos))  - b->x1;
    }
    else {
      b->x0= compute_size (cx1, cx2, 0, bx[i])- cx1[bx[i]];
    }
    
    if (bt[i].vflag) {
      SI by1= y2- compute_size (ry1, ry2, 0, by[i]+bh[i]);
      SI by2= y2- compute_size (ry1, ry2, 0, by[i])-b->h();
      b->y0= ((SI) (((double) by1)*(1-bt[i].vpos)+
		    ((double) by2)*bt[i].vpos))  - b->y1;
    }
    else {
      b->y0= y2- compute_size (ry1, ry2, 0, by[i]+bh[i])- ry1[by[i]];
    }
    
    x3= min (x3, b->x0+ b->x3);
    x4= max (x4, b->x0+ b->x4);
    y3= min (y3, b->y0+ b->y3);
    y4= max (y4, b->y0+ b->y4);
  }
}

void
matrix_box_rep::display (ps_device dev) {
  int i;
  for (i=0; i<nr; i++) {
    if (bt[i].filled) {
      SI bx1= compute_size (cx1, cx2, 0, bx[i]);
      SI bx2= compute_size (cx1, cx2, 0, bx[i]+bw[i]);
      SI by1= y2- compute_size (ry1, ry2, 0, by[i]+bh[i]);
      SI by2= y2- compute_size (ry1, ry2, 0, by[i]);
      dev->set_color (bt[i].col);
      dev->set_line_style (PIXEL);
      dev->fill (bx1, by1, bx2, by2);
    }
    // bs[i]->redraw (dev);
  }
}

path
matrix_box_rep::get_path (SI x, SI y, SI delta) {
  int I, J;
  if (outside (x, delta, x1, x2)) return std_path (x, y, delta);
  int xx= x1;
  for (I=0; I<nrx-1; I++) {
    xx += cx2[I]-cx1[I];
    if ((x<xx) || ((x==xx) && (delta<0))) break;
  }
  int yy= y2;
  for (J=0; J<nry-1; J++) {
    yy -= ry2[J]-ry1[J];
    if (y>=yy) break;
  }

  int i, d=MAX_SI, m=-1;
  for (i=0; i<nr; i++) {
    int d1= max (0, max (bx[i]-I, I-(bx[i]+bw[i]-1)));
    int d2= max (0, max (by[i]-J, J-(by[i]+bh[i]-1)));
    int dd= d1+d2;
    if ((dd<d) && (!bt[i].ro)) { d= dd; m= i; }
  }
  if (m==-1) return std_path (x, y, delta);
  else {
    path p= bs[m]->var_get_path (x- bs[m]->x0, y- bs[m]->y0, delta);
    return path (m, p);
  }
}

cursor
matrix_box_rep::get_cursor (path p) {
  if (nil (p->next)) return std_cursor (p);
  else {
    int i=p->item;
    cursor cu (bs[i]->var_get_cursor (p->next));
    cu->x0 += bs[i]->x0;
    cu->y0 += bs[i]->y0;
    return cu;
  }
}

SI
matrix_box_rep::sub_base () {
  int i, m=MAX_SI;
  for (i=0; i<nr; i++)
    m= min (m, bs[i]->y0+ bs[i]->sub_base ());
  return m;
}

SI
matrix_box_rep::sup_base () {
  int i, m=-MAX_SI;
  for (i=0; i<nr; i++)
    m= max (m, bs[i]->y0+ bs[i]->sup_base ());
  return m;
}

/******************************************************************************
* Mosaic items
******************************************************************************/

struct mosaic_item_box_rep: public composite_box_rep {
  mosaic_item_box_rep (box b);
  operator tree () { return tree ("mosaic item", (tree) bs[0]); }
  cursor get_cursor (path p);
  path get_path (SI x, SI y, SI delta);
  path left_path ();
  path right_path ();
};

mosaic_item_box_rep::mosaic_item_box_rep (box b) {
  insert (b, 0, 0);
  position ();
}

cursor
mosaic_item_box_rep::get_cursor (path p) {
  return composite_box_rep::get_cursor (p);
}

path
mosaic_item_box_rep::get_path (SI x, SI y, SI delta) {
  path p= composite_box_rep::get_path (x, y, delta);
  if (p == path (0)) return path (0, bs[0]->left_path ());
  if (p == path (1)) return path (0, bs[0]->right_path ());
  return p;
}

path
mosaic_item_box_rep::left_path () {
  return path (0, bs[0]->left_path ());
}

path
mosaic_item_box_rep::right_path () {
  return path (0, bs[0]->right_path ());
}

/******************************************************************************
* User interface
******************************************************************************/

box
matrix_box (int cols, int rows, array<box> b, SI hspace, SI vspace, SI h) {
  int x, y;
  matrix_box_rep* M= new matrix_box_rep (cols, rows);
  for (y=0; y<rows; y++)
    for (x=0; x<cols; x++)
      M->insert (b[y*cols+x], x, y, 1, 1);
  M->format ();
  for (x=0; x<cols; x++) {
    M->cx1[x]-=hspace>>1;
    M->cx2[x]+=hspace>>1;
  }
  for (y=0; y<rows; y++) {
    M->ry1[y]-=vspace>>1;
    M->ry2[y]+=vspace>>1;
  }
  M->ready ();

  int d= h- ((M->y1+M->y2)>>1), i;
  M->y1+=d; M->y2+=d;
  M->y3+=d; M->y4+=d;
  for (i=0; i<M->nr; i++)
    M->bs[i]->y0 += d;

  return M;
}

box
table_box (int cols, int rows, array<box> b,
	   SI hspace, SI vspace, SI h, color c)
{
  int x, y;
  matrix_box_rep* M= new matrix_box_rep (2*cols+1, 2*rows+1);
  for (y=0; y<rows; y++)
    for (x=0; x<cols; x++)
      M->insert (b[y*cols+x], 2*x+1, 2*y+1, 1, 1);
  M->format ();
  for (x=0; x<cols; x++) {
    M->cx1[2*x+1]-=hspace>>1;
    M->cx2[2*x+1]+=hspace>>1;
  }
  for (y=0; y<rows; y++) {
    M->ry1[2*y+1]-=vspace>>1;
    M->ry2[2*y+1]+=vspace>>1;
  }
  for (x=0; x<=cols; x++) {
    M->insert (empty_box (), 2*x, 0, 1, 2*rows+1);
    M->bt[M->nr-1].ro=TRUE;
    M->bt[M->nr-1].filled=TRUE;
    M->bt[M->nr-1].col=c;
    M->cx2[2*x] += PIXEL;
  }
  for (y=0; y<=rows; y++) {
    M->insert (empty_box (), 0, 2*y, 2*cols+1, 1);
    M->bt[M->nr-1].ro=TRUE;
    M->bt[M->nr-1].filled=TRUE;
    M->bt[M->nr-1].col=c;
    M->ry2[2*y] += PIXEL;
  }
  M->ready ();

  int d= h- ((M->y1+M->y2)>>1), i;
  M->y1+=d; M->y2+=d;
  M->y3+=d; M->y4+=d;
  for (i=0; i<M->nr; i++)
    M->bs[i]->y0 += d;

  return M;
}

box
mosaic_item_box (box b) {
  return new mosaic_item_box_rep (b);
}

box
mosaic_box (int cols, int rows, array<box> b, array<string> pos,
	    array<int> nr_x, array<int> nr_y, SI h, array<color> col)
{
  int x, y;
  matrix_box_rep* M= new matrix_box_rep (cols, rows);
  for (y=0; y<rows; y++)
    for (x=0; x<cols; x++) {
      int i= y*cols+x;
      if ((nr_x[i]*nr_y[i]) == 0) {
	M->insert (empty_box (), x, y, 1, 1);
	M->bt[M->nr-1].ro=TRUE;
      }
      else {
	M->insert (b[i], x, y, nr_x[i], nr_y[i]);
	mb_type* type= &(M->bt[M->nr-1]);
	if (col[i]!=-1) {
	  type->filled= TRUE;
	  type->col   = col[i];
	}
	type->hflag= TRUE;
	type->vflag= TRUE;
	if (pos[i][0] == 'n') type->vpos = 1;
	if (pos[i][0] == 'c') type->vpos = 0.5;
	if (pos[i][0] == '0') type->vflag= FALSE;
	if (pos[i][0] == 's') type->vpos = 0;
	if (pos[i][1] == 'w') type->hpos = 0;
	if (pos[i][1] == '0') type->hflag= FALSE;
	if (pos[i][1] == 'c') type->hpos = 0.5;
	if (pos[i][1] == 'e') type->hpos = 1;
      }
    }
  
  M->format ();
  M->ready ();
  
  int d= h- ((M->y1+M->y2)>>1), i;
  M->y1+=d; M->y2+=d;
  M->y3+=d; M->y4+=d;
  for (i=0; i<M->nr; i++)
    M->bs[i]->y0 += d;
  
  return M;
}

#endmodule // code_matrix_boxes
