
/******************************************************************************
* MODULE     : modifier.gen.cc
* DESCRIPTION: a modifier box modifies the behaviour of another box
*              as to cursor movements etc. , but displays in the same way
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <Boxes/modifier.gen.h>

#module code_modifier_boxes
#import modifier_boxes

/*****************************************************************************/
// Modifier boxes
/*****************************************************************************/

modifier_box_rep::modifier_box_rep (box b2): b(b2) {
  x1= b->x1; y1= b->y1;
  x2= b->x2; y2= b->y2;
  x3= b->x3; y3= b->y3;
  x4= b->x4; y4= b->y4;
}

modifier_box_rep::~modifier_box_rep () {}

int
modifier_box_rep::subnr () {
  return 1;
}

box
modifier_box_rep::subbox (int i) { (void) i;
  return b;
}

void
modifier_box_rep::display (ps_device dev) { (void) dev; }


path
modifier_box_rep::encode (path p) {
  if (atom (p) || (p->item != 0))
    fatal_error ("invalid path", "modifier_box_rep::encode");
  return p->next;
}

path
modifier_box_rep::decode (path p) {
  return path (0, p);
}

cursor
modifier_box_rep::get_cursor (path p) {
  p= decode (p);
  if (atom (p)) return std_cursor (p);
  cursor cu= b->get_cursor (p->next);
  cu->delta -= get_delta (cu->x0, b->x1, b->x2);
  cu->x0 += b->x0;
  cu->y0 += b->y0;
  return cu;
}

path
modifier_box_rep::get_path (SI x, SI y, SI delta) {
  return encode (path (0, b->get_path (x, y, delta)));
}

path
modifier_box_rep::left_path () {
  return encode (path (0, b->left_path ()));
}

path
modifier_box_rep::right_path () {
  return encode (path (0, b->right_path ()));
}

selection
modifier_box_rep::select (path start, path end) {
  selection sel= b->select (decode (start)->next, decode (end)->next);
  sel->rs      = translate (sel->rs, x0, y0);
  sel->start   = encode (path (0, sel->start));
  sel->end     = encode (path (0, sel->end));
  return sel;
}

modifier_box_rep::operator tree () {
  return (tree) b;
}

double modifier_box_rep::left_slope () { return b->left_slope (); }
double modifier_box_rep::right_slope () { return b->right_slope (); }
SI modifier_box_rep::sub_base () { return b->sub_base (); }
SI modifier_box_rep::sup_base () { return b->sup_base (); }
SI modifier_box_rep::left_correction () { return b->left_correction (); }
SI modifier_box_rep::right_correction () { return b->right_correction (); }
SI modifier_box_rep::left_sub_correction (double level) {
  return b->left_sub_correction (level); }
SI modifier_box_rep::left_sup_correction (double level) {
  return b->left_sup_correction (level); }
SI modifier_box_rep::right_sub_correction (double level) {
  return b->right_sub_correction (level); }
SI modifier_box_rep::right_sup_correction (double level) {
  return b->right_sup_correction (level); }

/*****************************************************************************/
// Group boxes
/*****************************************************************************/

class group_box_rep: public modifier_box_rep {
public:
  group_box_rep (box b);
  operator tree () { return tree ("group", subbox(0)); }
  path encode (path p);
  path decode (path p);
};

group_box_rep::group_box_rep (box b2): modifier_box_rep (b2) {}

path
group_box_rep::decode (path p) {
  if (p->item==0) return path (0, b->left_path ());
  else return path (0, b->right_path ());
}

path
group_box_rep::encode (path p) {
  p= p->next;
  while ((!nil(p)) && (p->item==0)) p= p->next;
  if (nil (p)) return path (0);
  else return path (1);
}

/*****************************************************************************/
// Symbol boxes
/*****************************************************************************/

class symbol_box_rep: public modifier_box_rep {
  int n;
public:
  symbol_box_rep (box b, int n);
  operator tree () { return tree ("symbol", subbox(0)); }
  path encode (path p);
  path decode (path p);
};

symbol_box_rep::symbol_box_rep (box b2, int n2):
  modifier_box_rep (b2), n (n2) {}

path
symbol_box_rep::decode (path p) {
  if (p->item<=(n>>1)) return path (0, b->left_path ());
  else return path (0, b->right_path ());
}

path
symbol_box_rep::encode (path p) {
  p=p->next;
  if (p->item<=0) return path (0);
  else return path (n);
}

/*****************************************************************************/
// Error boxes
/*****************************************************************************/

class error_box_rep: public modifier_box_rep {
  int n;
public:
  error_box_rep (box b, int n);
  operator tree () { return tree ("error", subbox(0)); }
  path encode (path p);
  path decode (path p);
};

error_box_rep::error_box_rep (box b2, int n2):
  modifier_box_rep (b2), n (n2) {}

path
error_box_rep::decode (path p) {
  return path (0, p);
}

path
error_box_rep::encode (path p) {
  p=p->next;
  if (p->item<=(n>>1)) return path (0);
  else return path (n);
}

/*****************************************************************************/
// Shorter boxes are used for hyphenation
/*****************************************************************************/

class shorter_box_rep: public modifier_box_rep {
  int max;
public:
  shorter_box_rep (box b, int max);
  operator tree () { return tree ("shorter", subbox(0)); }
  path encode (path p);
  path decode (path p);
};

shorter_box_rep::shorter_box_rep (box b2, int max2):
  modifier_box_rep (b2), max (max2) {}

path
shorter_box_rep::decode (path p) {
  return path (0, p);
}

path
shorter_box_rep::encode (path p) {
  p=p->next;
  if (p->item>max) return path (max);
  else return path (p->item);
}

/*****************************************************************************/
// box construction routines
/*****************************************************************************/

box
group_box (box b) {
  return new group_box_rep (b);
}

box
symbol_box (box b, int n) {
  return new symbol_box_rep (b, n);
}

box
error_box (box b, int n) {
  return new error_box_rep (b, n);
}

box
shorter_box (box b, int max) {
  return new shorter_box_rep (b, max);
}

#endmodule // code_modifier_boxes
