
/******************************************************************************
* MODULE     : math_data.gen.cc
* DESCRIPTION: the way data are organized for the math_editor
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <data.gen.h>

#module code_data
#import data

/******************************************************************************
* The hashmaps for getting properties of the math-editor constructors
******************************************************************************/

hashmap<string,int> math_primary (Unknown);
hashmap<string,int> math_secondary (0);
hashmap<string,int> math_ternary (0);
string EMPTY ("");

/******************************************************************************
* The special constructors from which our expressions are built up
******************************************************************************/

string UNKNOWN ("unknown");
string DOCUMENT ("document");
string PARAGRAPH ("paragraph");
string CONCAT ("concat");
string FORMAT ("format");
  string LINE_BREAK ("line break");
  string NEW_LINE ("new line");
  string LINE_SEP ("line separator");
  string NEXT_LINE ("next line");
  string NO_FIRST_INDENT ("no first indentation");
  string NO_LAST_INDENT ("no last indentation");
  string YES_FIRST_INDENT ("enable first indentation");
  string YES_LAST_INDENT ("enable last indentation");
  string PAGE_BREAK ("page break");
  string NEW_PAGE ("new page");
  string NO_PAGE_BREAK_BEFORE ("no page break before");
  string NO_PAGE_BREAK_AFTER ("no page break after");
string HSPACE ("hspace");
string VSPACE_BEFORE ("vspace*");
string VSPACE_AFTER ("vspace");
string SPACE ("space");
string HTAB ("htab");
string SPLIT ("split");
string MOVE ("move");
string RESIZE ("resize");

string LEFT_SUB ("lsub");
string LEFT_SUP ("lsup");
string RIGHT_SUB ("rsub");
string RIGHT_SUP ("rsup");
string FRAC ("frac");
string SQRT ("sqrt");
string WIDE ("wide");
string NEG ("neg");
string TREE ("tree");
string MATRIX ("matrix");
string TABLE ("table");
string MOSAIC ("mosaic");
string MOSAIC_ITEM ("mosaic item");

string INACTIVE ("inactive");
string SYMBOL ("<>");
string LATEX ("latex");
string HYBRID ("hybrid");
string BACKUP (";");
string QUOTE ("'");
string ASSIGN (":=");
string SET ("set");
string RESET ("reset");
string APPLY ("\\");
string BEGIN ("begin");
string END ("end");
string FUNCTION ("func");
string ENVIRONMENT ("env");
string LABEL ("label");
string REFERENCE ("reference");
string WRITE ("write");

string PLUS ("+");
string MINUS ("-");
string TIMES ("x");
string MERGE ("*");
string NUMBER ("nr");
string TRANSLATE ("translate");

string GRAPHICS ("graphics");
string POINT ("point");
string LINE ("line");
string ARC ("arc");
string BEZIER ("bezier");
string POSTSCRIPT ("postscript");

string& SUB (bool right) { return right? RIGHT_SUB: LEFT_SUB; }
string& SUP (bool right) { return right? RIGHT_SUP: LEFT_SUP; }

string CONSTRUCTOR_NAME[CONSTRUCTOR_NR]= {
  UNKNOWN, DOCUMENT, PARAGRAPH, CONCAT, FORMAT,
  HSPACE, VSPACE_BEFORE, VSPACE_AFTER, SPACE,
  HTAB, SPLIT, MOVE, RESIZE,

  LEFT_SUB, LEFT_SUP, RIGHT_SUB, RIGHT_SUP,
  FRAC, SQRT, WIDE, NEG, TREE,
  MATRIX, TABLE, MOSAIC, MOSAIC_ITEM,

  INACTIVE, SYMBOL, LATEX, HYBRID, BACKUP, QUOTE,
  ASSIGN, SET, RESET, APPLY, BEGIN, END,
  FUNCTION, ENVIRONMENT, LABEL, REFERENCE, WRITE,
  PLUS, MINUS, TIMES, MERGE, NUMBER, TRANSLATE,
  GRAPHICS, POINT, LINE, ARC, BEZIER, POSTSCRIPT
};

int CONSTRUCTOR_ARITY[CONSTRUCTOR_NR]= {
  -1, -1, -1, -1, -1,
  -1, -1, -1, -1,
  -1, -1, 2, 4,

  1, 1, 1, 1,
  2, -1, 1, 1, -1,
  -1, -1, -1, -1,

  1, 1, 1, 1, 1, 1,
  2, 2, 1, -1, -1, 1,
  2, 2, -1, -1, -1,
  2, 2, 2, 2, 2, 3,
  -1, -1, -1, -1, -1, -1
};

/******************************************************************************
* Get the rightmost path in tree
******************************************************************************/

int
right_index (tree t) {
  return (N(t)==0? N(t->label): 1);
}

/******************************************************************************
* [format] Text and formula formatting takes place in three modes
* ----------------------------------------------------------------------------
* - [document] The entire document is an array of paragraphs.
* - [paragraph] A typeset paragraph whose entities are separated by spaces.
* - [concat] An array of boxes which are glued together horizontally.
******************************************************************************/

bool
is_document (tree t) {
  return is_func (t, DOCUMENT);
}

bool
is_concat (tree t) {
  return is_func (t, CONCAT);
}

bool
is_format (tree t) {
  return is_document (t) || is_concat (t);
}

/******************************************************************************
* [math] The main classes of mathematical formulas are as follows:
* ----------------------------------------------------------------------------
* - [regular] composite formulas of the following types
*     * [frac, sqrt]
*     * [wide_modifier] underline, overline, hat, etc.
*     * [tabular] matrices and tables
*   An important feature of such formulas is that from a cursor positioning
*   point of view, the children are completely independent from
*   the entire formula.
* - [script] subscripts, superscripts, both and primes.
*   The leftmost cursor position can be attained in two ways:
*   as the left position of the entire script or
*   the leftmost position of the leftmost child.
* - [operator] exp, sin, lim, \sum, etc.
*   They behave in a special way relative to scripts and spacing.
******************************************************************************/

bool
is_tabular (tree t) {
  return
    is_func (t, MATRIX) ||
    is_func (t, TABLE);
}

bool
is_regular (tree t) {
  return
    is_func (t, FRAC) ||
    is_func (t, SQRT) ||
    is_tabular (t);
}

bool
is_script (tree t, bool& right) {
  if (is_func (t, LEFT_SUB) ||
      is_func (t, LEFT_SUP)) { right=FALSE; return TRUE; }
  if (is_func (t, RIGHT_SUB) ||
      is_func (t, RIGHT_SUP)) { right=TRUE; return TRUE; }
  return FALSE;
}

bool
is_math (tree t) {
  bool right;
  return
    is_regular (t) ||
    is_script (t, right) ||
    is_tabular (t);
}

/******************************************************************************
* [modifier]
* ----------------------------------------------------------------------------
* The only functions which are allowed to modify the environment
* are ASSIGN, SET and RESET. ASSIGN makes definitive changes,
* while changes made by SET can be recovered using RESET
******************************************************************************/

bool
is_modifier (tree t) {
  if (N(t)==0) return FALSE;
  string s= t->label;
  return (s==ASSIGN) || (s==SET) || (s==RESET);
}

bool
is_dynamic (tree t) {
  if (N(t)==0) return FALSE;
  string s= t->label;
  return
    (s==SPLIT) ||
    (s==LATEX) || (s==HYBRID) || (s==QUOTE) ||
    (s==ASSIGN) || (s==SET) || (s==RESET) ||
    (s==APPLY) || (s==BEGIN) || (s==END) ||
    (s==FUNCTION) || (s==ENVIRONMENT) ||
    (s==LABEL) || (s==REFERENCE) || (s==WRITE) ||
    (s==POSTSCRIPT);
}

/******************************************************************************
* brackets and bigops
******************************************************************************/

bool
is_opening (tree t) {
  return
    (N(t)==0) && (N(t->label)>6) &&
    (t->label (0, 6) == "<left-");
}

bool
is_middle (tree t) { (void) t;
  return
    (N(t)==0) && (N(t->label)>5) &&
    (t->label (0, 5) == "<mid-");
}

bool
is_closing (tree t) {
  return
    (N(t)==0) && (N(t->label)>7) &&
    (t->label (0, 7) == "<right-");
}

bool
is_bracket (tree t) {
  return is_opening (t) || is_middle (t) || is_closing (t);
}

bool
is_bigop (tree t) {
  return
    (N(t)==0) && (N(t->label)>5) &&
    (t->label (0, 5) == "<big-");
}

bool
is_prime (tree t) {
  if ((N(t)>0) || (N(t->label)==0)) return FALSE;
  int i;
  string s= t->label;
  if ((s[0]!='`') && (s[0]!='\'')) return FALSE;
  for (i=1; i<N(s); i++)
    if (s[i]!=s[0]) return FALSE;
  return TRUE;
}

/******************************************************************************
* initialization
******************************************************************************/

#define macro init_primitive

void
macro (string s, int primary, int secondary=0, int ternary=0) {
  math_primary  (s)= primary;
  math_secondary(s)= secondary;
  math_ternary  (s)= ternary;
}

static bool edit_data_initialized=FALSE;

void
initialize_data () {
  if (edit_data_initialized) return;
  edit_data_initialized=TRUE;

  macro (DOCUMENT, Document);
  macro (PARAGRAPH, Paragraph);
  macro (CONCAT, Concat);
  macro (FORMAT, Format);
  macro (HSPACE, Hspace);
  macro (VSPACE_BEFORE, Vspace_before);
  macro (VSPACE_AFTER, Vspace_after);
  macro (SPACE, Space);
  macro (HTAB, Htab);
  macro (SPLIT, Split);
  macro (MOVE, Move);
  macro (RESIZE, Resize);

  macro (LEFT_SUB, Left_sub);
  macro (LEFT_SUP, Left_sup);
  macro (RIGHT_SUB, Right_sub);
  macro (RIGHT_SUP, Right_sup);
  macro (FRAC, Frac);
  macro (SQRT, Sqrt);
  macro (WIDE, Wide);
  macro (NEG, Neg);
  macro (TREE, Tree);
  macro (MATRIX, Matrix);
  macro (TABLE, Table);
  macro (MOSAIC, Mosaic);
  macro (MOSAIC_ITEM, Mosaic_item);

  macro (INACTIVE, Inactive);
  macro (SYMBOL, Symbol);
  macro (LATEX, Latex);
  macro (HYBRID, Hybrid);
  macro (BACKUP, Backup);
  macro (QUOTE, Quote);
  macro (ASSIGN, Assign);
  macro (SET, Set);
  macro (RESET, Reset);
  macro (APPLY, Apply);
  macro (BEGIN, Begin);
  macro (END, End);
  macro (FUNCTION, Function);
  macro (ENVIRONMENT, Environment);
  macro (LABEL, Label);
  macro (REFERENCE, Reference);
  macro (WRITE, Write);

  macro (PLUS, Plus);
  macro (MINUS, Minus);
  macro (TIMES, Times);
  macro (MERGE, Merge);
  macro (NUMBER, Number);
  macro (TRANSLATE, Translate);

  macro (GRAPHICS, Graphics);
  macro (POINT, Point);
  macro (LINE, Line);
  macro (ARC, Arc);
  macro (BEZIER, Bezier);
  macro (POSTSCRIPT, Postscript);
}

#undef macro

#endmodule // code_data
