
/******************************************************************************
* MODULE     : page_breaker.gen.cc
* DESCRIPTION: Line breaking facility for paragraphs
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <hashmap.gen.cc>

#module code_page_breaker

/******************************************************************************
* The page_breaker class
******************************************************************************/

struct page_breaker_rep {
  array<page_item> l;
  int n;
  SI  page_height;

  array<int> best_prev;
  array<int> best_pen;
  array<SI>  best_spc_pen;

  int        the_start;
  int        the_end;
  int        the_height;
  space      the_spc;

  page_breaker_rep (array<page_item> l, SI page_height);

  void propose (int pos, int prev, int pen, SI spc_pen);
  void process_sub (int pass);
  void process (int pass);
  void make_breaks (array<int>& b, int pos);
  array<int> compute_breaks ();
};

/******************************************************************************
* Initialization
******************************************************************************/

page_breaker_rep::page_breaker_rep (array<page_item> l2, SI page_height2):
  l (l2), n (N(l)), page_height (page_height2),
  best_prev (n+1), best_pen (n+1), best_spc_pen (n+1)
{
  int i;
  for (i=0; i<n+1; i++) {
    best_prev [i]= 0;
    best_pen [i]    = HYPH_INVALID;
    best_spc_pen [i]= HYPH_INVALID;
  }
  best_pen [0]    = 0;
  best_spc_pen [0]= 0;
}

/******************************************************************************
* The main page breaking routine
******************************************************************************/

inline int square (int i) { return i*i; }

void
page_breaker_rep::propose (int pos, int prev, int pen, SI spc_pen) {
  if ((pen < best_pen [pos]) ||
      ((pen == best_pen [pos]) &&
       (spc_pen < best_spc_pen [pos])))
  {
    best_prev [pos]   = prev;
    best_pen [pos]    = pen;
    best_spc_pen [pos]= spc_pen;
  }
}

void
page_breaker_rep::process_sub (int pass) {
  int   end    = the_end;
  int   height = the_height;
  space spc    = the_spc;
  int   pen    = best_pen [the_start];
  SI    spc_pen= best_spc_pen [the_start];

  // determine first possible page break
  while ((end < n) &&
	 ((end == the_start) ||
	  ((spc->max < height) && (pass == 1)) ||
	  (l[end-1]->penalty == HYPH_INVALID)))
  {
    if (end == n) break;
    if ((end > the_start) && (l[end-1]->type == PAGE_CONTROL_ITEM)) {
      if (l[end-1]->t == tree (FORMAT, PAGE_BREAK)) break;
      if (l[end-1]->t == tree (FORMAT, NEW_PAGE)) break;
    }
    if (end > 0) spc= spc+ l[end-1]->spc;
    spc= spc+ space (l[end]->b->h());
    end++;
  }

  // prepare for next call to process_sub
  page_item& item (l[the_start]);
  the_start    ++;
  the_end    = end;
  the_height = height;
  the_spc    = spc- space (item->b->h())- item->spc;

  // propose better page breaks
  int i;
  for (i=end; i<n; i++) {
    if (spc->min > height) break;
    if (l[i-1]->type == PAGE_CONTROL_ITEM) {
      if (l[i-1]->t == tree (FORMAT, PAGE_BREAK))
	propose (i, the_start-1, pen, spc_pen);
      if (l[i-1]->t == tree (FORMAT, NEW_PAGE))
	propose (i, the_start-1, pen- 10000, spc_pen);
    }
    else {
      if (spc->max >= height)
	   propose (i, the_start-1, pen+ l[i-1]->penalty,
		    spc_pen+ square ((spc->def-the_height)/PIXEL));
      else propose (i, the_start-1, pen+ 10000,
		    spc_pen+ square ((spc->def-the_height)/PIXEL));
    }
    spc= spc+ l[i-1]->spc+ space (l[i]->b->h());
  }
  if (i==n) {
    int sq= square ((spc->def-the_height)/PIXEL);
    propose (i, the_start-1, pen, (spc_pen+ spc->def>the_height)? sq: 0);
  }
}

void
page_breaker_rep::process (int pass) { 
  the_start = 0;
  the_end   = 0;
  the_height= page_height;
  the_spc   = 0;
  while (the_start<n) process_sub (pass);
}

void
page_breaker_rep::make_breaks (array<int>& b, int pos) {
  if (pos>0) make_breaks (b, best_prev [pos]);
  b << pos;
}

array<int>
page_breaker_rep::compute_breaks () {
  process (1);
  if (best_pen [n] >= HYPH_INVALID) process (2);
  array<int> b (0);
  make_breaks (b, n);
  /*
  cout << "------------------------------------------------------------------------------\n";
  cout << "Page items: " << l << "\n";
  cout << "Previous  : " << best_prev << "\n";
  cout << "Penalties : " << best_pen << "\n";
  cout << "Spaces    : " << best_spc_pen << "\n";
  cout << "Breaks    : " << b << "\n";
  */
  return b;
}

/******************************************************************************
* The exported page breaking routine
******************************************************************************/

array<int>
page_breaks (array<page_item> l, SI page_height) {
  page_breaker_rep* H= new page_breaker_rep (l, page_height);
  array<int> b= H->compute_breaks ();
  delete H;
  return b;
}

#endmodule // code_page_breaker
