
/******************************************************************************
* MODULE     : pager.gen.cc
* DESCRIPTION: Control routines for typesetting paragraphs
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <Atom/typesetter.gen.h>
#include <Paragraph/liner.gen.h>
#include <Document/pager.gen.h>

#module code_pager
#import typesetter
#import liner
#import pager

/******************************************************************************
* Routines for the pager class
******************************************************************************/

pager_rep::pager_rep (edit_env env2): env (env2), style ("?") {
  style (PAGE_ODD_HEADER) = env->get (PAGE_ODD_HEADER);
  style (PAGE_ODD_FOOTER) = env->get (PAGE_ODD_FOOTER);
  style (PAGE_EVEN_HEADER)= env->get (PAGE_EVEN_HEADER);
  style (PAGE_EVEN_FOOTER)= env->get (PAGE_EVEN_FOOTER);
  style (PAGE_THIS_HEADER)= "";
  style (PAGE_THIS_FOOTER)= "";

  SI dummy;
  paper   = (env->get_string (PAGE_MEDIUM) == "paper");
  env->get_page_pars (dummy, text_height, width, height, odd, even, top, bot);
  head_sep= env->get_length (PAGE_HEAD_SEP);
  foot_sep= env->get_length (PAGE_FOOT_SEP);
  show_hf = env->get_bool (PAGE_SHOW_HF) && paper;

  cur_top    = 0;
  prev_vspc  = space (0);
  prev_nobr  = FALSE;
}

/******************************************************************************
* Typesetting a page
******************************************************************************/

void
pager_rep::start_page () {
  cur_spc = space (0);
  cur_sep = space (0);
  lines_bx= array<box> (0);
  lines_ht= array<space> (0);
  lines_st= array<path> (0);
  lines_en= array<path> (0);
  lines_fl= array<int> (0);
}

void
pager_rep::print (page_item item) {
  cur_spc= cur_spc+ cur_sep+ space (item->b->h());
  cur_sep= item->spc;
  if (item->type == PAGE_CONTROL_ITEM) {
    if (is_func (item->t, "env_page")) {
      if (((item->t[0]->label == PAGE_THIS_HEADER) ||
	   (item->t[0]->label == PAGE_THIS_FOOTER)) &&
	  is_func (item->t[1], "", 0)) style (item->t[0]->label)= " ";
      else style (item->t[0]->label)= copy (item->t[1]);
    }
    return;
  }
  lines_bx << item->b;
  lines_ht << item->spc;
  lines_st << item->p;
  lines_en << item->q;
  lines_fl << item->f;
}

void
pager_rep::end_page (bool flag) {
  int i;
  array<SI> spc (N(lines_bx));
  
  // stretching case
  if (paper && (!flag) && (cur_spc->def < text_height) &&
      (cur_spc->max > cur_spc->def)) {
    double f=
      ((double) (text_height - cur_spc->def)) /
      ((double) (cur_spc->max - cur_spc->def));
    for (i=0; i<N(lines_bx)-1; i++)
      spc[i]= lines_ht[i]->def+
	((SI) (f*((double) lines_ht[i]->max- lines_ht[i]->def)));
  }

  // shrinking case
  else if (paper && (cur_spc->def > text_height) &&
	   (cur_spc->def > cur_spc->min)) {
    double f=
      ((double) (cur_spc->def - text_height)) /
      ((double) (cur_spc->def - cur_spc->min));
    if (f>1.0) f=1.0;
    for (i=0; i<N(lines_bx)-1; i++)
      spc[i]= lines_ht[i]->def-
	((SI) (f*((double) lines_ht[i]->def- lines_ht[i]->min)));
  }

  // normal case
  else for (i=0; i<N(lines_bx)-1; i++) spc[i]= lines_ht[i]->def;

  box lb= new lines_box_rep (lines_bx, spc, lines_st, lines_en, lines_fl);
  SI  l = (N(pages)&1)==0? odd: even;
  SI  h = paper? height: (lb->h ()+ top+ bot);
  box pb= new page_box_rep (lb, width, h, l, top, top+ text_height,
			    make_header(), make_footer(), head_sep, foot_sep);

  for (i=0; i<N(lines_bx); i++) {
    int pnr= lines_st[i]->item;
    box sb = pb[0][0][i];
    SI ox= pb[0]->x0+ pb[0][0]->x0;
    SI oy= cur_top+ pb[0]->y0+ pb[0][0]->y0;
    rectangle r1= p[pnr]->r;
    rectangle r2= rectangle (sb->x3+ sb->x0+ ox, sb->y3+ sb->y0+ oy,
			     sb->x4+ sb->x0+ ox, sb->y4+ sb->y0+ oy);
    if (r1 == rectangle (0, 0, 0, 0)) p[pnr]->r= r2;
    else p[pnr]->r= least_upper_bound (rectangles (r1, r2));
  }

  cur_top -= h;
  pages    << pb;
  pages_st << lines_st[0];
  pages_en << lines_en[N(lines_en)-1];
  pages_fl << ((lines_fl[0]&1) | (lines_fl[N(lines_fl)-1]&2));
}

/******************************************************************************
* Typesetting a page
******************************************************************************/

void
pager_rep::insert_paragraph (paragraph par, int nr, bool empty) {
  int i= N(l)-1;
  while ((i>=0) && (l[i]->type == PAGE_CONTROL_ITEM)) i--;
  if (i>=0) {
    l[i]->spc = l[i]->spc + max (prev_vspc, par->vspc_before);
    if (prev_nobr || par->nobr_before) l[i]->penalty= HYPH_INVALID;
  }
  prev_vspc= par->vspc_after;
  prev_nobr= par->nobr_after;

  p << par;
  if (!empty)
    for (i=0; i<N(par->l); i++)
      l << page_item (par->l[i], nr);
  par->r= rectangle (0, 0, 0, 0);
}

box
pager_rep::make_header () {
  if (!show_hf) return empty_box ();
  env->set (PAGE_NR, as_string (N(pages)+1));
  typesetter T= new typesetter_rep (env);
  string which= (N(pages)&1)==0? PAGE_ODD_HEADER: PAGE_EVEN_HEADER;
  if (style [PAGE_THIS_HEADER] != "") which= PAGE_THIS_HEADER;
  box b= T->typeset (tree (PARAGRAPH, style [which]));
  style (PAGE_THIS_HEADER) = "";
  delete T;
  return b;
}

box
pager_rep::make_footer () {
  if (!show_hf) return empty_box ();
  env->set (PAGE_NR, as_string (N(pages)+1));
  typesetter T= new typesetter_rep (env);
  string which= (N(pages)&1)==0? PAGE_ODD_FOOTER: PAGE_EVEN_FOOTER;
  if (style [PAGE_THIS_FOOTER] != "") which= PAGE_THIS_FOOTER;
  box b= T->typeset (tree (PARAGRAPH, style [which]));
  style (PAGE_THIS_FOOTER) = "";
  delete T;
  return b;
}

box
pager_rep::make_pages () {
  array<int> breaks;
  if (paper) breaks= page_breaks (l, text_height);
  else { breaks= array<int> (2); breaks[0]= 0; breaks[1]= N(l); }

  int i;
  for (i=0; i<N(breaks)-1; i++) {
    int j, start= breaks[i], end= breaks[i+1];
    start_page ();
    for (j=start; j<end; j++) print (l[j]);
    bool newpage=
      (l[j-1]->type == PAGE_CONTROL_ITEM) &&
      (l[j-1]->t == tree (FORMAT, NEW_PAGE));
    end_page (newpage || (i==N(breaks)-2));
  }

  array<SI> spc (N(pages));
  for (i=0; i<N(pages); i++) spc[i]=0;

  return new pages_box_rep (pages, spc, pages_st, pages_en, pages_fl);
}

/******************************************************************************
* Typesetting a paragraph
******************************************************************************/

box
typesetter_rep::typeset_paragraph (tree t) {
  int i;
  liner L= new liner_rep (env);
  L->typeset_3_passes (t[0]);

  int n= N(L->par->l);
  array<box>  lines_bx (n);
  array<SI>   lines_ht (n);
  array<path> lines_st(n), lines_en(n);
  array<int>  lines_fl (n);
  for (i=0; i<n; i++) {
    page_item item (L->par->l[i], 0);
    lines_bx[i]= item->b;
    lines_ht[i]= item->spc->def;
    lines_st[i]= item->p;
    lines_en[i]= item->q;
    lines_fl[i]= item->f;
  }

  delete L;
  return new lines_box_rep (lines_bx, lines_ht, lines_st, lines_en, lines_fl);
}

/******************************************************************************
* Typesetting a document
******************************************************************************/

box
typesetter_rep::typeset_document (tree t) {
  int i;
  pager P= new pager_rep (env);
  for (i=0; i<N(t); i++) {
    liner L= new liner_rep (env);
    L->typeset_3_passes (t[i]);
    P->insert_paragraph (L->par, i, FALSE);
    // P->insert_paragraph (L->par, i, (i<N(t)-1) && is_func (t[i], "", 0));
    delete L;
  }
  env->save ();
  box b= P->make_pages ();
  env->recover ();
  delete P;
  return b;
}

#endmodule // code_pager
