
/******************************************************************************
* MODULE     : font.gen.cc
* DESCRIPTION: fonts
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <font.gen.h>

#module code_font
#import font
#import code_resource (font_rep, font)

/******************************************************************************
* Constructors for fonts
******************************************************************************/

font_rep::font_rep (display dis2, string s):
  rep<font> (s),
  dis       (dis2),
  enc       (always_enc),
  spc       (0),
  extra     (0)
{
}

font_rep::font_rep (display dis2, string s, font fn):
  rep<font>    (s),
  dis          (dis2),
  enc          (always_enc),
  design_size  (fn->design_size),
  display_size (fn->display_size),
  slope        (fn->slope),
  spc          (fn->spc),
  extra        (fn->extra)
{
  copy_math_pars (fn);
}

void
font_rep::copy_parameters (server_font fn) {
  design_size = fn->design_size;
  display_size= fn->display_size;
  slope       = fn->slope;
  spc->def    = fn->spc->def;
  spc->min    = fn->spc->min;
  spc->max    = fn->spc->max;
  extra       = fn->extra;
  extra->min  = fn->extra->min;
  extra->max  = fn->extra->max;

  y1    = fn->y1;
  y2    = fn->y2;
  ysub  = fn->ysub;
  ysup  = fn->ysup;
  yfrac = fn->yfrac;
  wpt   = fn->wpt;
  wquad = fn->wquad;
  wunit = fn->wunit;
  wfrac = fn->wfrac;
  wsqrt = fn->wsqrt;
  wneg  = fn->wneg;
}

void
font_rep::copy_math_pars (font fn) {
  y1     = fn->y1;
  y2     = fn->y2;
  yfrac  = fn->yfrac;
  ysub   = fn->ysub;
  ysup   = fn->ysup;
  wpt    = fn->wpt;
  wunit  = fn->wunit;
  wquad  = fn->wquad;
  wfrac  = fn->wfrac;
  wsqrt  = fn->wsqrt;
  wneg   = fn->wneg;
}

/******************************************************************************
* Joining fonts: the global font properties are taken from the first font.
* If a string needs to be converted to a box, we first try to convert
* it using the first font, and next using the second one.
******************************************************************************/

struct join_font_rep: font_rep {
  font fn1, fn2;

  join_font_rep (string name, font fn1, font fn2);
  box convert (string s, color col, SI len1, SI len2);
};

join_font_rep::join_font_rep (string name, font fn1b, font fn2b):
  font_rep (fn1b->dis, name, fn1b), fn1 (fn1b), fn2 (fn2b)
{
  enc= join (fn1->enc, fn2->enc);
}

box
join_font_rep::convert (string s, color col, SI len1, SI len2) {
  if (fn1->enc->valid (s)) return fn1->convert (s, col, len1, len2);
  else return fn2->convert (s, col, len1, len2);
}

font
join (font fn1, font fn2) {
  string name= fn1->res_name * "|" * fn2->res_name;
  return make<font> (name, new join_font_rep (name, fn1, fn2));
}

/******************************************************************************
* Restricting fonts: restricting of the fonts underlying encoding,
* we decrease the number of valid strings which can be printed in the font
******************************************************************************/

struct subfont_rep: font_rep {
  font fn;
  subfont_rep (string name, font fn, encoding enc);
  box convert (string s, color col, SI len1, SI len2);
};

subfont_rep::subfont_rep (string name, font fn2, encoding enc2):
  font_rep (fn2->dis, name, fn2), fn (fn2)
{
  enc= enc2;
}

box
subfont_rep::convert (string s, color col, SI len1, SI len2) {
  return fn->convert (s, col, len1, len2);
}

font
subfont (font fn, encoding enc) {
  string name= fn->res_name * ":" * enc->res_name;
  return make<font> (name, new subfont_rep (name, fn, enc));
}

/******************************************************************************
* Error font: used to display unindentified characters
******************************************************************************/

struct error_font_rep: font_rep {
  font fn1;
  font fn2;

  error_font_rep (string name, font fn1, font fn2);
  box convert (string s, color col, SI len1, SI len2);
};

error_font_rep::error_font_rep (string name, font fn1b, font fn2b):
  font_rep (fn1b->dis, name, fn1b), fn1 (fn1b), fn2 (fn2b)
{
  enc= fn1->enc;
}

box
error_font_rep::convert (string s, color col, SI len1, SI len2) { (void) col;
  if ((N(s) < 2) || (s[0] != '<') || (s[N(s)-1] != '>'))
    return error_box (fn1->convert (s, fn1->dis->red, len1, len2), N(s));
  else {
    array<box> a(3);
    array<SI>  spc(3);
    a[0]= fn2->convert ("<", fn2->dis->red);
    a[1]= fn1->convert (s (1, N(s)-1), fn1->dis->red, len1, len2);
    a[2]= fn2->convert (">", fn2->dis->red);
    spc[0]= spc[1]= spc[2]= 0;
    return error_box (concat_box (a, spc), N(s));
  }
}

font
error_font (font fn1, font fn2) {
  string name= "error-" * fn1->res_name * "-" * fn2->res_name;
  return make<font> (name, new error_font_rep (name, fn1, fn2));
}

/******************************************************************************
* Miscellaneous
******************************************************************************/

int
script (int sz, int level) {
  int i;
  if (level<0) level=0;
  if (level>2) level=2;
  for (i=0; i<level; i++) sz= (sz*2+2)/3;
  return sz;
}

#endmodule // code_font
