
/******************************************************************************
* MODULE     : math_font.gen.cc
* DESCRIPTION: the font used for mathematical typesetting is actually a list
*              of several fonts, which may be specified by the user.
*              If a symbol needs to printed, we search in the list
*              for a font which is capable of printing it.
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <font.gen.h>
#include <translator.gen.h>
#include <Boxes/construct.gen.h>
#include <hashmap.gen.h>
#include <iterator.gen.h>
#include <file.gen.h>
#include <program.gen.h>

#module code_math_font
typedef int SI;
#import font
#import translator
#import construct_boxes
#import hashmap (string, pointer)
#import hashmap (string, tree)
#import hashmap_iterator (string, string)
#import file
#import program

program lisp_expr_to_program (string s, int i);

/******************************************************************************
* The compound font class
******************************************************************************/

struct math_font_rep: font_rep {
  font base_fn;
  int  size, dpi;
  hashmap<string,tree> symbol_table;
  hashmap<string,pointer> font_table;

  math_font_rep (string name, tree t, int size, int dpi, font base_fn);
  void insert_symbol (string key, string fn, string im);
  void insert_rubber (string key, string fn, string enc);
  void compile (tree t);
  box convert (string s, color col, SI len1, SI len2);
  box evaluate (tree t, color col, SI len1, SI len2);
};

math_font_rep::math_font_rep (string n, tree t, int s, int d, font b):
  font_rep (b->dis, n, b), base_fn (b), size (s), dpi (d),
  symbol_table (""), font_table (NULL)
{
  // cout << "Making " << n << "\n";
  compile (t);
}

void
math_font_rep::insert_symbol (string key, string fn, string im) {
  // cout << key << " -> " << im << ":" << fn << "\n";
  if ((N(key)>7) && (key(0,7)=="<large-")) {
    string sub= key (7, N(key)-1);
    insert_symbol ("<left-" * sub * ">", fn, im);
    insert_symbol ("<mid-" * sub * ">", fn, im);
    insert_symbol ("<right-" * sub * ">", fn, im);
  }

  if (symbol_table->contains (key)) {
    tree old= symbol_table [key];
    if (is_func (old, "rubber", 4)) {
      symbol_table(key)= tree ("rubber", fn, im, old[2], old[3]);
      return;
    }
  }
  symbol_table(key)= tree ("font", fn, im);
}

void
math_font_rep::insert_rubber (string key, string fn, string enc) {
  // cout << key << " -> rubber " << fn << "\n";
  if ((N(key)>7) && (key(0,7)=="<large-")) {
    string sub= key (7, N(key)-1);
    insert_rubber ("<left-" * sub * ">", fn, enc);
    insert_rubber ("<mid-" * sub * ">", fn, enc);
    insert_rubber ("<right-" * sub * ">", fn, enc);
  }

  if (symbol_table->contains(key)) {
    tree old= symbol_table[key];
    if (is_func (old, "font", 2)) {
      symbol_table(key)= tree ("rubber", old[0], old[1], fn, enc);
      return;
    }
  }
  symbol_table(key)= tree ("rubber", fn, enc, fn, enc);
}

static tree
font_lisp_to_tree (program p) {
  int i;
  if (is_string (p)) return tree (p->s);
  if (!is_composite (p)) return "";
  if (N(p)==0) return "";

  string s;
  if (is_string (p[0])) s= p[0]->s;
  tree t (s, N(p)-1);
  for (i=1; i<N(p); i++)
    t[i-1]= font_lisp_to_tree (p[i]);
  return t;
}

void
math_font_rep::compile (tree t) {
  // cout << "  Compiling " << t << "\n";
  if (t == "void") return;
  if (is_func (t, "join")) {
    int i;
    for (i=0; i<N(t); i++) compile (t[i]);
    return;
  }
  if (is_func (t, "tex-symbol", 2)) {
    translator trl= load_translator (t[1]->label);
    iterator<string> it= iterate (trl->dict);
    while (it->busy()) {
      string key= it->next();
      insert_symbol (key, t[0]->label, trl->dict[key]);
    }
    return;
  }
  if (is_func (t, "tex-rubber", 2)) {
    translator trl= load_translator (t[1]->label);
    iterator<string> it= iterate (trl->dict);
    while (it->busy()) insert_rubber (it->next(), t[0]->label, t[1]->label);
    return;
  }
  if (is_func (t, "virtual", 1)) {
    int i;
    file f ("$TEXMACS_HOME_PATH/fonts/virtual:$TEXMACS_PATH/fonts/virtual",
	    t[0]->label * ".vfn", "r");
    f->check_open ("math_font_rep::compile", "math_font.gen.cc");
    string s; f >> s;
    tree t= font_lisp_to_tree (lisp_expr_to_program (s, 0));
    if (t->label != "virtual-font") return;
    for (i=0; i<N(t); i++)
      if (N(t[i])==1) {
	string s= t[i]->label;
	if (N(s)>1) s= "<" * s * ">";
	symbol_table (s)= t[i][0];
      }
    return;
  }
  cerr << "The required font is " << t << "\n";
  fatal_error ("Bad compound font", "math_font_rep::compile");
}

/******************************************************************************
* Conversion of strings
******************************************************************************/

box
math_font_rep::convert (string s, color col, SI len1, SI len2) {
  if (!symbol_table->contains (s))
    return base_fn->convert (s, col, len1, len2);
  tree t= symbol_table[s];

  if (is_func (t, "font", 2)) {
    string family= t[0]->label;
    if (!font_table->contains (family)) {
      font fn= tex_text_font (base_fn->dis, family, size, dpi);
      fn->copy_math_pars (base_fn);
      font_table (family)= (pointer) fn.rep;
    }
    font fn ((font_rep*) font_table [family]);
    return symbol_box (fn->convert (t[1]->label, col, len1, len2), N(s));
  }

  if (is_func (t, "rubber", 4)) {
    int n= N(s);
    if ((N(s)>6) && (s(0,6)=="<left-")) s= "<large-" * s(6,N(s));
    if ((N(s)>5) && (s(0,5)=="<mid-")) s= "<large-" * s(5,N(s));
    if ((N(s)>7) && (s(0,7)=="<right-")) s= "<large-" * s(7,N(s));
    if ((N(s)>7) && (s(0,7)=="<large-") &&
        ((t[0] != t[2]) || (t[1] != t[3]))) {
      string family= t[0]->label;
      if (!font_table->contains (family)) {
	font fn= tex_text_font (base_fn->dis, family, size, dpi);
	fn->copy_math_pars (base_fn);
	font_table (family)= (pointer) fn.rep;
      }
      font fn ((font_rep*) font_table [family]);
      box b= symbol_box (fn->convert (t[1]->label, col, len1, len2), n);
      if ((len2-len1) <= (b->y2-b->y1)) return b;
    }

    string family= t[2]->label;
    if (!font_table->contains (family * "*")) {
      font fn= tex_rubber_font (base_fn->dis, t[3]->label, family, size, dpi);
      fn->yfrac= base_fn->yfrac;
      // fn->copy_math_pars (base_fn);
      font_table (family * "*")= (pointer) fn.rep;
    }
    font fn ((font_rep*) font_table [family * "*"]);
    return symbol_box (fn->convert (s, col, len1, len2), n);
  }

  return symbol_box (evaluate (t, col, len1, len2), N(s));
}

box
math_font_rep::evaluate (tree t, color col, SI len1, SI len2) {
  // cout << "Evaluating " << t << "\n";
  if (N(t)==0) {
    string s= t->label;
    if (N(s)>1) s= "<" * s * ">";
    return convert (s, col, len1, len2);
  }

  if (is_func (t, "glue", 2)) {
    SI correction= (base_fn->wpt*19)>>4;
    array<box> bs(2);
    array<SI> x(2), y(2);

    bs[0]= evaluate (t[0], col, len1, len2);
    x[0] = 0;
    y[0] = 0;
    bs[1]= evaluate (t[1], col, len1, len2);
    x[1] = bs[0]->x4- bs[0]->x3- correction;
    y[1] = 0;
    return composite_box (bs, x, y);
  }

  if (is_func (t, "add", 2)) {
    array<box> bs(2);
    array<SI> x(2), y(2);

    bs[0]= evaluate (t[0], col, len1, len2);
    x[0] = 0;
    y[0] = 0;
    bs[1]= evaluate (t[1], col, len1, len2);
    x[1] = ((bs[0]->x3+ bs[0]->x4- bs[1]->x3- bs[1]->x4) >> 1);
    y[1] = 0;
    return composite_box (bs, x, y);
  }

  if (is_func (t, "join")) {
    int i;
    SI unit= base_fn->wquad;
    array<box> bs(N(t));
    array<SI> x(N(t)), y(N(t));

    for (i=0; i<N(t); i++) {
      if (N(t[i])<2)
	fatal_error ("too little arguments in join",
		     "math_font_rep::evaluate");
      bs[i]= evaluate (t[i][1], col, len1, len2);
      x[i] = (SI) (as_double (t[i]->label)*unit);
      y[i] = (SI) (as_double (t[i][0]->label)*unit);
    }
    return composite_box (bs, x, y);
  }

  if (is_func (t, "enlarge")) {
    SI unit= base_fn->wquad;
    box b= evaluate (t[0], col, len1, len2);
    if (N(t)>1) b->x1 -= (SI) (as_double (t[1]->label)*unit);
    if (N(t)>2) b->x2 += (SI) (as_double (t[2]->label)*unit);
    if (N(t)>3) b->y1 -= (SI) (as_double (t[3]->label)*unit);
    if (N(t)>4) b->y2 += (SI) (as_double (t[4]->label)*unit);
    return b;
  }

  fatal_error ("invalid virtual constructor", "math_font_rep::evaluate");
}

/******************************************************************************
* User interface
******************************************************************************/

string tree_to_string (tree t);

font
math_font (tree t, int size, int dpi, font base_fn) {
  string full_name=
    "compound-" * tree_to_string (t) * as_string (size) * "@" * as_string(dpi);
  return make<font> (full_name,
    new math_font_rep (full_name, t, size, dpi, base_fn));
}

#endmodule // code_math_font
