
/******************************************************************************
* MODULE     : tex_rubber_font.gen.cc
* DESCRIPTION: TeX rubber fonts
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <font.gen.h>
#include <translator.gen.h>

#module code_tex_rubber_font
#import font
#import translator

/*****************************************************************************/
// TeX rubber fonts
/*****************************************************************************/

struct tex_rubber_font_rep: font_rep {
  server_font   fn;     // the associated server tex font
  translator    ext;    // big delimiters

  tex_rubber_font_rep (display dis, string name, string trl_name,
		       string family, int size, int dpi, int dsize);

  int  get_delim_n (unsigned char c, SI y1, SI y2);
  int  get_sqrt_n (unsigned char c, SI y1, SI y2);
  int  get_wide_n (unsigned char c, SI x1, SI x2);
  box  make_bigop (unsigned char c, color col, int n);
  box  make_delim (unsigned char c, color col, SI y1, SI y2);
  box  make_sqrt (unsigned char c, color col, SI y1, SI y2);
  box  make_wide (unsigned char c, color col, SI x1, SI x2);
  box  convert (string s, color col, SI y1, SI y2);
};

/******************************************************************************
* The implementation of TeX rubber fonts
******************************************************************************/

tex_rubber_font_rep::tex_rubber_font_rep (display dis, string name,
  string trl_name, string family, int size, int dpi, int dsize):
  font_rep (dis, name), ext (load_translator (trl_name))
{
  fn= dis->server_tex_font (family, size, dpi, dsize);
  copy_parameters (fn);
}

int
tex_rubber_font_rep::get_delim_n (unsigned char c, SI Y1, SI Y2) {
  int n=1;
  SI last=-1;
  while (TRUE) {
    SI x1, y1, x2, y2, x3, y3, x4, y4;
    fn->get_extents (c, n, x1, y1, x2, y2, x3, y3, x4, y4);
    if ((y2-y1) >= (Y2-Y1)) {
      if ((Y2-Y1) > ((last+ y2- y1) >> 1)) return n;
      else return max (n-1, 1);
    }
    if (y2-y1==last) return max (n-1, 1);
    last= y2-y1;
    n++;
  }
}

int
tex_rubber_font_rep::get_sqrt_n (unsigned char c, SI Y1, SI Y2) {
  int n=1;
  SI last=-1;
  while (TRUE) {
    SI x1, y1, x2, y2, x3, y3, x4, y4;
    fn->get_extents (c, n, x1, y1, x2, y2, x3, y3, x4, y4);
    if ((y2-y1) >= (Y2-Y1)) return n;
    if (y2-y1==last) return max (n-1, 1);
    last= y2-y1;
    n++;
  }
}

int
tex_rubber_font_rep::get_wide_n (unsigned char c, SI X1, SI X2) {
  int n=1;
  SI last=-1;
  while (TRUE) {
    SI x1, y1, x2, y2, x3, y3, x4, y4;
    fn->get_extents (c, n, x1, y1, x2, y2, x3, y3, x4, y4);
    if ((x2-x1) >= (X2-X1)) return n-1;
    if (x2-x1==last) return n-1;
    last= x2-x1;
    n++;
  }
}

font
tex_rubber_font (display dis, string trl_name,
		 string family, int size, int dpi, int dsize) {
  string name= "tex-rubber:"*family * as_string (size) * "@" * as_string(dpi);
  return make<font> (name,
    new tex_rubber_font_rep (dis, name, trl_name, family, size, dpi, dsize));
}

/*****************************************************************************/
// Creating TeX rubber boxes
/*****************************************************************************/

box
tex_rubber_font_rep::make_delim (unsigned char c, color col, SI Y1, SI Y2) {
  int n= get_delim_n (c, Y1, Y2);
  SI x1, y1, x2, y2, x3, y3, x4, y4;
  fn->get_extents (c, n, x1, y1, x2, y2, x3, y3, x4, y4);
  SI y= (Y1+Y2-y1-y2) >> 1;
  return delim_box (c, n, fn, y, col);
}

box
tex_rubber_font_rep::make_sqrt (unsigned char c, color col, SI Y1, SI Y2) {
  int n= get_sqrt_n (c, Y1, Y2);
  SI x1, y1, x2, y2, x3, y3, x4, y4;
  fn->get_extents (c, n, x1, y1, x2, y2, x3, y3, x4, y4);
  SI y= (Y1+Y2-y1-y2) >> 1;
  return delim_box (c, n, fn, y, col);
}

box
tex_rubber_font_rep::make_wide (unsigned char c, color col, SI X1, SI X2) {
  int n= get_wide_n (c, X1, X2);
  return delim_box (c, n, fn, 0, col);
}

box
tex_rubber_font_rep::make_bigop (unsigned char c, color col, int n) {
  return bigop_box (c, n, fn, col);
}

box
tex_rubber_font_rep::convert (string s, color col, SI y1, SI y2) {
  // WARNING: no checking takes place whether the character actually exists
  //          in the TeX font in get_*_n and displaying the rubber box
  unsigned char c= ext->dict[s][0];
  if ((N(s)>5) && (s(0,5)=="<big-")) return make_bigop (c, col, y1);
  if ((N(s)>7) && (s(0,7)=="<large-")) return make_delim (c, col, y1, y2);
  if ((N(s)>6) && (s(0,6)=="<wide-")) return make_wide (c, col, y1, y2);
  if (s == "<sqrt>") return make_sqrt (c, col, y1, y2);

  cout << "Trying to convert " << s << "\n";
  fatal_error ("Invalid token for rubber font",
	       "tex_rubber_font_rep::convert");
}

#endmodule // code_tex_rubber_font
