
/******************************************************************************
* MODULE     : liner.gen.h
* DESCRIPTION: Typesetting paragraphs in three stages.
*                - produce an array of line items
*                - handle brackets and scripts
*                - hyphenation and return of page items
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public licence and comes WITHOUT
* ANY WARRENTY WHATSOEVER. See the file $TEXMACS_PATH/LICENCE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <Atom/typesetter.gen.h>
#include <Paragraph/line_item.gen.h>
#include <Document/paragraph.gen.h>

#module liner
#import typesetter
#import line_item
#import paragraph

#define MODE_JUSTIFY   0
#define MODE_LEFT      1
#define MODE_CENTER    2
#define MODE_RIGHT     3

class liner_rep:public typesetter_rep {
  array<line_item>     a;          // the line items (pass1 & pass 2)
  paragraph            par;        // the typesetted paragraph (pass3)
  hashmap<string,tree> style;      // the style parameters

  double       level1;     // level of subscripts
  double       level2;     // level of superscripts

  array<box>   items;      // the boxes on the line in order
  array<path>  items_st;   // the start paths of the boxes in the lines
  array<path>  items_en;   // the end paths of the boxes in the lines
  array<int>   items_fl;   // border flags for the boxes in the lines
  array<SI>    items_sp;   // the final spaces between boxes in lines
  array<path>  scripts;    // extra information for handling scripts
  array<int>   tabs;       // tab positions
  array<space> spcs;       // the spacing between the boxes of bs
  SI           cur_r;      // the current right offset of the last line unit
  space        cur_w;      // the current width of the line unit
  int          cur_start;  // index of the start of the line unit

  string       mode;       // justified, left, center or right
  string       hyphen;     // hyphenation quality
  SI           width;      // the width of the parbox
  SI           first;      // indentation of first line (added to left)
  SI           last;       // indentation of last line (substracted from right)
  SI           left;       // left indentation of paragraph lines
  SI           right;      // right indentation of paragraph lines
  SI           height;     // the minimal height of each line
  SI           sep;        // min. separation between ink of successive lines
  SI           hor_sep;    // min. hor. sep. when lines are shoved together
  space        line_spc;   // space between consecutive lines
  space        par_spc;    // space between consecutive paragraphs
  SI           bot;        // next line will not ascend higher than this
  SI           top;        // previous line will not descend lower than this
  SI           tab_sep;    // separation between columns in tabular

  // first pass routines
  void print (int type, box b, path p, path q, int data=0);
  void print (int type, box b, path p, string s, font fn);
  void control (tree t, path p);
  void ghost (string s, path p, int f=0);
  void ghost (string s, path p, int f, color col);
  void print (space spc);
  void italic_correction (box b);
  void penalty_min (int p);
  void penalty_max (int p);

  void typeset_substring (string s, path p);
  void typeset_bigop (string s, path p);
  void typeset_bracket (tree t, path p);
  void typeset_prime (string s, path p);
  void typeset_string (tree t, path p);
  void typeset_concat (tree t, path p);
  void typeset_hspace (tree t, path p);
  void typeset_script (tree t, path p, bool right);

  void typeset_inactive (tree t, path p);
  void typeset_inactive (string s, tree t, path p, int pos1=0, int pos2=0);
  void typeset_inactive_symbol (tree t, path p);
  void typeset_inactive_latex (tree t, path p);
  void typeset_inactive_hybrid (tree t, path p);
  void typeset_inactive_apply (tree t, path p);
  void typeset_inactive_string (string s, path p);
  void typeset_unknown (string which, tree t, path p, bool flag= FALSE);
  void typeset_set (tree t, path p);
  void typeset_reset (tree t, path p);
  void typeset_assign (tree t, path p);
  void typeset_apply (tree t, path p);
  void typeset_begin (tree t, path p);
  void typeset_end (tree t, path p);
  void typeset_executable (tree t, path p);
  void typeset_label (tree t, path p);
  void typeset_reference (tree t, path p);
  void typeset_write (tree t, path p);
  void typeset (tree t, path p, bool active_flag= TRUE);

  // second pass routines
  int  prec (int i);
  int  succ (int i);
  void pre_glue ();
  bool script_flag ();
  void glue (box b, int ref, int arg);
  void glue (box b, int ref, int arg1, int arg2);
  void eliminate_obsolete ();
  void handle_scripts (int start, int end);
  void handle_matching (int start, int end);
  void handle_brackets ();
  void kill_spaces ();
  void pass2 ();

  // third pass routines
  path treat_glue (line_item item);
  void line_print (line_item item);
  void line_print (line_item item, path start, path end);
  void line_print (path start, path end);

  void line_start ();
  void make_unit (string mode, SI the_width, bool break_flag);
  void line_unit (path start, path end, bool break_flag,
		  string mode, SI the_left, SI the_right);
  void line_end (space spc, int penalty);
  void line_units (int start, int end, bool is_start, bool is_end,
		   string mode, string hyphen,
		   SI the_left, SI the_right, SI the_first, SI the_last);

  void get_env_vars ();
  array<SI> compute_tabular_widths (int split, int start, int end);
  void typeset_table (int split, int start, int end, bool isstart, bool isend);
  void typeset_paragraph_unit (int start, int end);
  void pass3 ();

public:

  liner_rep (edit_env env);
  void typeset_2_passes (tree t);
  box  make_concat ();
  void typeset_3_passes (tree t);

  friend class par_line_rep;
  friend class typesetter_rep;
  friend class document_rep;
};

typedef liner_rep* liner;

#endmodule // liner
